#pragma once

// Frame library dependencies
#include <FrameL.h>

// ROOT+ library dependencies
#include <TClock.h>
#include <TPicasso.h>
#include <TVarexp.h>
#include "../TFrameTypes.h"

// thread dependencies
#include <thread>

#include <tuple>
#include <complex>
#include <utility>

// abi:cxx dependency
#include <cxxabi.h>

struct FrTrendData {
  char *name;      /* Channel name */
  FrAdcData *inadc;    /* Input ADC data */
  FrAdcData *nADC, *minADC, *maxADC, *meanADC, *rmsADC;
  int n;
  float min;
  float max;
  double sum;
  double sumsq;
  struct FrTrendData *next;
};

FrameH *FrNewFrame(const char *name, double GPStime, double frameStep);
std::vector<TString> FrFileNames(const char *pattern, int nFrames, double GPStime, int frameStep);
std::vector<TString> FrFileNames(FrFile *file);
FrFile *FrFileAllocation(
        const char *prefix, const char *option, 
        const char *title, int nFrames, double GPStime, int frameStep, int compression
);

FrTrendData *FrTrendDataNew(int fs, FrAdcData *);
void FrTrendDataFree(FrTrendData *trendData);

int FrScanFile(FrFile *file, bool decompress = false);
int FrScanFile(FrFile *file, bool decompress, double gpstime0, double gpstime1);


/* Based on MyFrVectCmp from Frame_Compare_Channels.c by Ben Johnson */
int FrVectDiff(FrVect *vect0, FrVect *vect1, char* name1, char* name2, int debugLvl, double minSampleRate, bool useMinLength);

void FrameDump(FrameH *frame, int level = 1);

// A helper to check if T is a complex type
template <typename T>
struct is_complex : std::false_type {};

template <typename T>
struct is_complex<std::complex<T>> : std::true_type {};

template <typename T>
inline constexpr bool is_complex_v = is_complex<T>::value;

template<typename T2, typename T3 = T2, typename T1>
inline std::vector<T3> VcCast(std::vector<T1> v, int n = -1, bool realonly = false)
{
        if (n < 0 || n > v.size()) n = v.size();

        std::vector<T2> casted;
        casted.reserve((n * sizeof(T1)) / sizeof(T2));

        const char* byte_data = reinterpret_cast<const char*>(v.data());
        casted.resize((n * sizeof(T1)) / sizeof(T2));

        for (size_t i = 0; i < (n * sizeof(T1)) / sizeof(T2); ++i) {
                casted[i] = *reinterpret_cast<const T2*>(byte_data + i * sizeof(T2));
        }

        if constexpr (std::is_same_v<T2, T3>) return casted;
        else if constexpr (std::is_arithmetic_v<T2> && is_complex_v<T3>) {

                std::vector<T3> result;
                result.reserve(casted.size());
                for (const auto& element : casted) {
                        result.push_back(T3(element, T2(0)));
                }

                return result;
        
        } else if constexpr (is_complex_v<T2> && std::is_arithmetic_v<T3>) {
                
                std::vector<T3> result;
                result.reserve(realonly ? casted.size() : 2 * casted.size());
                for (const auto& element : casted) {
                        if (realonly) {
                                result.push_back(element.real());
                        } else {
                                result.push_back(element.real());
                                result.push_back(element.imag());
                        }
                }
                return result;

        } else {
                return std::vector<T3>(casted.begin(), casted.end());
        }
}

FrVect* FrVectNew(int type, std::vector<unsigned int> nx, std::vector<double> dx, std::vector<std::string> unit = {});
std::vector<double> FrVectRead(FrVect *ptrFr);
TString FrVectType(FrVect *ptrFr);
TString FrVectType(FrAdcData *ptrFr);

int FrCompressionLevel(FrFile *file);

inline std::vector<double> FrReadData(FrVect *ptrFr) { return FrVectRead(ptrFr); }

FrameFormat::TFrAdcData  *RootCast(FrAdcData  *ptrFr);
FrameFormat::TFrDetector *RootCast(FrDetector *ptrFr);
FrameFormat::TFrEvent    *RootCast(FrEvent    *ptrFr);
FrameFormat::TFrHistory  *RootCast(FrHistory  *ptrFr);
FrameFormat::TFrMsg      *RootCast(FrMsg      *ptrFr);
FrameFormat::TFrProcData *RootCast(FrProcData *ptrFr);
FrameFormat::TFrRawData  *RootCast(FrRawData  *ptrFr);
FrameFormat::TFrSerData  *RootCast(FrSerData  *ptrFr);
FrameFormat::TFrSimData  *RootCast(FrSimData  *ptrFr);
FrameFormat::TFrSimEvent *RootCast(FrSimEvent *ptrFr);
FrameFormat::TFrStatData *RootCast(FrStatData *ptrFr);
FrameFormat::TFrSummary  *RootCast(FrSummary  *ptrFr);
FrameFormat::TFrTable    *RootCast(FrTable    *ptrFr);
FrameFormat::TFrVect     *RootCast(FrVect     *ptrFr);

FrAdcData  *FrCast(FrameFormat::TFrAdcData  *rootObject);
FrDetector *FrCast(FrameFormat::TFrDetector *rootObject);
FrEvent    *FrCast(FrameFormat::TFrEvent    *rootObject);
FrHistory  *FrCast(FrameFormat::TFrHistory  *rootObject);
FrMsg      *FrCast(FrameFormat::TFrMsg      *rootObject);
FrProcData *FrCast(FrameFormat::TFrProcData *rootObject);
FrRawData  *FrCast(FrameFormat::TFrRawData  *rootObject);
FrSerData  *FrCast(FrameFormat::TFrSerData  *rootObject);
FrSimData  *FrCast(FrameFormat::TFrSimData  *rootObject);
FrSimEvent *FrCast(FrameFormat::TFrSimEvent *rootObject);
FrStatData *FrCast(FrameFormat::TFrStatData *rootObject);
FrSummary  *FrCast(FrameFormat::TFrSummary  *rootObject);
FrTable    *FrCast(FrameFormat::TFrTable    *rootObject);
FrVect     *FrCast(FrameFormat::TFrVect     *rootObject);

inline int    FrDebugLvl;
void FrLibSetLvlT(int level);

int FrBits(FrameFormat::TFrVect::TYPE type);

FrameH *FrameReadT(FrFile *iFile, double gtime, const char *tag);
FrEndOfFile *FrEndOfFileReadAndReturn(FrFile *iFile);
FrEndOfFrame *FrEndOfFrameReadAndReturn(FrFile *iFile);