#pragma once

#include "FrameFormat/Abstract/TFrameAbstract.h"
#include "FrameFormat/TFrameProxy.h"

namespace FrameFormat {

        struct GPS {
    
                double start, end;
    
                GPS(double _start = 0, double _end = 0) : start(_start), end(_end) {}
    
                bool ConnectsTo(const GPS& other) const {
                        return !(other.end < start || other.start > end);
                }
                
                GPS ProtractsWith(const GPS& other) const {

                        double _start = ROOT::IOPlus::Helpers::EpsilonEqualTo(start, 0.) ? other.start : start;
                        double _end = ROOT::IOPlus::Helpers::EpsilonEqualTo(end, 0.) ? other.end : end;

                        return GPS(std::min(_start, other.start), std::max(_end, other.end));
                }
    
                // Equality: A == B if both start and end match within tolerance
                inline bool operator==(const GPS &other) const {
                        return ROOT::IOPlus::Helpers::EpsilonEqualTo(start, other.start) && ROOT::IOPlus::Helpers::EpsilonEqualTo(end, other.end);
                }
    
                // Inequality: A != B if either start or end does not match within tolerance
                inline bool operator!=(const GPS &other) const {
                        return !(*this == other);
                }
    
                // Strict Less Than: A < B if A ends strictly before B starts (no overlap)
                inline bool operator<(const GPS &other) const {
                        return end < other.start;
                }
    
                // Strict Greater Than: A > B if A starts strictly after B ends (no overlap)
                inline bool operator>(const GPS &other) const {
                        return start > other.end;
                }
    
                // Less Than or Equal to: A <= B if A starts before or overlaps B
                inline bool operator<=(const GPS &other) const {
                        return start <= other.end;
                }
    
                // Greater Than or Equal to: A >= B if A ends after or overlaps B
                inline bool operator>=(const GPS &other) const {
                        return end >= other.start;
                }
    
                // Includes: Returns true if this frame fully contains another frame
                inline bool Includes(const GPS &other) const {
                        return start <= other.start && end >= other.end;
                }
    
                // Overlaps: Returns true if this frame overlaps with another frame
                inline bool Overlaps(const GPS &other) const {
                        return (start < other.end) && (end > other.start);
                }
    
                // The frame is strictly before the GPS time
                inline bool operator<(double gpsTime) const {
                        return end < gpsTime;
                }
    
                // The frame is strictly after the GPS time
                inline bool operator>(double gpsTime) const {
                        return start > gpsTime;
                }
    
                // The frame starts before or includes the GPS time
                inline bool operator<=(double gpsTime) const {
                        return start <= gpsTime;
                }
    
                // The frame ends after or includes the GPS time
                inline bool operator>=(double gpsTime) const {
                        return end >= gpsTime;
                }
        };
    
        class TFrame;
        class TFrameManager;
        class TFrameChain: public TFrameAbstract
        {
                using FrameContainer = std::vector<std::vector<TFrame*>>;

                class Iterator {

                        public:
                        using OuterIt = FrameContainer::iterator;
                        using InnerIt = std::vector<TFrame*>::iterator;
                
                        Iterator(OuterIt outer, OuterIt outerEnd, InnerIt inner): outer_(outer), outerEnd_(outerEnd), inner_(inner) {
                                next();
                        }
                
                        TFrame* operator*() const { return *inner_; }
                        TFrame* operator->() const { return *inner_; }
                
                        Iterator& operator++() {
                                ++inner_;
                                next();
                                return *this;
                        }
                
                        bool operator!=(const Iterator& other) const {
                                return outer_ != other.outer_ || (outer_ != outerEnd_ && inner_ != other.inner_);
                        }
                
                        private:
                        OuterIt outer_, outerEnd_;
                        InnerIt inner_;

                        void next() {
                                while (outer_ != outerEnd_ && inner_ == outer_->end()) {

                                        ++outer_;
                                        if (outer_ != outerEnd_)
                                                inner_ = outer_->begin();
                                }
                        }
                };

                protected:
                        
                        friend class TFrameManager;
                        TFrameManager *fManager = NULL;
                        int fileID = 0;
                        int frameID = 0;

                        int padWidth = 892;
                        int padHeight = 512;
                         
                        using TObject::Dump;
                        using TObject::Print;
                        using TObject::Write;

                        TString fName;
                        int compression = 0;
                        FrameContainer frames;
                        
                        Option_t *option;
                        TString tags;
                        std::vector<FrType> dataClasses;
                        int decimate;
                        std::vector<int> channelGroups;
                        std::vector<int> channelNumbers;

                public:

                        TFrameChain() = default;
                        TFrameChain(std::vector<TFrame*> frames, Verbosity verbosity = Verbosity::kUnchanged);
                        TFrameChain(const char *fName, Option_t *option = "", TString tags = "*", std::vector<FrType> dataClasses = {}, int decimate = 1, std::vector<int> channelGroups = {}, std::vector<int> channelNumbers = {}, Verbosity verbosity = Verbosity::kUnchanged);
                        TFrameChain(std::vector<TString> fName, Option_t *option = "", TString tags = "*", std::vector<FrType> dataClasses = {}, int decimate = 1, std::vector<int> channelGroups = {}, std::vector<int> channelNumbers = {}, Verbosity verbosity = Verbosity::kUnchanged);

                        TFrameChain &Reallocate();
                        ~TFrameChain() { Close(); };


                        // Access frames using global index
                        TFrame* operator[](size_t globalIndex) {
                                size_t currentIndex = 0;
                                for (auto& row : frames) {
                                        if (globalIndex < currentIndex + row.size()) {
                                                return row[globalIndex - currentIndex];
                                        }
                                        currentIndex += row.size();
                                }
                                return nullptr;
                        }

                        const TFrame* operator[](size_t globalIndex) const {
                                size_t currentIndex = 0;
                                for (const auto& row : frames) {
                                        if (globalIndex < currentIndex + row.size()) {
                                                return row[globalIndex - currentIndex];
                                        }
                                        currentIndex += row.size();
                                }
                                return nullptr;
                        }

                        Iterator begin() {
                                return frames.empty() ? Iterator(frames.end(), frames.end(), {}) 
                                                      : Iterator(frames.begin(), frames.end(), frames.begin()->begin());
                        }
                
                        Iterator end() {
                                return frames.empty() ? Iterator(frames.end(), frames.end(), {}) 
                                                      : Iterator(frames.end(), frames.end(), frames.end()->begin());
                        }

                        TFrameChain& Open(std::vector<TString> fName, Option_t *option = "", TString tags = "*", std::vector<FrType> dataClasses = {}, int decimate = 1, std::vector<int> channelGroups = {}, std::vector<int> channelNumbers = {}, Verbosity verbosity = Verbosity::kUnchanged);
                        TFrameChain& Open(const char *fName, Option_t *option = "", TString tags = "*", std::vector<FrType> dataClasses = {}, int decimate = 1, std::vector<int> channelGroups = {}, std::vector<int> channelNumbers = {}, Verbosity verbosity = Verbosity::kUnchanged);
                        bool IsOpen();
                        void Close();

                        inline TFrameManager &GetManager() { return *this->fManager; }
                        TString GetFileName(int fileID = -1) const;
                        int GetNFrames(int fileID = -1) const ;
                        int GetNFiles() const ;
                        int GetEntries() const ;

                        std::vector<TFrame *> FindByGPS(GPS gps);
                        std::vector<TFrame *> FindByFileName(TString filename);

                        GPS GetGPS(int fileID = -1, int frameID = -1) const ;
                        TString GetDetector  (int fileID = -1, int frameID = -1) const ;
                        double GetTimeLength(int fileID = -1, int frameID = -1) const ;

                        bool VerifyChecksum(bool decompress = false);

                        int GetFileID(int globalID) const ;
                        std::vector<int> GetFileID(std::vector<int> globalID) const ;
                        int GetFrameID(int globalID) const ;
                        std::vector<int> GetFrameID(std::vector<int> globalID) const ;

                        int GetGlobalID(int fileID, int localID) const;
                        int GetGlobalID(TFrame *frame) const;

                        TFrame    *GetFrame(GPS) const;
                        TFrame    *GetFrame(int globalID = -1) const;
                        TFrame    *GetFrame(int fileID, int frameID) const;

                        inline int GetCompression() const { return this->compression; }
                        void SetCompression(int compression) { this->compression = compression; }

                        void Free();

                        std::vector<int> Select          (TString formula, std::vector<double> params = {});
                        std::vector<int> SelectByFrame(std::vector<int> globalIDs);
                        std::vector<int> SelectByFrame(int firstGlobalID = -1, int lastGlobalID = -1);
                        std::vector<int> SelectByGPS     (std::vector<double> gps);
                        std::vector<int> SelectByGPS     (GPS gps = GPS());
                        std::vector<int> SelectByGPS     (double gpstime0, double gpstime1) { return this->SelectByGPS(GPS(gpstime0, gpstime1)); };
                        std::vector<int> SelectByMaxDuration(double duration);
                        std::vector<int> SelectByMinDuration(double duration);
                        std::vector<int> SelectByDQ      (unsigned int dataQuality);

                        void Print(Option_t *opt = "", int globalID = -1);
                        void Print(Option_t *opt, std::vector<int> globalIDs);

                        inline void SetPadHeight(int padHeight) { this->padHeight = padHeight; }
                        inline void SetPadWidth(int padWidth) { this->padWidth = padWidth; }

                        void DumpTOC();
                        void DumpTOC(int globalID);

                        bool AddFriend(TFrameChain *c);
                        void RemoveFriend(TFrameChain *c);

                        void Reload();
                        void Set(TString name);
                        TString Get() const;

                        void Dump(std::vector<int> globalIDs);
                        void Dump(int globalID = -1);

                        bool IsContiguous();

                        TObject* GetObject(TString objName, GPS gpstime) { return this->GetObject(objName, this->SelectByGPS(gpstime)); }
                        TObject* GetObject(TString objName, std::vector<int> globalIDs = {});
                        TObject* GetObject(TString objName, TString branchName, Option_t *option, GPS gpstime) { return this->GetObject(objName, branchName, option, this->SelectByGPS(gpstime)); }
                        TObject* GetObject(TString objName, TString branchName, Option_t *option = "", std::vector<int> globalIDs = {});

                        inline std::vector<TString> Write(TString filename, Option_t *opt,            int framesPerFile, GPS gpstime) { return this->Write  (filename, opt, framesPerFile, this->SelectByGPS(gpstime) ); };
                               std::vector<TString> Write(TString filename, Option_t *opt,            int framesPerFile, std::vector<int> globalIDs);
                               std::vector<TString> Write(TString filename, Option_t *opt = "update", int framesPerFile = 1);

                        static TFrameChain *Allocate(
                                const char *fname, const char *detector, const char *title, 
                                int nFrames, double gpstime, int timeLength, 
                                int compression = 1, Verbosity verbosity = Verbosity::kUnchanged
                        );

                        TFrameChain *Merge(TFrameChain *f, GPS gpstime = GPS());
                        inline TFrameChain *Extract(GPS gpstime) { return this->Extract(this->SelectByGPS(gpstime)); }
                               TFrameChain *Extract(std::vector<int> globalIDs = {});
                        inline TFrameChain *Trend  (int fs, Option_t *opt, GPS gpstime) { return this->Trend  (fs, opt, this->SelectByGPS(gpstime), gpstime); }
                               TFrameChain *Trend  (int fs, Option_t *opt = "", std::vector<int> globalIDs = {}, GPS gpstime = GPS() );
                        inline TFrameChain *Reshape(double dt, Option_t *opt, GPS gpstime) { return this->Reshape(dt, opt, this->SelectByGPS(gpstime), gpstime); }
                               TFrameChain *Reshape(double dt, Option_t *opt = "", std::vector<int> globalIDs = {}, GPS gpstime = GPS() );
  
                        TFrame *AddFrame(int nFrames, int fileID = -1);
                        TFrame *AddFrame(TFrame *frame);
                        std::vector<TFrame *> AddFrame(std::vector<TFrame *> frames);

                        TFrame *RemoveFrame(int fileID, int frameID);
                        TFrame *RemoveFrame(TFrame *frame);
                        std::vector<TFrame *> RemoveFrame(std::vector<TFrame *> frames);

                        template <typename T>
                        bool AddChannel(const char *channel, double fs, std::vector<T> adcData, double timestamp = -1, TString unitX = "Time (in s)", TString unitY = "");
                        void RemoveChannel(const char *channel);
                        
                        TFrameChain &Commit(Option_t *opt = "", int framesPerFile = 1);

                protected:

                        std::vector<TString> WriteGWF (TString name, Option_t *opt, GPS gpstime) { return this->WriteGWF(name, opt, this->SelectByGPS(gpstime)); }
                        std::vector<TString> WriteGWF (TString name, Option_t *opt = "", std::vector<int> globalIDs = {});
                        std::vector<TString> WriteROOT(TString name, Option_t *opt, GPS gpstime) { return this->WriteROOT(name, opt, this->SelectByGPS(gpstime)); }
                        std::vector<TString> WriteROOT(TString name, Option_t *opt = "", std::vector<int> globalIDs = {});
                        static std::vector<TString> WriteROOT(TFile *f, TFrame *frame);

                ClassDef(TFrameChain,1);
        };

        using TFrameFile = TFrameChain;
}

inline std::ostream& operator<<(std::ostream& os, const FrameFormat::GPS& gps) {

        if(ROOT::IOPlus::Helpers::EpsilonEqualTo(gps.end, 0.)) {
                os << Form("GPS %.2f", gps.start);
        } else {
                os << Form("GPS [%.2f - %.2f]", gps.start, gps.end);
        }
        return os;
}