#pragma once 

#include <string>
#include <vector>
#include <typeindex>
#include <variant>

// CHAR   = char           = 1 byte
// CHAR_U = unsigned char  = 1 byte
// INT_2S = short          = 2 bytes
// INT_2U = unsigned short = 2 bytes
// INT_4S = int            = 4 bytes
// INT_4U = unsigned int   = 4 bytes
// INT_8S = long          = 8 bytes
// INT_8U = unsigned long = 8 bytes
// REAL_4 = float         = 4 bytes
// REAL_8 = double        = 8 bytes
// STRING = char[]          = < 65536 bytes
// DOUBLE = double          = 6 bytes
// COMPLEX_8  = <float,float>   = 8 bytes
// COMPLEX_16 = <double,double> = 16 bytes

namespace FrameFormat {

        enum class FrType
        {
                FrVect   , FrTable   , FrDetector,
                FrRawData, FrProcData, FrSimData , FrStatData,
                FrEvent  , FrSimEvent, FrSummary ,
                FrAdcData, FrSerData , FrHistory , FrMsg

        };

        struct TFrAbstract { 

                std::string name; 
                virtual TFrAbstract& operator=(const TFrAbstract& other) {
                        if (this != &other) { name = other.name; }
                        return *this;
                }
        };

        struct TFrVect: public TFrAbstract
        {
                // @TODO: a safe solution for handling TFrVect, and have access to ROOT display features
                //        would be to provide std::span interface or operator.
                std::vector<double> data;

                short compress;

                enum class TYPE { 
                        FR_VECT_C = 0, FR_VECT_2S = 1, FR_VECT_8R = 2, FR_VECT_4R = 3, 
                        FR_VECT_4S = 4, FR_VECT_8S = 5, FR_VECT_8C = 6, FR_VECT_16C = 7, 
                        FR_VECT_STRING = 8, FR_VECT_2U = 9, FR_VECT_4U = 10,FR_VECT_8U = 11,
                        FR_VECT_1U = 12 
                };
                TYPE type;

                unsigned long nData;
                unsigned long nBytes;

                unsigned int nDim;
                std::vector<unsigned int> nx;
                std::vector<double> dx;
                std::vector<double> startX;
                std::vector<std::string> unitX;

                std::string unitY;

                TFrVect& operator=(const TFrVect& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                compress = other.compress;
                                type = other.type;
                                nData = other.nData;
                                nBytes = other.nBytes;
                                data = other.data;
                                nDim = other.nDim;
                                nx = other.nx;
                                dx = other.dx;
                                startX = other.startX;
                                unitX = other.unitX;
                                unitY = other.unitY;
                        }
                        return *this;
                }
        };

        struct TFrTable: public TFrAbstract
        {
                std::string comment;
                unsigned short nColumn;
                unsigned int nRow;

                std::vector<std::string> columnName;
                TFrVect column;

                TFrTable& operator=(const TFrTable& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                nColumn = other.nColumn;
                                nRow = other.nRow;
                                columnName = other.columnName;
                                column = other.column;
                        }
                        return *this;
                }
        };

        struct TFrAdcData: public TFrAbstract
        {
                std::string comment;
                unsigned int channelGroup;
                unsigned int channelNumber;

                unsigned int nBits;
                float bias;
                float slope;
                std::string units;

                double sampleRate;
                double timeOffset;
                double fShift;
                float phase;

                unsigned short dataValid;

                TFrVect data;
                TFrVect aux;

                TFrAdcData& operator=(const TFrAdcData& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                channelGroup = other.channelGroup;
                                channelNumber = other.channelNumber;
                                nBits = other.nBits;
                                bias = other.bias;
                                slope = other.slope;
                                units = other.units;
                                sampleRate = other.sampleRate;
                                timeOffset = other.timeOffset;
                                fShift = other.fShift;
                                phase = other.phase;
                                dataValid = other.dataValid;
                                data = other.data;
                                aux = other.aux;
                        }
                        return *this;
                }
        };

        struct TFrDetector: public TFrAbstract
        {
                char  prefix[2];

                double longitude;
                double latitude;
                float elevation;

                float armXazimuth;
                float armYazimuth;

                float armXaltitude;
                float armYaltitude;

                float armXmidpoint;
                float armYmidpoint;

                int localTime;

                TFrVect aux;
                TFrTable table;

                TFrDetector& operator=(const TFrDetector& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                std::copy(std::begin(other.prefix), std::end(other.prefix), std::begin(prefix));
                                longitude = other.longitude;
                                latitude = other.latitude;
                                elevation = other.elevation;
                                armXazimuth = other.armXazimuth;
                                armYazimuth = other.armYazimuth;
                                armXaltitude = other.armXaltitude;
                                armYaltitude = other.armYaltitude;
                                armXmidpoint = other.armXmidpoint;
                                armYmidpoint = other.armYmidpoint;
                                localTime = other.localTime;
                                aux = other.aux;
                                table = other.table;
                        }
                        return *this;
                }
        };

        struct TFrEvent: public TFrAbstract
        {
                std::string comment;
                std::string inputs;

                unsigned int GTimeS;
                unsigned int GTimeN;
                float timeBefore;
                float timeAfter;
                unsigned int eventStatus;

                float amplitude;
                float probability;
                std::string statistics;

                unsigned int nParam;
                std::vector<double> parameters;
                std::vector<std::string> parameterNames;

                TFrVect data;
                TFrTable table;

                TFrEvent& operator=(const TFrEvent& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                inputs = other.inputs;
                                GTimeS = other.GTimeS;
                                GTimeN = other.GTimeN;
                                timeBefore = other.timeBefore;
                                timeAfter = other.timeAfter;
                                eventStatus = other.eventStatus;
                                amplitude = other.amplitude;
                                probability = other.probability;
                                statistics = other.statistics;
                                nParam = other.nParam;
                                parameters = other.parameters;
                                parameterNames = other.parameterNames;
                                data = other.data;
                                table = other.table;
                        }
                        return *this;
                }
        };

        struct TFrHistory: public TFrAbstract
        {
                unsigned int time;
                std::string comment;

                TFrHistory& operator=(const TFrHistory& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                time = other.time;
                                comment = other.comment;
                        }
                        return *this;
                }
        };

        struct TFrMsg: public TFrAbstract
        {
                std::string& alarm = name; // Make alarm a reference to name
                std::string message;
                unsigned int severity;
                unsigned int GTimeS;
                unsigned int GTimeN;

                TFrMsg& operator=(const TFrMsg& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                message = other.message;
                                severity = other.severity;
                                GTimeS = other.GTimeS;
                                GTimeN = other.GTimeN;
                        }
                        return *this;
                }
        };

        struct TFrProcData: public TFrAbstract
        {
                std::string comment;

                enum class TYPE { Unknown = 0, TimeSeries = 1, FrequencySeries = 2, OtherSeries = 3, TimeFrequency = 4, Wavelets = 5, MultiDimensional = 6 };
                TYPE type;

                enum class SUBTYPE { Unknown = 0, DFT = 1, ASD = 2, PSD = 3, CSD = 4, Coherence = 5, TF = 6};
                SUBTYPE subType;

                double timeOffset;
                double tRange;
                double fShift;
                float phase;
                double fRange;
                double BW;

                unsigned short nAuxParam;
                std::vector<double> auxParam;
                std::vector<std::string> auxParamNames;

                TFrVect data;
                TFrVect aux;
                TFrTable table;
                TFrHistory history;

                TFrProcData& operator=(const TFrProcData& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                type = other.type;
                                subType = other.subType;
                                timeOffset = other.timeOffset;
                                tRange = other.tRange;
                                fShift = other.fShift;
                                phase = other.phase;
                                fRange = other.fRange;
                                BW = other.BW;
                                nAuxParam = other.nAuxParam;
                                auxParam = other.auxParam;
                                auxParamNames = other.auxParamNames;
                                data = other.data;
                                aux = other.aux;
                                table = other.table;
                                history = other.history;
                        }
                        return *this;
                }
        };

        struct TFrSerData: public TFrAbstract
        {
                unsigned int timeSec;
                unsigned int timeNsec;
                double sampleRate;

                std::string data;
                TFrVect serial;
                TFrTable table;

                TFrSerData& operator=(const TFrSerData& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                timeSec = other.timeSec;
                                timeNsec = other.timeNsec;
                                sampleRate = other.sampleRate;
                                data = other.data;
                                serial = other.serial;
                                table = other.table;
                        }
                        return *this;
                }
        };

        struct TFrRawData: public TFrAbstract
        {
                std::string name;
                TFrMsg logMsg;
                TFrVect more;

                // TFrTable *firstTable;
                TFrRawData& operator=(const TFrRawData& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                name = other.name;
                                logMsg = other.logMsg;
                                more = other.more;
                        }
                        return *this;
                }
        };

        struct TFrSimData: public TFrAbstract
        {
                std::string comment;

                double sampleRate;
                double timeOffset;
                double fShift;
                float phase;

                TFrVect data;
                TFrVect input;
                TFrTable table;

                TFrSimData& operator=(const TFrSimData& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                sampleRate = other.sampleRate;
                                timeOffset = other.timeOffset;
                                fShift = other.fShift;
                                phase = other.phase;
                                data = other.data;
                                input = other.input;
                                table = other.table;
                        }
                        return *this;
                }
        };

        struct TFrSimEvent: public TFrAbstract
        {
                std::string comment;
                std::string inputs;

                unsigned int GTimeS;
                unsigned int GTimeN;
                float timeBefore;
                float timeAfter;
                float amplitude;

                unsigned int nParam;
                std::vector<double> parameters;
                std::vector<std::string> parameterNames;
                TFrVect data;
                TFrTable table;

                TFrSimEvent& operator=(const TFrSimEvent& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                inputs = other.inputs;
                                GTimeS = other.GTimeS;
                                GTimeN = other.GTimeN;
                                timeBefore = other.timeBefore;
                                timeAfter = other.timeAfter;
                                amplitude = other.amplitude;
                                nParam = other.nParam;
                                parameters = other.parameters;
                                parameterNames = other.parameterNames;
                                data = other.data;
                                table = other.table;
                        }
                        return *this;
                }
        };

        struct TFrStatData: public TFrAbstract
        {
                std::string comment;
                std::string representation;

                unsigned int timeStart;
                unsigned int timeEnd;
                unsigned int version;

                TFrDetector detector;
                TFrVect data;
                TFrTable table;

                TFrStatData& operator=(const TFrStatData& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                representation = other.representation;
                                timeStart = other.timeStart;
                                timeEnd = other.timeEnd;
                                version = other.version;
                                detector = other.detector;
                                data = other.data;
                                table = other.table;
                        }
                        return *this;
                }
        };

        struct TFrSummary: public TFrAbstract
        {
                std::string comment;
                std::string test;

                unsigned int GTimeS;
                unsigned int GTimeN;

                TFrVect moments;
                TFrTable table;

                TFrSummary& operator=(const TFrSummary& other) {
                        if (this != &other) {
                                TFrAbstract::operator=(other);
                                comment = other.comment;
                                test = other.test;
                                GTimeS = other.GTimeS;
                                GTimeN = other.GTimeN;
                                moments = other.moments;
                                table = other.table;
                        }
                        return *this;
                }
        };

        using TFrAbstractT = std::variant<
                FrameFormat::TFrAdcData*, 
                FrameFormat::TFrDetector*,
                FrameFormat::TFrEvent*,
                FrameFormat::TFrHistory*,
                FrameFormat::TFrMsg*,
                FrameFormat::TFrProcData*,
                FrameFormat::TFrRawData*,
                FrameFormat::TFrSerData*,
                FrameFormat::TFrSimData*,
                FrameFormat::TFrSimEvent*,
                FrameFormat::TFrStatData*,
                FrameFormat::TFrSummary*,
                FrameFormat::TFrTable*,
                FrameFormat::TFrVect*
        >;
}
