TCanvas *C;
TPad *p1, *p2, *p3;
TH1F *h1, *h2, *h3;

bool GetAxisIndex(TH1 *h, TAxis *axis) {

      if(h != NULL && axis != NULL)  {

            if(h->GetXaxis() == axis) return 0;
            if(h->GetYaxis() == axis) return 1;
            if(h->GetZaxis() == axis) return 2;
      }

      return -1;
}

TAxis* GetAxisByIndex(TH1* h, int index) {
    
   if (!h) return nullptr;
   switch (index) {
         case 0: return h->GetXaxis(); // X-axis
         case 1: return h->GetYaxis(); // Y-axis (for 2D histograms)
         case 2: return h->GetZaxis(); // Z-axis (for 3D histograms)
         default:
            std::cerr << "Invalid axis index!" << std::endl;
            return nullptr;
   }
}

TList* ExtractObjectListFrom(TList* list, TString className = "TObject", TString objName = "") {

    if (!list) {
        std::cerr << "Pad is null!" << std::endl;
        return nullptr;
    }

    TList* extractedList = new TList();
    for (auto obj : *list) {

         if( !obj->InheritsFrom(className) ) continue;
         if( !TString(obj->GetName()).EqualTo(objName) ) continue;
         if( TObject* h = dynamic_cast<TObject*>(obj) ) extractedList->Add(h);
    }

    for (auto obj : *list) {
        if (TVirtualPad* subPad = dynamic_cast<TVirtualPad*>(obj)) {
            TList* sublist = ExtractObjectListFrom(subPad->GetListOfPrimitives(), className, objName);
            if (sublist) {
                extractedList->AddAll(sublist);
                delete sublist;  // Clean up to avoid memory leak
            }
        }
    }

    return extractedList;
}

void MotherPadPropagator()
{
   if (gPad && gPad->GetMother() != gPad) {
      
      TList *execList = gPad->GetMother()->GetListOfExecs();
      if (execList) {

         for (int i = 0; i < execList->GetSize(); i++) {
            TNamed *exec = (TNamed*)execList->At(i);
            exec->Draw();
         }
      }
   }
}

void ComputeTimeAxis(TString hName, double t0 = 0, int precision = -1, TVirtualPad *pad = nullptr)
{
   pad = pad ? pad : gPad;

   TH1 *h = (TH1*) pad->FindObject(hName);
   if(h == NULL) return;

   int first = h->GetXaxis()->GetFirst();
   double ufirst = h->GetXaxis()->GetBinLowEdge(first);
   int last = h->GetXaxis()->GetLast();
   double ulast = h->GetXaxis()->GetBinUpEdge(last);
      
   Int_t divxo  = 0;
   Double_t bwx = 0.;
   Double_t x1o = 0.;
   Double_t x2o = 0.;

   if(h->GetXaxis()->GetNdivisions() <= 0 ) return;
   
   TString precision_str = precision < 0 ? "%f" : "%."+TString::Itoa(precision,10)+"f";
   THLimitsFinder::Optimize(ufirst, ulast, h->GetXaxis()->GetNdivisions()%100,x1o,x2o,divxo,bwx,"");
   for(int i = 0; i <= h->GetXaxis()->GetNlabels()+5; i++) {

      double bw = (x2o-x1o)/divxo;
      h->GetXaxis()->SetMaxDigits(1024); // remove 10^N
      h->GetXaxis()->ChangeLabel(i+1, -1, -1, -1, kRed, -1, Form(precision_str, x1o + i*bw - t0));
      h->GetXaxis()->SetTitle( Form("t0 = "+precision_str, t0) );
   }
}

void GlobalZoom()
{
   TVirtualPad *gPadBak = gPad;
   if (!gPadBak) return;

   gPad->GetMother()->cd();
   std::cout << "Selected pad:" << gPad->GetName() << std::endl;

   TList *l = ExtractObjectListFrom(gPad->GetListOfPrimitives(), "TH1");
          l->Print();

   Int_t event = gPad->GetEvent();
   int px = gPad->GetEventX();
   int py = gPad->GetEventY();
   TObject *select = gPad->GetSelected();

   if (select && select->InheritsFrom(TAxis::Class())) {

      TAxis *axis = (TAxis*) select;
      double first = axis->GetFirst();
      double last  = axis->GetLast();

      std::cout << axis << std::endl;
      std::cout << "Selected Axis: (" << first << "," << last << ")" << std::endl;
   }

   gPadBak->cd();
}

void zoomdiv2() {

   TCanvas *C = new TCanvas("C", "C", 500, 900);
   C->Divide(1,3,0,0);

   h1 = new TH1F("h1", "h1", 100, 1726982609, 1726982908) ; h1->FillRandom("gaus", 10000);
   h2 = new TH1F("h2", "h2", 100, 1726982609, 1726982908) ; h2->FillRandom("gaus", 10000);
   h3 = new TH1F("h3", "h3", 100, 1726982609, 1726982908) ; h3->FillRandom("gaus", 10000);

   p1 = (TPad*)C->cd(1); h1->Draw();
   p2 = (TPad*)C->cd(2); h2->Draw();
   p3 = (TPad*)C->cd(3); h3->Draw();

   h3->GetListOfFunctions()->Add(new TExec("ComputeTimeAxis", "ComputeTimeAxis(\""+TString(h3->GetName())+"\", 1726982709.5, 2);"));

   C->AddExec("GlobalZoom","GlobalZoom();");   
   TExec *exec = new TExec("MotherPadPropagator","MotherPadPropagator();"); 
          exec->Draw();
}
