/**
 **********************************************
 *
 * \file TFrameChain.cc
 * \brief Source code of the TFrameChain class
 * \author Marco Meyer \<marco.meyer@cern.ch\>
 *
 *********************************************
 */

#include "FrameFormat/TFrameChain.h"
#include "FrameFormat/TFrameManager.h"
#include "FrameFormat/TFrame.h"
ClassImp(FrameFormat::TFrameChain)

#include "FrameFormat/Impl/TFrameHeader.h"
#include <TFileReader.h>

namespace FrameFormat {

    TFrameChain::TFrameChain(std::vector<TString> fName, Option_t *option, TString tags, std::vector<FrType> dataClasses, int decimate, std::vector<int> channelGroups, std::vector<int> channelNumbers, Verbosity verbosity)
               : TFrameChain(ROOT::IOPlus::Helpers::Implode(" ", fName), option, tags, dataClasses, decimate, channelGroups, channelNumbers, verbosity) { }

    TFrameChain::TFrameChain(const char *fName, Option_t *option, TString tags, std::vector<FrType> dataClasses, int decimate, std::vector<int> channelGroups, std::vector<int> channelNumbers, Verbosity verbosity)
    {
        this->Open(fName, option, tags, dataClasses, decimate, channelGroups, channelNumbers, verbosity);
    }

    TFrameChain& TFrameChain::Open(std::vector<TString> fName, Option_t *option, TString tags, std::vector<FrType> dataClasses, int decimate, std::vector<int> channelGroups, std::vector<int> channelNumbers, Verbosity verbosity)
    {
        return this->Open(ROOT::IOPlus::Helpers::Implode(" ", fName), option, tags, dataClasses, decimate, channelGroups, channelNumbers, verbosity);
    }

    TFrameChain& TFrameChain::Open(const char *fName, Option_t *option, TString tags, std::vector<FrType> dataClasses, int decimate, std::vector<int> channelGroups, std::vector<int> channelNumbers, Verbosity verbosity)
    {
        if(this->IsOpen()) this->Close();

        if(verbosity > Verbosity::kUnchanged) this->SetVerbosity(verbosity);
        if(verbosity > Verbosity::kUnchanged) this->SetDebug(this->GetVerbosity());

        this->fName = ROOT::IOPlus::Helpers::Implode(" ", TFileReader::RecursiveExpandTextFile(fName));
        this->fName = ROOT::IOPlus::Helpers::Implode(" ", TFileReader::ExpandWildcardUrl(ROOT::IOPlus::Helpers::Explode(" ", this->fName)));

        TString _fName = "";
        for(auto &f : ROOT::IOPlus::Helpers::Explode(" ", this->fName)) {
            if(f.EndsWith(".ffl")) _fName += ROOT::IOPlus::Helpers::Implode(" ", TFileReader::ReadTextFile(f)) + " ";
            else _fName += f + " ";
        }

        this->fName          = _fName.Strip();
        this->option         = option;
        this->tags           = tags;
        this->dataClasses    = dataClasses;
        this->decimate       = decimate;
        this->channelGroups  = channelGroups;
        this->channelNumbers = channelNumbers;

        TString fNameStr = gPrint->Truncate(this->fName, 50, ETruncMode::kSpace);
        gPrint->Debug(1, __METHOD_NAME__, "Opening new frame chain: `%s`", fNameStr.Data());    

        FrFile *fMaster = FrFileINew(this->fName);
        if(fMaster == NULL) return *this;

        TString opt = TString(option);
                opt.ToLower();

        std::vector<std::vector<GPS>> _frameGPS;
        std::vector<std::vector<unsigned int>> _frameDQ;

        bool keepDuplicates = opt.Contains("d");
        std::vector<TString> _fileNames = ROOT::IOPlus::Helpers::Explode(" ", this->fName);
 
        int fileID = -1;       
        do {

            fileID = fileID + 1;
            
            FrFileH *file = fMaster->current;
            if(file == NULL) continue;
            
            if(!gPrint->IsSuperQuiet()) {

                gPrint->SetSingleCarriageReturn();
                gPrint->ProgressBar(Form("Reading TOC file #%d", fileID+1), "", fileID+1, _fileNames.size());
            }

            if(verbosity > Verbosity::kUnchanged) this->SetDebug(0);
            FrTOC *toc = FrTOCReadFull(fMaster); // Just to know the number of frames contained in current file
            if(verbosity > Verbosity::kUnchanged) this->SetDebug(GetVerbosity());

            if(toc == NULL) {

                gPrint->Warning(__METHOD_NAME__, Form("Failed to read TOC file #%d.. skip", fileID+1));
                continue; // Failed to load file.. @TODO implement an alternative loading procedure (if TOC not found)
            }

            bool bFileFound = this->FindByFileName(file->fileName).size() > 0;
            bool bGpsFound  = this->FindByGPS(file->tStart).size() > 0;
            if(!keepDuplicates && (bFileFound || bGpsFound)) {

                gPrint->Warning(__METHOD_NAME__, Form("Duplicates found for file #%d.. skip `%s`", fileID+1, file->fileName));
                continue;
            }

            int nFrames = toc->nFrame;
            if(nFrames < 1) continue;

            this->frames.push_back(std::vector<TFrame*>(nFrames, NULL));
            for(int frameID = 0; frameID < nFrames; frameID++) {
                
                GPS gpstime = GPS(
                    gClockGPS->Timestamp(toc->GTimeS[frameID], toc->GTimeN[frameID]),
                    gClockGPS->Timestamp(toc->GTimeS[frameID], toc->GTimeN[frameID]) + (*toc->dt)
                );

                this->frames[fileID][frameID] = new TFrame(this, fileID, frameID, tags, dataClasses, decimate, channelGroups, channelNumbers);
                this->frames[fileID][frameID]->SetGPS(gpstime);
                this->frames[fileID][frameID]->SetFileName(file->fileName);
                this->frames[fileID][frameID]->SetDataQuality(toc->dataQuality[frameID]);
                this->frames[fileID][frameID]->SetDetector(toc->detector ? toc->detector->name : "");
            }

        } while( FrFileINext(fMaster, 0., 0., NULL, FR_NO) == 0 );
        
        FrFileIEnd(fMaster);
        return Reallocate();
    }

    TFrameChain::TFrameChain(std::vector<TFrame *> frames, Verbosity verbosity)
    {
        if(verbosity > Verbosity::kUnchanged) this->SetVerbosity(verbosity);
        if(verbosity > Verbosity::kUnchanged) this->SetDebug(this->GetVerbosity());

        if(this->GetVerbosity() >= Verbosity::kDebug) {
            TFrameAbstract::Logging(gPrint->CreateTemporaryPath()+".log");
        }

        this->frames.clear();
        std::vector<std::vector<TFrame*>> frameGroups;
        std::vector<TString> fileNames;
        
        for(auto &frame : frames) {
            auto it = std::find(fileNames.begin(), fileNames.end(), frame->GetFileName());
            if(it == fileNames.end()) {
                fileNames.push_back(frame->GetFileName());
                frameGroups.push_back({frame});
            } else {
                frameGroups[it - fileNames.begin()].push_back(frame);
            }
        }
        
        for(auto &group : frameGroups) {
            this->frames.push_back(group);
        }

        this->fName = "";
        for(auto &f : this->frames) {
            this->fName += f[0]->GetFileName() + " ";
        }
        this->fName = this->fName.Strip();
        
        Reallocate();
    }

    TFrameChain& TFrameChain::Reallocate()
    {
        if(this->fManager != NULL) {
        
            delete this->fManager;
            this->fManager = NULL;
        }
        
        // Reorder vectors
        std::vector<int> indices(this->frames.size());
        std::iota(indices.begin(), indices.end(), 0);
        std::sort(indices.begin(), indices.end(), [&](int A, int B) -> bool { return this->frames[A][0]->GetGPS() < this->frames[B][0]->GetGPS(); });
        TFileReader::Reorder(this->frames,      indices);

        this->fName = "";
        for(auto &f : this->frames) {
            this->fName += f[0]->GetFileName() + " ";
        }

        std::vector<int> _bufferAllocation;
        for(auto &file : this->frames) {
            _bufferAllocation.push_back(file.size());
        }
        
        this->fManager = new TFrameManager(this->fName, _bufferAllocation);
        return *this;
    }

    std::vector<TFrame*> TFrameChain::FindByGPS(GPS gpstime) {

        std::vector<TFrame*> frames;
        for(auto frame : *this) {

            if (frame && gpstime.Overlaps(frame->GetGPS()))
            frames.push_back(frame);

            if (frame->GetGPS() > gpstime) break; // Stop searching because frames are sorted by GPS
        }

        return frames;
    }

    std::vector<TFrame*> TFrameChain::FindByFileName(TString filename) {

        std::vector<TFrame*> matchingFrames;

        for(auto frame : *this) {

            if (!frame->GetFileName().EqualTo(filename)) continue;
            matchingFrames.push_back(frame);
        }

        return matchingFrames;
    }

    void TFrameChain::Free()
    {
         for(int fileID = 0, nFiles = this->frames.size(); fileID < nFiles; fileID++) {

            for(int frameID = 0, nFrames = this->frames[fileID].size(); frameID < nFrames; frameID++) {
                this->frames[fileID][frameID]->Free();
            }
        }
    }

    bool TFrameChain::IsOpen()
    {
        return this->fManager != NULL && (&this->fManager->GetMaster()) != NULL;
    }

    void TFrameChain::Close()
    {
        for(int i = 0, I = this->frames.size(); i < I; i++) {

            for(int j = 0, J = this->frames[i].size(); j < J; j++) {

                if(this->frames[i][j]->IsOwnedBy(this))
                    delete this->frames[i][j];
            }
        }

        this->frames.clear();

        if(this->fManager != NULL) { 

            FrFile *fMaster = &this->fManager->GetMaster();
            if(fMaster != NULL) {    
                FrFileIEnd(fMaster);
            }

            delete this->fManager;
            this->fManager = NULL;
        }
    }

    TFrame *TFrameChain::AddFrame(int nFrames, int fileID)
    {
        if (fileID < 0 || fileID >= this->frames.size()) {
            this->frames.push_back({});
        }

        for(int i = 0; i < nFrames; i++) {
            this->frames[fileID].push_back(new TFrame(this, fileID, this->frames[fileID].size()));
        }

        Reallocate();
        
        return this->frames[fileID].back();
    }

    TFrame *TFrameChain::AddFrame(TFrame *frame)
    {
        std::vector<TFrame*> framesWithSameFileName = this->FindByFileName(frame->GetFileName());

        int fileID = !framesWithSameFileName.empty() ? framesWithSameFileName.front()->GetFileID() : -1;
        if(fileID < 0) {
            fileID = this->frames.size();
            this->frames.push_back({});
        }

        this->frames[fileID].push_back(frame);
        Reallocate();
        
        return this->frames[fileID].back();
    }
    
    TFrame *TFrameChain::RemoveFrame(int fileID, int frameID)
    {
        if (fileID < 0 || fileID >= this->frames.size()) return nullptr;
        if (frameID < 0 || frameID >= this->frames[fileID].size()) return nullptr;

        TFrame* frame = this->frames[fileID][frameID];
        this->frames[fileID].erase(this->frames[fileID].begin() + frameID);

        return frame;
    }

    TFrame *TFrameChain::RemoveFrame(TFrame* frame)
    {
        if (!frame) return nullptr;

        for (auto &file : frames) {
            auto it = std::find(file.begin(), file.end(), frame);
            if (it != file.end()) {
                TFrame* removedFrame = *it;
                file.erase(it);
                return removedFrame;
            }
        }

        return nullptr;
    }

    std::vector<TFrame *> TFrameChain::RemoveFrame(std::vector<TFrame *> frames)
    {
        std::unordered_map<int, std::vector<TFrame *>> framesToRemove;

        for (auto &frame : frames) {
            int fileID = frame->GetFileID();
            if (fileID >= 0 && fileID < this->frames.size()) {
                framesToRemove[fileID].push_back(frame);
            }
        }

        for (auto &[fileID, frameList] : framesToRemove) {
            auto &fileFrames = this->frames[fileID];
            for (auto &frame : frameList) {
                auto it = std::find(fileFrames.begin(), fileFrames.end(), frame);
                if (it != fileFrames.end()) {
                    fileFrames.erase(it);
                }
            }
        }

        Reallocate();

        std::vector<TFrame *> removedFrames;
        for (auto &frame : frames) {
            
            if (!this->FindByFileName(frame->GetFileName()).size()) {
                removedFrames.push_back(frame);
            }
        }

        return removedFrames;
    }

    bool TFrameChain::AddFriend(TFrameChain *c)
    {
        if(c == NULL) return false;
        if (this->GetNFrames() != c->GetNFrames()) {

            std::cerr << "Failed to chain `" << this->fName << "` and `" << c->fName << "`: same number of frame expected" << std::endl;
            return false;
        }

        std::vector<TFrame *> siblings;
        for(int i = 0, N = this->GetNFrames(); i < N; i++) {
            
            TFrame *f1 = this->GetFrame(this->GetFileID(i), this->GetFrameID(i));
            GPS GPSrange = f1->GetGPS();

            TFrame *f2 = c->GetFrame(f1->GetGPS());
            if (f2 == NULL) {

                std::cerr << "Failed to chain `" << this->fName << "` and `" << c->fName << "`: GPS [" << GPSrange.start << "," << GPSrange.end << "] not found" << std::endl;
                return false;
            }
            
            siblings.push_back(f2);
        }
        
        for(int i = 0, N = this->GetNFrames(); i < N; i++) {
            TFrame *f1 = this->GetFrame(this->GetFileID(i), this->GetFrameID(i));
            f1->AddFriend(siblings[i]);
        }

        return true;
    }

    void TFrameChain::RemoveFriend(TFrameChain *c)
    {
        if(c == NULL) return;

        for(int i = 0, N = this->GetNFrames(); i < N; i++) {
            TFrame *f1 = this->GetFrame(this->GetFileID(i), this->GetFrameID(i));
            TFrame *f2 = c->GetFrame(f1->GetGPS());
            if (f2 == NULL) {
                std::cerr << "Failed to unchain `" << this->fName << "` and `" << c->fName << "`: GPS [" << f1->GetGPS().start << "," << f1->GetGPS().end << "] not found" << std::endl;
                continue;
            }
            f1->RemoveFriend(f2);
        }
    }

    void TFrameChain::Print(Option_t *opt, int globalID)
    {
        std::vector<int> globalIDs;
        for(int i = 0, N = this->GetNFrames(); i < N; i++) {
            globalIDs.push_back(i);
        }

        return Print(opt, globalID < 0 ? globalIDs : std::vector<int>({globalID}));
    }

    void TFrameChain::Print(Option_t *opt, std::vector<int> globalIDs)
    {
        TString option = opt;
                option.ToLower();

        bool bRead = option.Contains("read");
        bool bDump = option.Contains("dump");
        bool bWait = option.Contains("wait");
        bool bTOC  = option.Contains("toc");

        bool bInteractive = option.Contains("interactive");

        if(bTOC) {

            this->DumpTOC(); 
            
            gPrint->PressAnyKeyToContinue(10, __METHOD_NAME__, "Start reading frames..");
            if (gPrint->IsDebugEnabled(10)) gPrint->ClearPage();       
        }

        TVirtualPad *gPadBak = gPad;
        int padWidth  = this->padWidth;
        int padHeight = this->padHeight;

        int globalID_first = globalIDs.size() ? *std::min_element(globalIDs.begin(), globalIDs.end()) : 0;
        for(int globalID = globalID_first, globalN = this->GetNFrames(); globalID < globalN; globalID++)
        {
            bool isValid = std::find(globalIDs.begin(), globalIDs.end(), globalID) != globalIDs.end();
            bool bDraw = option.Contains("draw");
            bool bDrawLimit = false;

            int fileID  = this->GetFileID(globalID);
            int frameID = this->GetFrameID(globalID);
            
            TString fileName = this->GetFileName(fileID);
            int nFiles = this->frames.size();
            int nFrames = this->frames[fileID].size();

            int count = 0;
            GPS GPSrange(0,0);
            if(bInteractive) {
    
                TFrame *frame = this->frames[fileID][frameID];
                        frame->Print("interactive");

                if (GPSrange.start == GPSrange.end)
                    GPSrange = frame->GetGPS();

                if(isValid) std::cout << ">>> selectable frame: " << gPrint->kGreen << "✓" << gPrint->kNoColor;
                else std::cout << gPrint->kOrange << ">>> skipped in selection: " << gPrint->kRed << "✗" << gPrint->kNoColor;
                if(isValid) std::cout << std::endl << std::endl;
                else std::cout << std::endl;
                
                enum class MyAction { PrevFrame, NextFrame, PrevFile, NextFile, PrevBatch, NextBatch, ReadTOC, ReadFrame, Clear, Quit };
                std::map<TKeyboard::KeyEvent, MyAction> keyEvents;
                keyEvents[TKeyboard::KeyEvent({27, 91, 68})] = MyAction::PrevFrame; // LEFT
                keyEvents[TKeyboard::KeyEvent({27, 91, 67})] = MyAction::NextFrame; // RIGHT
                keyEvents[TKeyboard::KeyEvent({27, 91, 49, 59, 50, 68})] = MyAction::PrevBatch; // Shift+LEFT
                keyEvents[TKeyboard::KeyEvent({27, 91, 49, 59, 50, 67})] = MyAction::NextBatch; // Shift+RIGHT
                keyEvents[TKeyboard::KeyEvent({27, 91, 65})] = MyAction::PrevFile ; // UP
                keyEvents[TKeyboard::KeyEvent({27, 91, 66})] = MyAction::NextFile ; // DOWN
                keyEvents[TKeyboard::KeyEvent({9})         ] = MyAction::ReadTOC  ; // TAB
                keyEvents[TKeyboard::KeyEvent({10})        ] = MyAction::ReadFrame; // ENTER
                keyEvents[TKeyboard::KeyEvent({127})       ] = MyAction::Clear    ; // DEL
                keyEvents[TKeyboard::KeyEvent({27})        ] = MyAction::Quit     ; // ESC

                std::vector<std::vector<unsigned char>> sequences;
                for( auto event : keyEvents)
                    sequences.push_back(event.first.GetSequence());
                
                if(gPadBak) gPadBak->cd(1);

                bool first = true;
                while(1) {

                    if(!first) std::cout << std::endl;
                    
                    TString fileStr  = nFiles  > 1 ? "[UP|DOWN] to change file\n" : "";
                    TString frameStr = nFrames > 1 ? "[LEFT|RIGHT] to change frame (use shift to skip faster)\n" : "";
                    TString readableStr = isValid ? "[TAB] to read TOC, [DEL] to clear the page, [ESC] to exit\n" : "Press [ESC] to exit\n";
                    TString blinkingStr = isValid ? "Press [ENTER] to " + TString(first ? "read" : "re-read") + " the frame.\n" : "";

                    // int keyPressed = first && frame->GetN() < 32 ? 1 /* MyAction::NextFrame */ : gPrint->WaitForKeyPress(__METHOD_NAME__, sequences, blinkingStr, fileStr + frameStr + readableStr); 
                    int keyPressed = gPrint->WaitForKeyPress(__METHOD_NAME__, sequences, blinkingStr, fileStr + frameStr + readableStr); 
                    if(!first) gPrint->ClearLine(4);

                    first = false;
                    if(gPadBak) gPadBak->cd();
                    bool bRead = false;
                    bool bTOC  = false;
                    bool bInteractiveReset = true;

                    int nObjects = frame->GetN();
                    auto it = keyEvents.find(TKeyboard::KeyEvent(sequences[keyPressed]));
                    switch(it->second) {

                        case MyAction::PrevFile:
                            globalID = TMath::Max(0, this->GetGlobalID(TMath::Max(0,fileID-1), 0));
                            if(globalID < 0) globalID = 0;
                            break;

                        case MyAction::NextFile:

                            globalID = this->GetGlobalID(TMath::Min(nFiles, fileID+1), 0);
                            if (globalID < 0) globalID = globalN;

                            break;
                        
                        case MyAction::PrevBatch:
                            globalID = TMath::Max(0, globalID - 32);
                            break;

                        case MyAction::NextBatch:
                            globalID = TMath::Min(globalN, globalID + 32);
                            break;

                        case MyAction::PrevFrame: [[fallthrough]];
                            globalID = TMath::Max(0, globalID-1);
                            break;

                        case MyAction::Quit:
                            globalID = globalN+1;
                            break;

                        case MyAction::ReadFrame:
                            bInteractiveReset = false;
                            bRead = true;
                            break;

                        case MyAction::ReadTOC:
                            bInteractiveReset = false;
                            bTOC = true;
                            break;

                        case MyAction::Clear:
                            break;

                        case MyAction::NextFrame: [[fallthrough]];
                        default:
                            globalID = TMath::Min(globalN, globalID+1);
                            break;
                    }

                    if(bInteractiveReset) {

                        if(globalID < 0) globalID = 0;
                        else if(globalID == globalN) globalID = globalID-1; // Prevent to exit when reaching the last frame.
                        
                        globalID--; // Compensate for loop index increase.
                        break;
                    }

                    if(!isValid) continue;

                    if(bTOC) this->DumpTOC(this->GetGlobalID(fileID,0));
                    if(bRead) {

                        TFrame *frame = this->GetFrame(fileID,frameID);

                        bool isEmpty = true;
                        std::vector<TFrameProxyT> proxies = frame->ReadProxy();

                        for(int k = 0, K = proxies.size(); k < K; k++) {

                            std::visit([&count, &isEmpty, &bDraw, &bDump, &bDrawLimit, &GPSrange, &nObjects, &gPadBak, &padWidth, &padHeight](auto* obj) -> void { 

                                if(obj->GetType().Contains("FrameFormat::TFrRawData")) return;
                                isEmpty &= obj->GetNBaskets() == 0;

                                if(obj->GetNBaskets()) std::cout << TPrint::kPurple << "-----" << TPrint::kNoColor << std::endl;
                                for(int basket = 0, B = obj->GetNBaskets(); basket < B; basket++) {

                                    if(bDraw && !bDrawLimit) {

                                        int iPad = ROOT::IOPlus::Helpers::GetPadIndex(gPad)+1;
                                        if (iPad > 0) {
                                            
                                            int nPads = ROOT::IOPlus::Helpers::CountPads(gPad->GetMother());
                                            if (iPad >= nPads) {

                                                gPrint->Warning(__METHOD_NAME__, "Too many objects to draw. Only " + TString::Itoa(nPads,10) + " available pad(s).");
                                                bDrawLimit = true;
                                                continue;
                                            }

                                            gPad->cd(iPad+1);
                                        }

                                        TTree *tree = (TTree *) obj->GetObject(basket);
                                        if(tree != NULL) {

                                            TH1 *hData = TFrameProxyAbstract::Histogram(tree, obj->GetBranchName(), "delete");
                                            if (hData) {

                                                if(hData->GetEntries() > 16384) {
                                                    gPrint->Warning(__METHOD_NAME__, "Histogram is too large, reduced performances expected for '" + TString(hData->GetName()) + "'.");
                                                }

                                                if(gPad) gPad->GetMother()->cd(iPad+1);
                                                else {

                                                    int maxPads = int(gPicasso->GetDisplayHeight()/padHeight);
                                                    int nPads = TMath::Min(maxPads, nObjects);
                                                    TCanvas *c = TCanvas::MakeDefCanvas();
                                                            c->SetWindowSize(padWidth, padHeight*nPads);
                                                    
                                                    gKeyboard->Connect(c);
                                                    gPadBak = gPad;

                                                    gPad->Divide(1, nPads);
                                                    gPad->cd(1);
                                                } 

                                                hData->Draw();
                                                if (gPad) {
                                                    
                                                    gPad->Modified();
                                                    gPad->Update();
                                                }
                                            }
                                        }
                                    }
                                    
                                    obj->Print(basket);
                                    if(gPrint->IsDebugEnabled(1) || bDump) {
                                        obj->Dump(basket);
                                    }

                                    if(bDraw) gSystem->ProcessEvents();

                                    count++;
                                }

                            }, proxies[k]);
                        }

                        frame->PrintBuffer();

                        if(isEmpty && frame->GetDataBuffer().size() < 1) {
                            gPrint->Warning(__METHOD_NAME__, "No object found in this frame.");
                        }

                        if (gPad) {

                            gPad->Modified();
                            gPad->Update();
                        }
                    }
                }

                if(gPadBak) gPadBak->cd();
                
            } else {

                if(std::find(globalIDs.begin(), globalIDs.end(), globalID) == globalIDs.end())
                    continue;

                TString fileName = this->GetFileName(fileID);
                int nFrame = this->frames[fileID].size();

                TString header = gPrint->kGreen + "Master #" + TString::Itoa(fileID+1,10) + "/" + TString::Itoa(nFiles,10) + gPrint->kNoColor + "; " + gPrint->MakeItShorter(fileName, 64) + "; " + gPrint->kRed + TString::Itoa(nFrames,10) + " frame" + (nFrame > 1 ? "s" : "") + gPrint->kNoColor;
                        header += gPrint->kNoColor + " <<< from " + gPrint->kOrange + gClockGPS->UTC(this->GetGPS(fileID).start) + gPrint->kNoColor + " to " + gPrint->kOrange + gClockGPS->UTC(this->GetGPS(fileID).end) + gPrint->kNoColor;

                if(bDraw) {
                    gPrint->Warning(__METHOD_NAME__, "Drawing is only permitted in interactive mode.");
                }

                if(globalID == globalID_first || frameID == 0) {

                    if(bTOC) {

                        if (gPrint->IsDebugEnabled(bTOC ? 10 : 100)) {

                            std::cout << header << std::endl;
                            gPrint->PressAnyKeyToContinue(__METHOD_NAME__, "Table of content will be read.");
                            gPrint->ClearPage();
                        }

                        std::cout << header << std::endl;
                        this->DumpTOC(globalID);
                    }
                }

                TFrame *frame = this->frames[fileID][frameID];
                if(!bRead && !bTOC && frame) frame->Print();
                else if(bRead) {

                    std::vector<TFrameProxyT> proxies = frame->ReadProxy();
                    for(int k = 0, K = proxies.size(); k < K; k++) {

                        std::visit([&count, &bDump](auto* obj) -> void { 

                            if(obj->GetNBaskets()) std::cout << TPrint::kPurple << "-----" << TPrint::kNoColor << std::endl;
                            
                            for(int basket = 0, B = obj->GetNBaskets(); basket < B; basket++) {
                                
                                obj->Print(basket);
                                if(gPrint->IsDebugEnabled(1) || bDump) {
                                    obj->Dump(basket);
                                }
                                count++;
                            }

                        }, proxies[k]);

                    }

                    std::vector<TFrAbstractT> fDataBuffer = frame->GetDataBuffer();
                    for (auto& obj : fDataBuffer)
                    {
                        std::visit([](auto* concreteObj) {
                            using T = std::decay_t<decltype(*concreteObj)>;
                            if constexpr (std::is_base_of<TFrAbstract, T>::value) {
                                TFrameProxy<T>::Print(concreteObj);
                            }
                        }, obj);

                        count++;
                    }
                    
                    if(count < 1) {
                        std::cout << TPrint::kPurple << "-----" << TPrint::kNoColor << std::endl;
                        std::cout << "No object found in this frame." << std::endl;
                    }

                    if (globalID+1 < globalN && count > 0) {

                        if(gPrint->IsDebugEnabled(bWait ? 1 : 10)) {
                            std::cout << std::endl;
                        }

                        gPrint->PressAnyKeyToContinue(bWait ? 1 : 10, __METHOD_NAME__, "Go to next the frame..");
                        if(gPrint->IsDebugEnabled(bWait ? 1 : 10)) {
                            gPrint->ClearLine(bWait ? 1 : 2);
                        }
                    }
                }
            }
        }
    }

    int TFrameChain::GetEntries() const 
    {
        return this->GetNFrames();
    }

    int TFrameChain::GetNFrames(int fileID) const 
    {
        if(fileID < 0) {

            int total = 0;
            for(int i = 0, N = this->GetNFiles(); i < N; i++)
                total += this->GetNFrames(i);

            return total;
        }

        return fileID < this->frames.size() ? this->frames[fileID].size() : 0;
    }

    int TFrameChain::GetGlobalID(TFrame *frame) const 
    {
        return this->GetGlobalID(frame->GetFileID(), frame->GetFrameID());
    }

    int TFrameChain::GetNFiles() const 
    {
        return this->frames.size();
    }

    TString TFrameChain::GetFileName(int fileID) const 
    {
        if(fileID < 0 || fileID >= this->frames.size())
            throw std::invalid_argument("Unexpected file ID ("+TString::Itoa(fileID,10)+") provided.");
    
        return this->frames[fileID][0]->GetFileName();
    }

    std::vector<int> TFrameChain::GetFileID(std::vector<int> globalIDs) const 
    {
        std::vector<int> fileIDs;
        for(int i = 0, N = globalIDs.size(); i < N; i++)
        {
            int fileID = this->GetFileID(globalIDs[i]);
            if(fileID < 0) continue;
            
            fileIDs.push_back(fileID);
        }

        return fileIDs;
    }

    std::vector<int> TFrameChain::GetFrameID(std::vector<int> globalIDs) const 
    {
        std::vector<int> frameIDs;
        for(int i = 0, N = globalIDs.size(); i < N; i++)
        {
            int frameID = this->GetFrameID(globalIDs[i]);
            if(frameID < 0) continue;
            
            frameIDs.push_back(frameID);
        }

        return frameIDs;
    }

    void TFrameChain::DumpTOC()
    {
        std::cout   << "==> Table of content at:" << static_cast<void*>(this)
                        << ", name=\"" << gPrint->MakeItShorter(TString(this->fName)) << "\""
                        << ", class=\"" << this->Class() << "\""<< std::endl;

        if(this->IsOpen()) std::cout   << "    This chain contains " << this->GetNFrames() << " frame(s) out of " << this->GetNFiles() << " file(s)" << std::endl;
        else {
                std::cout << "    File closed." << std::endl;
                return;
        }

        std::cout << std::endl;
    }

    void TFrameChain::DumpTOC(int globalID)
    {
        this->GetManager().DumpTOC(&this->GetManager().GetMaster(), this->GetFileID(globalID));
        std::cout << std::endl << std::flush;
    }

    void TFrameChain::Dump(int globalID)
    {
        std::vector<int> globalIDs;
        for(int i = 0, N = this->GetNFrames(); i < N; i++) {
            globalIDs.push_back(i);
        }

        return Dump(globalID < 0 ? globalIDs : std::vector<int>({globalID}));
    }

    bool TFrameChain::IsContiguous()
    {
        int nFiles = this->GetNFiles();
        if(nFiles == 0) return false;

        TFrame *frame = this->GetFrame(0,0);
        if(frame == NULL) return false;

        do {

            if(!frame->IsContiguous()) return false;

        } while( (frame = frame->Next() ));

        return true;
    }

    void TFrameChain::Dump(std::vector<int> globalIDs)
    {
        std::cout   << "==> Dumping object at:" << static_cast<void*>(this)
                    << ", name=\"" << TString(this->fName) << "\""
                    << ", class=\"" << this->Class() << "\""<< std::endl;

        TString contiguous = !this->IsContiguous() ? gPrint->kRed + " [DISCONTINUED]" + gPrint->kNoColor : "";
        std::cout   << "    This object contains in total " << this->GetNFrames() << " frame(s), out of " << this->GetNFiles() << " file(s)" << contiguous << std::endl;

        for(int fileID = 0, nFiles = this->GetNFiles(); fileID < nFiles; fileID++)
        {
            //
            // TABLE OF CONTENT
            {
                TString frameID = "==========  " + gPrint->kNoColor + "TABLE OF CONTENT #"+ TString::Itoa(fileID+1,10) + gPrint->kGreen+"  ==========";
                TString spacer = ROOT::IOPlus::Helpers::Spacer(frameID.Length()-11, '=');

                std::cout << gPrint->kGreen << std::endl;
                std::cout << spacer << std::endl;
                std::cout << frameID << std::endl;
                std::cout << spacer << gPrint->kNoColor << std::endl;

                std::cout << std::endl;
                std::cout << "==> Table of content at:" << static_cast<void*>(this)
                          << ", name=\"" << TString(this->fName) << "\""
                          << ", class=\"" << this->Class() << "\""<< std::endl;

                this->GetManager().ReadTOC(&this->GetManager().GetMaster(), fileID);
            }

            //
            // DUMPING ALL FRAMES
            for(int frameID = 0, nFrames = this->GetNFrames(fileID); frameID < nFrames; frameID++) {

                int globalID = this->GetGlobalID(fileID, frameID);
                if(std::find(globalIDs.begin(), globalIDs.end(), globalID) == globalIDs.end())
                    continue;

                std::cout << std::endl << "==> Reading frame #" << (frameID+1) << " at:" << static_cast<void*>(this) << std::endl;
                TFrame *frame = this->GetFrame(fileID, frameID);
                if(frame) frame->Dump();
            }

            //
            // END OF FILE
            //
            {
                TString frameID = "==========  " + gPrint->kNoColor + "END OF FILE #"+ TString::Itoa(fileID+1,10) + gPrint->kOrange+"  ==========";
                TString spacer = ROOT::IOPlus::Helpers::Spacer(frameID.Length()-11, '=');
                if(this->GetNFrames(fileID) < 1) continue;

                std::cout << gPrint->kOrange << std::endl;
                std::cout << spacer << std::endl;
                std::cout << frameID << std::endl;
                std::cout << spacer << gPrint->kNoColor << std::endl;

                FrEndOfFile *eof = this->GetManager().ReadEOF(&this->GetManager().GetMaster(), fileID);
                delete eof; // Delete it right away.. we just want to set the cursor at the right place and display debug information
            }
        }
    }

    int TFrameChain::GetFileID(int globalID) const 
    {
        if(globalID < 0 && this->fManager != NULL) 
            return this->fManager->GetFileID(&this->fManager->GetMaster());

        for(int i = 0, N = this->GetNFiles(); i < N; i++) {

            int maxLocalID = this->frames[i].size();
            if(globalID < maxLocalID) return i;

            globalID -= maxLocalID;
        }

        return -1;
    }

    int TFrameChain::GetFrameID(int globalID) const 
    {
        if(globalID < 0 && this->fManager != NULL)
            return this->fManager->GetFrameID(&this->fManager->GetMaster());

        for(int i = 0, N = this->frames.size(); i < N; i++) {

            int maxLocalID = this->frames[i].size();
            if(globalID < maxLocalID) return globalID;

            globalID -= maxLocalID;
        }

        return -1;
    }

    int TFrameChain::GetGlobalID(int fileID, int localID) const 
    {
        int globalID = 0;
        int maxGlobalID = 0;

        for(int i = 0, N = this->frames.size(); i < N; i++) {

            maxGlobalID += this->frames[i].size();
            if(i < fileID)
                globalID += this->frames[i].size();
        }

        globalID += localID;

        return globalID < maxGlobalID ? globalID : -1;
    }

    TFrame* TFrameChain::GetFrame(GPS GPSrange) const 
    {
        for(int i = 0, N = this->GetNFrames(); i < N; i++) {
            
            TFrame *f = this->GetFrame(this->GetFileID(i), this->GetFrameID(i));
            GPS _GPSrange = f->GetGPS();

            if(GPSrange.start != _GPSrange.start) continue;
            if(GPSrange.end != _GPSrange.end) continue;

            return f;
        }

        return NULL;
    }

    TFrame* TFrameChain::GetFrame(int fileID, int frameID) const 
    {
        return this->GetFrame(this->GetGlobalID(fileID, frameID));
    }

    TFrame* TFrameChain::GetFrame(int globalID) const
    {
        int fileID = this->GetFileID(globalID);
        if (fileID < 0) return NULL;

        int frameID = this->GetFrameID(globalID);
        if (frameID < 0) return NULL;
        
        return this->frames[fileID][frameID];
    }

    std::vector<int> TFrameChain::Select(TString expr, std::vector<double> parameters)
    {
        std::vector<int> entryList;

        expr = expr.ReplaceAll("*", "true");
        TFormula *formula = expr.EqualTo("") ? NULL : new TFormula(ROOT::IOPlus::Helpers::GetRandomName("formula"), expr);
        if (formula == NULL) return entryList;

        if (!formula->IsValid()) {
            std::cerr << "Failed to compute formula `" << expr.Data() << "`" << std::endl;
            throw std::exception();
        }

        if (parameters.size()) {
            formula->SetParameters(&parameters[0]);
        }

        double duration = 0;
        for(int fileID = 0, N = this->GetNFiles(); fileID < N; fileID++) {

            for(int frameID = 0, M = this->GetNFrames(fileID); frameID < M; frameID++) {

                int globalID = this->GetGlobalID(fileID, frameID);

                int x = globalID;
                double y = this->GetGPS(fileID, frameID).start;

                duration += this->GetTimeLength(fileID, frameID);
                double z = duration;

                double t = 0;
                if((expr.Contains("x[3]") || expr.Contains("t")) && !expr.EqualTo("true")) {

                    TFrame *frame = this->GetFrame(fileID, frameID);
                    if(gPrint->WarningIf(frame == NULL, __METHOD_NAME__, (TString) "Failed to find frame #"+TString::Itoa(globalID+1,10)))
                        throw std::exception();

                    t = frame->GetDataQuality();
                }

                if(formula->Eval(x, y, z, t)) {
                    if(!expr.EqualTo("true")) gPrint->Debug(100, __METHOD_NAME__, (TString) "Frame #"+TString::Itoa(globalID+1,10)+" got a valid selection criteria.. " + expr);
                } else {
                    gPrint->Debug(100, __METHOD_NAME__, (TString) "Frame #"+TString::Itoa(globalID+1,10)+" doesn't meet selection criteria.. " + expr);
                    continue;
                }

                entryList.push_back(globalID);
            }
        }

        return entryList;
    }

    std::vector<int> TFrameChain::SelectByFrame(std::vector<int> frames)
    {
        int nFrames = this->GetNFrames();

        TString formula = "";
        for(int x = 0, N = frames.size(); x < N; x++) {

            if(frames[x] < 1 || frames[x] > nFrames)
                throw std::invalid_argument("Frame #"+TString::Itoa(frames[x],10)+" requested, but TFrameChain must range between 1 and "+TString::Itoa(nFrames, 10)+".");

            formula += (x > 0 ? " && " : "") + ("x == " + TString::Itoa(frames[x]-1,10));
        }

        return this->Select(formula);
    }

    std::vector<int> TFrameChain::SelectByFrame(int first, int last)
    {
        int nFrames = this->GetNFrames();
        if(first < 1 || first > nFrames)
            throw std::invalid_argument("Frame #"+TString::Itoa(first,10)+" requested, but TFrameChain must range between 1 and "+TString::Itoa(nFrames, 10)+".");
        if(last < 1 || last > nFrames)
            throw std::invalid_argument("Frame #"+TString::Itoa(last,10)+" requested, but TFrameChain must range between 1 and "+TString::Itoa(nFrames, 10)+".");

        TString formula = "";
        if(first >= 0) formula += TString("x >= "+TString::Itoa(first-1,10));
        if(last  >= 0) formula += (formula.EqualTo("") ? "" : " && ") + ("x <= "+TString::Itoa(last-1,10));

        return this->Select(formula);
    }

    std::vector<int> TFrameChain::SelectByGPS(std::vector<double> gpsList)
    {
        TString formula = "";
        for(int i = 0, N = gpsList.size(); i < N; i++)
            formula += (i > 0 ? " && " : "") + ("y == ["+TString::Itoa(i,10)+"]");

        return this->Select(formula, gpsList);
    }

    std::vector<int> TFrameChain::SelectByGPS(GPS gpstime)
    {
        TString formula = "";
        if(gpstime.start >= 0) formula += TString(gpstime.start >= 0 ? "y >= [0]" : "");
        if(gpstime.end  >= 0) formula += (formula.EqualTo("") ? "" : " && ") + TString(gpstime.end  >= 0 ? "y <= [1]" : "");

        return this->Select(formula, std::vector<double>({gpstime.start, gpstime.end}));
    }

    std::vector<int> TFrameChain::SelectByMaxDuration(double duration)
    {
        TString formula = (duration ? "z <= [0]" : "");
        return this->Select(formula, std::vector<double>({duration}));
    }

    std::vector<int> TFrameChain::SelectByMinDuration(double duration)
    {
        TString formula = (duration ? "z >= [0]" : "");
        return this->Select(formula, std::vector<double>({duration}));
    }

    std::vector<int> TFrameChain::SelectByDQ(unsigned int dataQuality)
    {
        TString formula = ("t == " + TString::UItoa(dataQuality, 10));
        return this->Select(formula);
    }

    GPS TFrameChain::GetGPS(int fileID, int frameID) const 
    {
        GPS gpstime = GPS();

        if (fileID < 0) {

            for (int i = 0; i < this->GetNFiles(); ++i) {
                gpstime = gpstime.ProtractsWith(this->GetGPS(i, frameID));
            }

        } else if (frameID < 0) {
           
            for (int i = 0; i < this->GetNFrames(fileID); ++i) {
                gpstime = gpstime.ProtractsWith(this->GetGPS(fileID, i));
            }

        } else {

            TFrame* frame = this->GetFrame(fileID, frameID);
            if (frame) gpstime = frame->GetGPS();
        }

        return gpstime;
    }

    double TFrameChain::GetTimeLength(int fileID, int frameID) const 
    {
        GPS gpstime = this->GetGPS(fileID, frameID);
        return gpstime.end-gpstime.start;
    }

    bool TFrameChain::VerifyChecksum(bool decompress)
    {
        FrScanFile(&this->fManager->GetMaster(), decompress); // Scan files and compute chkSum
        return this->fManager->GetMaster().chkSumFi == this->fManager->GetMaster().chkSumFiRead;
    }

    std::vector<TString> TFrameChain::Write(TString fName, Option_t *opt, int framesPerFile)
    {
        std::vector<int> globalIDs;
        for (int i = 0; i < this->GetNFrames(); ++i) {
            globalIDs.push_back(i);
        }

        return Write(fName, opt, framesPerFile, globalIDs);
    }

    std::vector<TString> TFrameChain::Write(TString fName, Option_t *opt, int framesPerFile, std::vector<int> globalIDs)
    {
        std::vector<TString> writtenFiles;
        if(gPrint->ErrorIf(fName.EqualTo(""), __METHOD_NAME__, (TString) "No output file provided. Cannot save selected frames."))
            return writtenFiles;

        if(globalIDs.size() < 1) {
            gPrint->Warning(__METHOD_NAME__, (TString) "No frame selected for writing.");
            return writtenFiles;
        }

        if(framesPerFile < 0) framesPerFile = globalIDs.size();
        for(int b = 0, B = globalIDs.size()/framesPerFile; b < B; b++) {

            std::vector<int> _globalIDs = std::vector<int>(globalIDs.begin() + framesPerFile*b, globalIDs.begin() + framesPerFile*(b+1));
            if (fName.EndsWith(".gwf")) {
                std::vector<TString> _writtenFiles = WriteGWF(fName, opt, _globalIDs);
                for( auto _fName : _writtenFiles )
                    writtenFiles.push_back(_fName);
                    
                continue;
            }

            if (fName.EndsWith(".root")) {
                std::vector<TString> _writtenFiles = WriteROOT(fName, opt, _globalIDs);
                for( auto _fName : _writtenFiles )
                    writtenFiles.push_back(_fName);

                continue;
            }

            gPrint->Error(__METHOD_NAME__, (TString) "Wrong output extension provided for %s. Please use \".gwf\" or \".root\" extension..", fName.Data());
        }

        return writtenFiles;
    }

    void TFrameChain::Set(TString name)
    {
        this->fName = name;
        this->Reload();
    }

    TString TFrameChain::Get() const
    {
        return this->fName;
    }

    void TFrameChain::Reload()
    {
        this->Open(this->fName, this->option, this->tags, this->dataClasses, this->decimate, this->channelGroups, this->channelNumbers, this->verbosity);
    }

    std::vector<TString> TFrameChain::WriteGWF(TString fName, Option_t *opt, std::vector<int> globalIDs)
    {
        std::vector<TString> writtenFiles;

        if(gPrint->ErrorIf(fName.EqualTo(""), __METHOD_NAME__, (TString) "No output file provided. Cannot save selected frames."))
            return writtenFiles;

        if(TFileReader::HasPattern(fName)) {
            gPrint->Error(__METHOD_NAME__, "Output contains a pattern, which is not supported for GWF files. These files already include GPS time and the number of frames.");
            return writtenFiles;
        }

        TString fDir = TFileReader::DirName(fName);
        TString fBasename = TFileReader::BaseName(fName, ".gwf");
        gSystem->mkdir(fDir, true);

        TString tmpDir = gPrint->CreateTemporaryPath();
        TString tmpOutput = tmpDir+"/"+fBasename+".gwf";
        gSystem->mkdir(tmpDir,true);

        TString option = opt;
                option.ToLower();

        if(option.Contains("log") && TFrameAbstract::logFile != NULL) {
            gSystem->Rename(TFrameAbstract::logFileName, fDir+"/"+fBasename+".log");
        }

        bool bTree  = option.Contains("tree");
        bool bMore = option.Contains("more");
        bool bVerbose = option.Contains("verbose");

        int compression = TMath::Min(255, TMath::Max(0, this->compression));
        FrFile *oFile = FrFileONew(tmpOutput, compression);
        
        if(gPrint->ErrorIf(oFile == NULL, __METHOD_NAME__, (TString) "Cannot open output file \""+fName+"\""))
            return writtenFiles;

        int lastFileID = -1;
        int timeReference = 0, length = 0;
        int nFrames = 0;
        
        // Remove duplicates
        std::sort(globalIDs.begin(), globalIDs.end());
        auto last = std::unique(globalIDs.begin(), globalIDs.end());
        globalIDs.erase(last, globalIDs.end());

        for(int i = 0, N = globalIDs.size() ? globalIDs.size() : this->GetNFrames(); i < N; i++)
        {
            int globalID = globalIDs.size() ? globalIDs[i] : i;
            int fileID = this->GetFileID(globalID);
            int frameID = this->GetFrameID(globalID);

            if(gPrint->WarningIf(fileID < 0 || frameID < 0, __METHOD_NAME__, (TString) "Failed to find global frame #"+TString::Itoa(globalID+1, 10)))
                continue;
            
            if(lastFileID < 0) lastFileID = fileID;

            TFrame* frame = this->GetFrame(globalID);
            if(gPrint->WarningIf(frame == NULL, __METHOD_NAME__, (TString) "Failed to find frame #"+gPrint->Pad(frameID+1, this->GetNFrames())+" from file #"+gPrint->Pad(fileID+1, this->GetNFiles())))
                continue;

            int ret = this->GetManager().Write(oFile, frame);
            if(nFrames < 1) timeReference = frame->GetGPS().start;
            
            TString _fName = this->GetFileName(fileID);
            if(ret != FR_OK) gPrint->Warning(__METHOD_NAME__, (TString) "Failed to write frame #"+gPrint->Pad(frameID+1, this->GetNFrames())+" from file #"+gPrint->Pad(fileID+1, this->GetNFiles())+"/"+TString::Itoa(this->GetNFiles(), 10)+" ("+_fName+").");
            else {

                if(bMore) gPrint->Message(__METHOD_NAME__, (TString) "Frame #"+gPrint->Pad(frameID+1, this->GetNFrames())+" from file #"+gPrint->Pad(i+1, N)+"/"+TString::Itoa(N, 10)+" ("+_fName+") saved.");
                else {
                
                    TString fNameWithBrackets = fName;
                            fNameWithBrackets.ReplaceAll(".gwf","-[..]-[..].gwf");

                    gPrint->ProgressBar((TString) "Saving frames from master #"+gPrint->Pad(fileID+1, this->GetNFiles())+"/"+TString::Itoa(N, 10)+" into \""+fNameWithBrackets+"\"", "", frameID+1, N);
                }

                nFrames++;
            }

            length += frame->GetTimeLength();
        }

        if(bMore) std::cout << std::endl;

        FrFileOEnd(oFile);

        if(bTree) { 
            fDir += "/" + TString::Itoa(timeReference,10)(0,6);
            gSystem->mkdir(fDir, true);
        }

        TString output = fDir+"/"+fBasename+"-"+TString::Itoa(timeReference,10)+"-"+TString::Itoa(length,10)+".gwf";        

        if(nFrames > 0) {
            
            int ret = gSystem->Rename(tmpOutput, output.Data());
            if (ret < 0) gPrint->Error(__METHOD_NAME__, (TString) "Failed to save frames into `"+output.Data()+"` (do you have access right?).");
            else {
                if(bVerbose) gPrint->Message(__METHOD_NAME__, (TString) "%d frame(s) saved into `"+output.Data()+"`.", nFrames);
                writtenFiles.push_back(output);
            }

            gSystem->Unlink(tmpDir);
        }

        return writtenFiles;
    }

    std::vector<TString> TFrameChain::WriteROOT(TString fName, Option_t *opt, std::vector<int> globalIDs)
    {
        std::vector<TString> writtenFiles;

        if(gPrint->ErrorIf(fName.EqualTo(""), __METHOD_NAME__, (TString) "No output file provided. Cannot save selected frames."))
            return writtenFiles;

        TFile *f = NULL;
        TString option = opt;
                option.ToLower();

        TList *outputList = new TList;
        TList *dirList = new TList;

        for(int k = 0, N = globalIDs.size() ? globalIDs.size() : this->GetNFrames(); k < N; k++)
        {
            int globalID = globalIDs.size() ? globalIDs[k] : k;
            if (globalIDs.size() && std::find(globalIDs.begin(), globalIDs.end(), globalID) == globalIDs.end()) continue;

            TFrame* frame = this->GetFrame(globalID);
            int fileID = this->GetFileID(globalID);
            int frameID = this->GetFrameID(globalID);

            if(frame->GetDataBuffer().size()) {
                // @TODO: Implement data buffer saving (if needed)
                gPrint->Warning(__METHOD_NAME__, (TString) "Data buffer is not empty (these objects will be ignored). Commit first, then convert to ROOT.");
            }
            
            if(gPrint->WarningIf(frame == NULL, __METHOD_NAME__, (TString) "Failed to find frame #"+gPrint->Pad(frameID+1, this->GetNFrames())+" from file #"+gPrint->Pad(fileID+1, this->GetNFiles())))
                continue;

            gPrint->ProgressBar(Form((TString) "Saving frame into `%s`.", fName.Data()), "noclean", k+1, N);
            std::vector<TFrameProxyT> proxyMap = frame->ReadProxy();

            TString _fName = fName;
            if(TFileReader::HasPattern(fName)) {

                std::vector<TString> identifiers;
                                     identifiers.push_back(TString::Itoa(frame->GetGPS().start,10));
                                     identifiers.push_back(TString::Itoa(frame->GetTimeLength(),10));

                _fName = TFileReader::SubstituteIdentifier(fName, identifiers);
            }

            TString _fopt = "update";
            if(option.Contains("recreate")) _fopt = "recreate";
            if(option.Contains("read_without_globalregistration")) _fopt = "read_without_globalregistration";
            if(option.Contains("new")) _fopt = "new";
            if(option.Contains("net")) _fopt = "net";
            if(option.Contains("web")) _fopt = "web";

            if(f == NULL || !_fName.EqualTo(f->GetName())) {

                if(f) {
                            
                    TIter Next(dirList);
                    TDirectory *structureDir = NULL;
                    while( (structureDir = (TDirectory*) Next()) )
                        structureDir->Write();
    
                    dirList->Clear();
                    f->Close();
                }

                f = new TFile(_fName, _fopt);
                writtenFiles.push_back(_fName);
            }
            
            bool bHisto = option.Contains("histo");
            for(int i = 0, I = proxyMap.size(); i < I; i++) {

                TFrameProxyAbstract *proxy = std::visit([](auto obj) -> TFrameProxyAbstract* { return (TFrameProxyAbstract*) obj; }, proxyMap[i]);
                for(int j = 0, J = proxy->GetNBaskets(); j < J; j++) {

                    TString name = proxy->GetName(j);
        
                    if(J > 512)
                        gPrint->ProgressBar(Form((TString) "Saving basket `%s`.", name.Data()), "", i+1, I);

                    TTree *tree = (TTree*) outputList->FindObject(name);
                    if(tree == NULL) {

                        TDirectory *dir = TFileReader::RecursiveMkdir(proxy->GetIdentifier(), f->GetDirectory("/"));
                        if(!dirList->FindObject(dir)) dirList->Add(dir);

                        outputList->Add(proxy->EmptyTree(name));
                        tree = (TTree*) outputList->FindObject(name);
                        if(tree == NULL) continue;

                        tree->SetDirectory(dir);
                    }

                    proxy->Hydrate(j, tree);
                }

                if(bHisto) {

                    for(int j = 0, J = outputList->GetEntries(); j < J; j++) {
                        
                        TTree *tree = (TTree*) outputList->At(j);
                        if(tree == NULL) continue;

                        TDirectory *dir = TFileReader::RecursiveMkdir(proxy->GetIdentifier(), f->GetDirectory("/"));
                        TH1 *hData = TFrameProxyAbstract::Histogram(tree, proxy->GetBranchName());
                             hData->SetDirectory(dir);
                    }
                }
            }

            gPrint->ClearLine();
        }

        if(f) {

            TIter Next(dirList);
            TDirectory *structureDir = NULL;
            while( (structureDir = (TDirectory*) Next()) )
                structureDir->Write();

            dirList->Clear();
            f->Close();
        }

        delete outputList;
        delete dirList;

        return writtenFiles;
    }

    TObject* TFrameChain::GetObject(TString objName, std::vector<int> globalIDs)
    {
        TTree *tree = NULL;

        for(int i = 0, N = globalIDs.size() ? globalIDs.size() : this->GetNFrames(); i < N; i++)
        {
            int globalID = globalIDs.size() ? globalIDs[i] : i;
            if (globalIDs.size() && std::find(globalIDs.begin(), globalIDs.end(), globalID) == globalIDs.end()) continue;

            TFrame* frame = this->GetFrame(globalID);
            int fileID = this->GetFileID(globalID);
            int frameID = this->GetFrameID(globalID);

            if(gPrint->WarningIf(frame == NULL, __METHOD_NAME__, (TString) "Failed to find frame #"+gPrint->Pad(frameID+1, this->GetNFrames())+" from file #"+gPrint->Pad(fileID+1, this->GetNFiles())))
                continue;

            std::vector<TFrameProxyT> proxyMap = frame->ReadProxy();
            for(int i = 0, I = proxyMap.size(); i < I; i++) {

                TFrameProxyAbstract *proxy = std::visit([](auto obj) -> TFrameProxyAbstract* { return (TFrameProxyAbstract*) obj; }, proxyMap[i]);
                for(int j = 0, J = proxy->GetNBaskets(); j < J; j++) {

                    TString _objName = proxy->GetName(j);
                    TString _objTitle = proxy->GetMessage(j);
                    if(!objName.EqualTo(_objName)) continue;
                    
                    if (tree == NULL)
                        tree = proxy->EmptyTree(_objName, _objTitle);

                    proxy->Hydrate(j, tree);
                }
            }
        }

        return (TObject*) tree;
    }

    TObject *TFrameChain::GetObject(TString objName, TString branchName, Option_t *option, std::vector<int> globalIDs)
    {
        TTree *tree = (TTree *) this->GetObject(objName, globalIDs);
        if(tree == NULL) return NULL;
            
        TH1 *hData = TFrameProxyAbstract::Histogram(tree, branchName, option);
        tree->Delete();

        return (TObject *) hData;
    }


    TString TFrameChain::GetDetector  (int fileID, int frameID) const
    {
        TFrame *frame = this->GetFrame(fileID, frameID);
        if(frame == NULL) return "";

        return frame->GetDetector();
    }
    
    TFrameChain *TFrameChain::Allocate(const char *fName, const char *detector, const char *ftitle, int nFrames, double timestamp, int timeLength, int compression, Verbosity verbosity)
    {
        if (!TString(fName).EndsWith(".gwf")) {
            gPrint->Error(__METHOD_NAME__, "%s only supports GWF files, not `%s`", __METHOD_NAME__.Data(), fName);
            return NULL;
        }
        
        if(TFileReader::HasPattern(fName)) {
            gPrint->Error(__METHOD_NAME__, "Output contains a pattern, which is not supported for GWF files.");
            return NULL;
        }
        
        if( timeLength < 1 ) {
            gPrint->Error(__METHOD_NAME__, "Time length must be greater than 0.");
            return NULL;
        }

        // Allocate frames
        TString fDir = TFileReader::DirName(fName);
        TString fBasename = TFileReader::BaseName(fName, ".gwf");
        gSystem->mkdir(fDir, true);

        if (verbosity >= Verbosity::kDebug) {
            std::cout << "==> Allocating " << nFrames << " frames under the path \"" << (fDir+"/"+fBasename+"-%d-%d.gwf") << "\".." << timestamp << std::endl;
            std::cout << TString(fDir+"/"+fBasename) << std::endl;
        }
       
        SetDebug(0);
        FrFile *f = FrFileAllocation(fDir+"/"+fBasename+"-", detector, ftitle, nFrames, timestamp, timeLength, compression);
        SetDebug(GetVerbosity());

        std::vector<TString> fNames = FrFileNames(fDir+"/"+fBasename+"-%.0f-%d.gwf", nFrames, timestamp, timeLength);
       
        SetDebug(0);
        TFrameChain *chain = new TFrameChain(ROOT::IOPlus::Helpers::Implode(" ", fNames));
        SetDebug(GetVerbosity());

        return chain;
    }

    TFrameChain *TFrameChain::Merge(TFrameChain *f, GPS gpstime)
    {
        std::vector<TFrame *> frameList;
        std::vector<int> globalIDs;

        if(gpstime != GPS()) globalIDs = this->SelectByGPS(gpstime);
        else globalIDs.clear();
        for (int i = 0, N = globalIDs.size(); i < N; i++) {
            frameList.push_back(this->GetFrame(globalIDs[i]));
        }

        if(gpstime != GPS()) globalIDs = f->SelectByGPS(gpstime);
        else globalIDs.clear();
        for (int i = 0, N = globalIDs.size(); i < N; i++) {
            frameList.push_back(f->GetFrame(globalIDs[i]));
        }

        return new TFrameChain(frameList);
    }

    TFrameChain *TFrameChain::Extract(std::vector<int> globalIDs)
    {
        std::vector<TFrame *> frameList;
        for (int i = 0, N = globalIDs.size(); i < N; i++) {
            frameList.push_back(this->GetFrame(globalIDs[i]));
        }

        return new TFrameChain(frameList);
    }
   
    TFrameChain *TFrameChain::Reshape(double dt, Option_t *option, std::vector<int> globalIDs, GPS gpstime)
    {
        TString tmpDir = ROOT::IOPlus::Helpers::TempDirectory(6);
        gSystem->mkdir(tmpDir, true);

        gpstime = (gpstime == GPS()) ? this->GetGPS() : gpstime;

        // Create a buffer allocation to get proper frame structure.
        FrameFormat::TFrameChain *f0 = FrameFormat::TFrameChain::Allocate(tmpDir+"/buffer.gwf", this->GetDetector(), "", (gpstime.end - gpstime.start) / dt, gpstime.start, dt, this->GetCompression());
        if(f0 == NULL) {
            gPrint->Error(__METHOD_NAME__, (TString) "Failed to allocate buffer for reshaping frames.");
            return NULL;
        }

        // Loop over available frames and reshape candidates from this chain
        for( auto frame0 : *f0) {

            frame0->Print();
            this->SetDebug(0);            
            FrameH *frame0H = f0->GetManager().Read(&frame0->GetRHead(), frame0->GetFileID(), frame0->GetFrameID());
            this->SetDebug(f0->GetVerbosity());

            if(frame0H == NULL) {
                gPrint->Error(__METHOD_NAME__, (TString) "Failed to read reference frame with GPS time %.2f", frame0->GetGPS().start);
                continue;
            }

            std::vector<TFrame *> frames = this->FindByGPS(frame0->GetGPS());
            frames.erase(
                std::remove_if(frames.begin(), frames.end(), [&](TFrame* frame) {
                    int globalID = this->GetGlobalID(frame);
                    return globalIDs.size() && std::find(globalIDs.begin(), globalIDs.end(), globalID) == globalIDs.end();
                }),
                frames.end()
            );
            
            if( this->fManager->Reshape(frames, frame0H) == nullptr) {

                gPrint->Error(__METHOD_NAME__, (TString) "Failed to reshape frame with GPS time %.2f", frame0->GetGPS().start);
                continue;

            }

            FrameH *h = f0->GetManager().Read(&frame0->GetRHead(), frame0->GetFileID(), frame0->GetFrameID());
            
            if(frame0->Write()) {
                gPrint->Debug(1, __METHOD_NAME__, (TString) "Frame with GPS time %.2f reshaped and saved.", frame0->GetGPS().start);
            } else {
                gPrint->Error(__METHOD_NAME__, (TString) "Failed to save reshaped frame with GPS time %.2f", frame0->GetGPS().start);
            }

            frame0->Free();
        }

        f0->Reload();
        return f0;
    }

    TFrameChain *TFrameChain::Trend(int fs, Option_t *option, std::vector<int> globalIDs, GPS gpstime)
    {
        UNUSED(option); // ROOT convention: option is not used in this method

        if(gPrint->ErrorIf(fName.EqualTo(""), __METHOD_NAME__, (TString) "No output file provided. Cannot save selected frames."))
            return NULL;

        TString tmpDir = ROOT::IOPlus::Helpers::TempDirectory(6);

        gPrint->Debug(1, __METHOD_NAME__, (TString) "Computing trending curves (fs = "+TString::Itoa(fs,10)+"Hz) ..");
        FrameFormat::TFrameChain *f = NULL;

        std::vector<FrAdcData*> adcList;
        for( auto frame : *this) {

            int globalID = this->GetGlobalID(frame);
            if(globalIDs.size() && std::find(globalIDs.begin(), globalIDs.end(), globalID) == globalIDs.end()) 
                continue;

            std::vector<TFrameProxyT> proxies = frame->ReadProxy();
            for(int k = 0, K = proxies.size(); k < K; k++) {

                std::visit([&adcList](auto* obj) -> void { 

                    if(!obj->GetType().Contains("FrameFormat::TFrAdcData")) return;
                    for(int b = 0, B = obj->GetNBaskets(); b < B; b++) {
                        adcList.push_back(FrCast((TFrAdcData*) obj->Get(b)));
                    }

                }, proxies[k]);
            }

            if (f == NULL && adcList.size() > 0) {

                double timestamp  = (gpstime == GPS()) ? frame->GetGPS().start : gpstime.start;
                double timeLength = (gpstime == GPS()) ? frame->GetTimeLength() : gpstime.end - gpstime.start;
                f = FrameFormat::TFrameChain::Allocate(tmpDir+"/buffer.gwf", this->GetDetector(), "Trending Curves", this->GetNFrames(), timestamp, timeLength, this->GetCompression());
            }

            std::vector<FrTrendData*> trendList = this->fManager->Trend(fs, adcList, gpstime);
            for(int i = 0, N = trendList.size(); i < N; i++) {

                f->GetFrame(0, this->GetGlobalID(frame))->Add(RootCast(trendList[i]->nADC));
                f->GetFrame(0, this->GetGlobalID(frame))->Add(RootCast(trendList[i]->minADC));
                f->GetFrame(0, this->GetGlobalID(frame))->Add(RootCast(trendList[i]->maxADC));
                f->GetFrame(0, this->GetGlobalID(frame))->Add(RootCast(trendList[i]->meanADC));
                f->GetFrame(0, this->GetGlobalID(frame))->Add(RootCast(trendList[i]->rmsADC));
            }

            adcList.clear();

            for(auto trend : trendList) {
                FrTrendDataFree(trend);
            }
            
            trendList.clear();
        }

        return f;
    }

    TFrameChain &TFrameChain::Commit(Option_t *opt, int framesPerFile)
    {
        if(this->fName.EqualTo("")) {
            gPrint->Error(__METHOD_NAME__, (TString) "Cannot commit without file support. Please write frames first.");
            return *this;
        }

        if(this->GetEntries() < 1) {
            gPrint->Error(__METHOD_NAME__, (TString) "No frames to commit. Please add frames first.");
            return *this;
        }
        
        TFrame* firstFrame = *this->begin();
        if (firstFrame != nullptr) {
            
            std::vector<TString> files = this->Write(ROOT::IOPlus::Helpers::TempDirectory(6)+"/buffer.gwf", opt, framesPerFile);
            this->Open(ROOT::IOPlus::Helpers::Implode(" ", files), "", 
                firstFrame->GetTags(), firstFrame->GetDataClasses(), firstFrame->GetDecimate(),
                firstFrame->GetChannelGroups(), firstFrame->GetChannelNumbers(), verbosity
            );
        }

        return *this;
    }

    template <typename T>
    bool TFrameChain::AddChannel(const char *channel, double fs, std::vector<T> adcData, double timestamp, TString unitX, TString unitY)
    {
        double gpstime = timestamp < 0 ? this->GetGPS().start : timestamp;
        std::vector<TFrame*> frames = this->FindByGPS(gpstime);

        TFrame *frame = !frames.empty() ? frames.front() : nullptr;
        if(frame == NULL) return false;

        int gpstime0 = frame->GetGPS().start;
        int idx = int((gpstime - gpstime0) * fs);

        while (!adcData.empty()) {

            int remainingSamples = frame->GetTimeLength() * fs - idx;
            int samplesToAdd = TMath::Min(remainingSamples, static_cast<int>(adcData.size()));
            
            TFrAdcData *adc = new TFrAdcData();
                        adc->name       = channel;
                        adc->sampleRate = fs;
                        adc->nBits      = sizeof(T);
                        adc->data.name  = channel;
                        adc->data.type  = TFrVect::TYPE::FR_VECT_8R;
                        adc->data.nData = samplesToAdd;
                        adc->data.data  = VcCast<double>(adcData, samplesToAdd);
                        adc->data.nDim  = 1;
                        adc->data.nBytes = samplesToAdd * sizeof(T);

                        adc->data.nx.push_back(samplesToAdd);
                        adc->data.dx.push_back(1.0 / fs);
                        adc->data.startX.push_back(gpstime0 + idx / fs);
                        adc->data.unitX.push_back(unitX.Data());

                        adc->data.unitY = unitY.Data();

            frame->Add(adc);

            adcData.erase(adcData.begin(), adcData.begin() + samplesToAdd);
            idx = 0;

            if (!adcData.empty()) {
                frame = this->GetFrame(this->GetFileID(this->GetGlobalID(frame) + 1), 0);
                if (frame == NULL) return false;
                gpstime0 = frame->GetGPS().start;
            }
        }

        return true;
    }

    void TFrameChain::RemoveChannel(const char *channel)
    {
        for(int i = 0, N = this->GetNFrames(); i < N; i++) {
            
            TFrame *f = this->GetFrame(this->GetFileID(i), this->GetFrameID(i));
            f->Remove(channel);
        }
    }

    /// \cond NOPE
    template bool TFrameChain::AddChannel<float>(const char*, double, std::vector<float>, double, TString, TString);
    template bool TFrameChain::AddChannel<double>(const char*, double, std::vector<double>, double, TString, TString);
    template bool TFrameChain::AddChannel<char>(const char*, double, std::vector<char>, double, TString, TString);
    template bool TFrameChain::AddChannel<short>(const char*, double, std::vector<short>, double, TString, TString);
    template bool TFrameChain::AddChannel<int>(const char*, double, std::vector<int>, double, TString, TString);
    template bool TFrameChain::AddChannel<long>(const char*, double, std::vector<long>, double, TString, TString);
    template bool TFrameChain::AddChannel<unsigned long>(const char*, double, std::vector<unsigned long>, double, TString, TString);
    template bool TFrameChain::AddChannel<std::complex<float>>(const char*, double, std::vector<std::complex<float>>, double, TString, TString);
    template bool TFrameChain::AddChannel<std::complex<double>>(const char*, double, std::vector<std::complex<double>>, double, TString, TString);
    /// \endcond

};