#include "FrameFormat/Impl/TFrameHeader.h"

// CHAR   = char           = 1 byte
// CHAR_U = unsigned char  = 1 byte
// INT_2S = short          = 2 bytes
// INT_2U = unsigned short = 2 bytes
// INT_4S = int            = 4 bytes
// INT_4U = unsigned int   = 4 bytes
// INT_8S = long          = 8 bytes
// INT_8U = unsigned long = 8 bytes
// REAL_4 = float         = 4 bytes
// REAL_8 = double        = 8 bytes
// STRING = char[]          = < 65536 bytes
// DOUBLE = double          = 6 bytes
// COMPLEX_8  = <float,float>   = 8 bytes
// COMPLEX_16 = <double,double> = 16 bytes

FrameH *FrNewFrame(const char *name, double timestamp, double frameStep) {

        FrameH *frame = FrameNew(const_cast<char*>(name));
        if (frame == NULL) {
                fprintf( stderr, "*** Cannot create output frame\n" );
                return NULL;
        }

        frame->run = 0;
        frame->frame = 0;
        frame->dataQuality = 0;
        frame->GTimeS = gClockGPS->TimeS(timestamp);
        frame->GTimeN = gClockGPS->TimeN(timestamp);
        frame->ULeapS = gClockGPS->GetLeapSeconds(timestamp);
        frame->dt = (double) frameStep;

        return frame;
}

std::vector<TString> FrFileNames(const char *pattern, int nFrames, double timestamp, int frameStep) {

        std::vector<TString> expandedNames;

        double extraTime = timestamp - int(timestamp); // Compensate extraTime by adding an extra frame
        if(extraTime > 0) nFrames += 1; 

        for (int i = 0; i < nFrames; ++i) {
                TString buffer = Form(pattern, timestamp + i * frameStep, frameStep);
                expandedNames.push_back(buffer);
        }

        return expandedNames;
}

std::vector<TString> FrFileNames(FrFile *file)
{
        std::vector<TString> fileNames;
        if(file == NULL) return fileNames;

        FrFileH *fileHeader = file->fileH;
        do { fileNames.push_back(fileHeader->fileName); }
        while ((fileHeader = fileHeader->next));

        return fileNames;
}


int FrCompressionLevel(FrFile *file)
{
        if (file == NULL) return -1;
        return file->compress;
}

FrFile *FrFileAllocation(const char *prefix, const char *detector, const char *title, int nFrames, double timestamp, int frameStep, int compression) {
        
        TString name = "";

        double extraTime = timestamp - int(timestamp); // Compensate extraTime by adding an extra frame
        if(extraTime > 0) nFrames += 1; 

        for (int i = 0; i < nFrames; ++i) {
        
		const char *outfilename = Form("%s%.0f-%d.gwf", prefix, timestamp + i * frameStep, frameStep );
                name += TString(outfilename) + " ";

                FrFile *ffo = FrFileONew(outfilename, compression);
                if ( ffo == NULL ) {
                        fprintf( stderr, "*** Cannot open output frame file %s\n",outfilename);
                        return NULL;
                }

                FrameH *outframe = FrNewFrame(detector, timestamp + i * frameStep, frameStep);
                if ( outframe == NULL ) return NULL;

                FrFileOSetMsg( ffo, const_cast<char *>(title) );

                /*-- Write the output frame to the file --*/
                if( FrameWrite(outframe,ffo) != FR_OK ) {
                        fprintf( stderr, "*** Error writing output frame to file %s\n",
                                outfilename );
                        return NULL;
                }

                /*-- Close output frame file and free output frame --*/
                FrFileOEnd( ffo );
                FrameFree( outframe );
        }
        
        return FrFileINew(name.Data());
}

FrTrendData *FrTrendDataNew(int fs, FrAdcData* channel)
{
        if(channel == NULL || channel->data == NULL) return NULL;

        double fsOld = channel->sampleRate;
        size_t dataSize = channel->data->nData;
        if (dataSize == 0 || fs <= 0 || fsOld <= 0) return NULL;

        double minVal = NAN;
        double maxVal = NAN;
        double sum = 0.0;
        double sumsq = 0.0;

        std::vector<double> data = FrReadData(channel->data);
        for (size_t i = 0; i < dataSize; ++i)
        {
                double value = data[i];
                minVal = ROOT::IOPlus::Helpers::IsNaN(minVal) ? value : std::min(value, minVal);
                maxVal = ROOT::IOPlus::Helpers::IsNaN(maxVal) ? value : std::max(value, maxVal);
                sum += value;
                sumsq += value * value;
        }

        double mean = sum / dataSize;
        double rms = std::sqrt(sumsq / dataSize);

        // Calculate the number of data points for the new sampling rate
        size_t newDataSize = static_cast<size_t>(std::ceil(dataSize * fs / fsOld));

        // Allocate and populate trend data
        FrTrendData *trend = new FrTrendData();
        trend->name = strdup(channel->name);
        trend->inadc = channel;
        trend->n = dataSize;
        trend->min = minVal;
        trend->max = maxVal;
        trend->sum = sum;
        trend->sumsq = sumsq;

        trend->nADC    = FrAdcDataNew(NULL, strdup((std::string(channel->name) + ".nADC").c_str()), fs, newDataSize, 32);
        trend->minADC  = FrAdcDataNew(NULL, strdup((std::string(channel->name) + ".minADC").c_str()), fs, newDataSize, -64);
        trend->maxADC  = FrAdcDataNew(NULL, strdup((std::string(channel->name) + ".maxADC").c_str()), fs, newDataSize, -64);
        trend->meanADC = FrAdcDataNew(NULL, strdup((std::string(channel->name) + ".meanADC").c_str()), fs, newDataSize, -64);
        trend->rmsADC  = FrAdcDataNew(NULL, strdup((std::string(channel->name) + ".rmsADC").c_str()), fs, newDataSize, -64);

        // Populate ADC data by aggregating over intervals
        for (size_t i = 0; i < newDataSize; ++i) {

                size_t startIdx = static_cast<size_t>(i * fsOld / fs);
                size_t enddx = static_cast<size_t>((i + 1) * fsOld / fs);
                if (startIdx >= dataSize) break;

                double minVal = NAN;
                double maxVal = NAN;
                double sum = 0.0;
                double sumsq = 0.0;
                size_t count = 0;

                for (size_t j = startIdx; j < enddx && j < dataSize; ++j) {

                        double value = data[j];

                        minVal = ROOT::IOPlus::Helpers::IsNaN(minVal) ? value : std::min(value, minVal);
                        maxVal = ROOT::IOPlus::Helpers::IsNaN(maxVal) ? value : std::max(value, maxVal);
                
                        sum += value;
                        sumsq += value * value;

                        count++;
                }

                trend->nADC->data->dataI[i] = count;
                trend->minADC->data->dataD[i] = minVal;
                trend->maxADC->data->dataD[i] = maxVal;
                trend->meanADC->data->dataD[i] = count > 0 ? sum / count : 0.0;
                trend->rmsADC->data->dataD[i] = count > 0 ? std::sqrt(sumsq / count) : 0.0;
        }

        return trend;
}

void FrTrendDataFree(FrTrendData *trendata)
{
        if(trendata == NULL) return;

        if(trendata->nADC    != NULL) FrAdcDataFreeOne(trendata->nADC);
        if(trendata->minADC  != NULL) FrAdcDataFreeOne(trendata->minADC);
        if(trendata->maxADC  != NULL) FrAdcDataFreeOne(trendata->maxADC);
        if(trendata->meanADC != NULL) FrAdcDataFreeOne(trendata->meanADC);
        if(trendata->rmsADC  != NULL) FrAdcDataFreeOne(trendata->rmsADC);
}

FrEndOfFile *FrEndOfFileReadAndReturn(FrFile *iFile)
{
        unsigned int nFrames, chkSumFlag, chkSum, nBytes4, offset4,
        chkSumFrHeader, chkSumFile;

        FRLONG offset, nBytesEnd;

        /*--------open a real file or set pointers for in-memory read if needed---*/
        if(iFile->frfd == NULL && iFile->buf == NULL) FrFileIOpen(iFile);
        if(iFile->error != FR_OK) return(NULL);

        if(iFile->inMemory == FR_NO && iFile->frfd == NULL) return(NULL);

                        /*--do not compute file checksum when using random access ---*/
 
        iFile->chkSumFiFlag = FR_NO;

        /*---------------- read the FrEndOfFile data for differnet format version---*/

        if(iFile->fmtVersion < 6) {

                if(FrFileIOSetFromEnd(iFile, -20) == -1) {return(NULL);}
                FrReadIntU(iFile, &nFrames);
                FrReadIntU(iFile, &nBytes4);
                iFile->nBytesE = nBytes4;
                FrReadIntU(iFile, &chkSumFlag);
                FrReadIntU(iFile, &chkSum);
                FrReadIntU(iFile, &offset4);
                offset = offset4;
      
        }    /*--- skip the record length(INT4) and type(INT2) */

        else if(iFile->fmtVersion < 8) {
                
                if(FrFileIOSetFromEnd(iFile, -28) == -1) {return(NULL);}
                FrReadIntU(iFile, &nFrames);
                FrReadLong(iFile, (FRLONG*) &iFile->nBytesE);
                FrReadIntU(iFile, &chkSumFlag);
                FrReadIntU(iFile, &chkSum);
                FrReadLong(iFile, &offset);
        
        } else {

                if(FrFileIOSetFromEnd(iFile, -32) == -1) {return(NULL);}
                nBytesEnd = FrFileIOTell(iFile)+32;
                if(nBytesEnd == FRIOBSIZE) nBytesEnd = iFile->frfd->nBytesRead;
                FrReadIntU(iFile, &nFrames);
                FrReadLong(iFile, (FRLONG*) &iFile->nBytesE);
                if(nBytesEnd != iFile->nBytesE) {return(NULL);}
                FrReadLong(iFile, &offset);
                FrReadIntU(iFile, &chkSumFrHeader);
                FrReadIntU(iFile, &chkSum);
                FrReadIntU(iFile, &chkSumFile);
                chkSumFlag = 0;
        }

        FrEndOfFile *eof = new FrEndOfFile();
        eof->classe = FrEndOfFrameDef();
        eof->nFrames = nFrames;
        eof->nBytes = iFile->nBytesE;
        eof->seekTOC = offset;

        /*---------------------------------------------- check the file checksum---*/

        if(iFile->chkTypeFiRead == 0) {}
        else if(iFile->chkSumFiFlag == FR_NO) {}
        else if(iFile->chkSumFiRead  == iFile->chkSumFi) {}
        else {
        iFile->error = FR_ERROR_CHECKSUM;
        return eof;}

        eof->chkType = iFile->chkTypeFiRead;
        eof->chkSum = iFile->chkSumFi;
        return eof;
}

void FrLibSetLvlT(int level) { 
        FrDebugLvl = level;
        FrLibSetLvl(level);
}

// This is a partial reading of the frame mainly for rawData reading
// NB: Implementation for other variables should be verified and/or implemented (detector, auxiliary data, ..)
FrameH *FrameReadT(FrFile *iFile, double gtime, const char *tag)
{
        FrLibSetLvl(0);
        
        /** Prepare an FrameH with minimal information **/
        FrameH *frame = FrameHReadT(iFile,  gtime);
        if(frame == NULL) return(NULL);
        if(strlen(tag) == 0) return frame;

        FrVect *snr, *gps;
        int nData, i, index;
        FrTag* word, *listOfWord;
        char star[] = "*";
        char *name;
        double sampleRate;

        FrTOCdet *detector = iFile->toc->detector;
        if(detector != nullptr) frame->detectProc = FrDetectorNew(detector->name);

        /** ADC Channel with specific tag only */
        FrRawDataNew(frame);
        if(frame->rawData != NULL) {
                
                FrAdcData *firstAdc = FrAdcDataReadT(iFile, const_cast<char *>(tag), gtime);
                if (firstAdc != NULL) frame->rawData->firstAdc = firstAdc;
                FrSerData *firstSer = FrSerDataReadT(iFile, const_cast<char *>(tag), gtime);
                if (firstSer != NULL) frame->rawData->firstSer = firstSer;
        }

        frame->procData = FrProcDataReadT(iFile, const_cast<char *>(tag), gtime);    
        frame->simData  = FrSimDataReadT (iFile, const_cast<char *>(tag), gtime);
        frame->simEvent = FrSimEventReadT(iFile, const_cast<char *>(tag), gtime, frame->dt, 0., 1.e37);
        frame->event    = FrEventReadT   (iFile, const_cast<char *>(tag), gtime, frame->dt, 0., 1.e37);

        if(strstr(const_cast<char *>(tag),"EVENT_SNR:") != 0) {
                listOfWord = FrTagNew(const_cast<char *>(tag));

                for(word = listOfWord; word != NULL; word = word->next)
                {if(strncmp(word->start,"EVENT_SNR:",10) == 0) 
                {name = word->start+10;
                        if(strncmp(name,"ALL",3) == 0) name = star;
                        gps = FrFileIGetEventInfo(iFile, name, gtime, frame->dt, 0., 1.e37);}
                else if(strncmp(word->start,"SIM_EVENT_SNR:",14) == 0) 
                {name = word->start+14;
                        if(strncmp(name,"ALL",3) == 0) name = star;
                        gps = FrFileIGetSimEventInfo(iFile, name, gtime, frame->dt, 0, 1.e37);}
                else {continue;}

                sampleRate = 1000.;
                nData = frame->dt*sampleRate + 0.1;
                snr = FrVectNewTS(word->start, sampleRate, nData, -32);
                if(snr == NULL) return(frame);
                snr->GTime = gtime;

                if(gps != NULL)
                {for(i=0; i<gps->nData; i++)
                {index = FrVectGetIndex(snr, gps->dataD[i]-gtime);
                snr->dataF[index] += gps->next->dataF[i];}}

                FrProcDataNewV(frame, snr);
                FrVectFree(gps);}

                FrTagFree(listOfWord);
        }

        return(frame);
}

int FrScanFile(FrFile *file, bool decompress) // Sequential access
{
        FrameH *frame = NULL;
        
        if(file == nullptr) return 0;
        FrFileIRewind(file);

        short compress = file->compress;
        if(!decompress) file->compress = -1;

        int nFrame = 0, nPrint = 1;

        if (FrDebugLvl >0) printf("\n  Sequential file reading. Please wait....\n");
        file->chkSumFiFlag = FR_YES;
        file->chkSumFrFlag = FR_NO;

        while((frame = FrameReadRecycle(file, frame)) != NULL) {

                nFrame++;
                if(nFrame != nPrint) continue;
                nPrint *= 2;
                if(FrDebugLvl) printf("%9d frames read\n",nFrame);
        }
        
        if(FrDebugLvl > 0) printf("   %d frame(s) in file.",nFrame);
        if(file->error != FR_OK) {

                fprintf(stderr," Read Error. Last errors are:\n%s\n",FrErrorGetHistory());
                printf("\n");

                file->compress = compress;
                return 0;

        } else if(FrDebugLvl > 0) {
                printf(" No read error.");
        }

        if(file->chkTypeFiRead == 0) {

                if(FrDebugLvl > 0) printf(" This file does not have a file checksum\n");
        
        } else {
        
                if(file->chkSumFiRead == file->chkSumFi) {
                        if(FrDebugLvl > 0) printf(" File Checksum OK (%x)\n",file->chkSumFiRead);
                } else {
                        if(FrDebugLvl > 0) printf(" File Checksum error:\n       read = %x\n"
                                                "   computed = %x\n", file->chkSumFiRead, file->chkSumFi);

                        file->compress = compress;
                        return 0;
                }
        }

        file->compress = compress;
        return(nFrame);
}

int FrScanFile(FrFile *file, bool decompress, double gpstime0, double gpstime1) // Random access
{
        FrameH *frame = NULL;
        int nFrame = 0, nPrint = 1;
        double gTimeS;

        if(file == nullptr) return 0;
        FrFileIRewind(file);

        short compress = file->compress;
        if(!decompress) file->compress = -1;

        if (FrDebugLvl >0) printf("\n  Reading using the TOC. Please wait....\n");
        file->chkSumFiFlag = FR_NO;
        file->chkSumFrFlag = FR_YES;

        if(FrDebugLvl > 0)
        printf("   Requested time range: %9.0f - %9.0f\n",gpstime0, gpstime1);

        gTimeS = FrFileITStart(file);
        if(gTimeS < 0) {
                if(FrDebugLvl > 0) printf(" No TOC available\n");
                file->compress = compress;
                return 0;
        }
        
        if(FrDebugLvl > 0) printf("        File time range: %9.0f - %9.0f\n", gTimeS, FrFileITEnd(file));
        if(gTimeS < gpstime0) gTimeS = gpstime0;

        while((frame = FrameReadT(file, gTimeS)) != NULL) 
        {
                nFrame++;
                if(nFrame == nPrint)
                {
                        nPrint *= 2;
                        if(FrDebugLvl) printf("%9d frames read\n",nFrame);
                }

                gTimeS = FrFileITNextFrame(file, gTimeS);
                if(gTimeS >= gpstime1 || gTimeS < 0.) break;
                FrameFree(frame);
        }
        
        if(FrDebugLvl > 0) printf("   %d frame(s) in file.",nFrame);
        if(file->error != FR_OK) {

                fprintf(stderr," *** Read Error for GPS: %.0f\n Last errors are:\n%s", gTimeS, FrErrorGetHistory());
                file->compress = compress;
                return 0;

        } else if(FrDebugLvl > 0) {
                printf(" No read error.");
        }
        
        if(FrDebugLvl > 0) 
        {
                if(file->chkTypeFrRead == 0) {
                        printf(" This file does not have structure checksums\n");}
                else {
                        if(file->fmtVersion < 8) printf(" Frame Checksums OK\n");
                        else                     printf(" Structure Checksums OK\n");
                }
        } 
        
        file->compress = compress;
        return(nFrame);
}


/* Based on MyFrVectCmp from Frame_Compare_Channels.c by Ben Johnson */
int FrVectDiff(FrVect *vect0, FrVect *vect1, char* name1, char* name2, int debugLvl, double minSampleRate, int useMinLength)
{
        FRULONG i, nData;
        double mytime = 0;
        FRULONG *d0,*d1;

        if (vect0 == NULL) {
                if (debugLvl > 0) fprintf(stderr,"Null pointer to data for channel %s in file1.\n",name1);
                return 0;
        }
        if (vect1 == NULL) {
                if (debugLvl > 0) fprintf(stderr,"Null pointer to data for channel %s in file2.\n",name2);
                return 0;
        }

        if (minSampleRate > 0.0) {
        
                if (vect0->dx[0] > 1.0/minSampleRate) {
                        if (debugLvl > 1) fprintf(stdout,"Ignoring channel %s in file1 with sample rate %0.9lf < %0.9lf .\n",name1,1.0/vect0->dx[0],minSampleRate);
                        return 0;
                }

                if (vect1->dx[0] > 1.0/minSampleRate) {
                        if (debugLvl > 1) fprintf(stdout,"Ignoring channel %s in file2 with sample rate %0.9lf < %0.9lf .\n",name1,1.0/vect1->dx[0],minSampleRate);
                        return 0;
                }
        }

        if (debugLvl > 2) fprintf(stderr,"Data for channel %s have sample rates in file1 and file2: %0.9lf %0.9lf respectively.\n",name1,1.0/vect0->dx[0],1.0/vect1->dx[0]);

        if (vect0->dx[0] != vect1->dx[0]) {

                if (debugLvl > 0) fprintf(stderr,"Data for channel %s have different sample rates in file1 and file2: %0.9lf %0.9lf respectively.\n",name1,1.0/vect0->dx[0],1.0/vect1->dx[0]);
                return 0; 
        }

        if (useMinLength) {

                /* only compare data for the length of the shorter frame */
                if (vect1->nData < vect0->nData) nData = vect1->nData;
                else nData = vect0->nData;
        
        } else {

                if (vect0->nData != vect1->nData) {

                        if (debugLvl > 0) fprintf(stderr,"Data for channel %s have different lengths in file1 and file2: %li %li respectively.\n",name1,vect0->nData,vect1->nData);
                        return 0;
                }
                
                nData = vect0->nData;
        }

        if(vect0->GTime != 0) mytime = vect0->GTime;

        if (debugLvl > 1) fprintf(stdout,"Checking %lu data points for channel %s.\n",nData,name1);

        /*-------------------------------------- data part ---------------*/

        if(vect0->compress != 0) {

                if (debugLvl > 0) fprintf(stderr,"Data for channel %s is compressed and cannot be checked.\n",name1);
                return -1;
        }

        if (vect0->type == FR_VECT_4R) {
                
                float  *d0,*d1;

                d0 = (float *) vect0->data;
                d1 = (float *) vect1->data;

                for(i=0; i<nData; i++) {

                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {

                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %g\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if (vect0->type == FR_VECT_8R) {
       
                double *d0,*d1;

                d0 = (double *) vect0->data;
                d1 = (double *) vect1->data;

                for(i=0; i<nData; i++) {
                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {

                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %g\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_C) {

                char   *d0,*d1;

                d0 = (char   *)vect0->data;
                d1 = (char   *)vect1->data;

                for(i=0; i<nData; i++) {

                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %c, file2 value = %c .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %c\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_2S) {
                
                short  *d0,*d1;

                d0 = (short *) vect0->data;
                d1 = (short *) vect1->data;

                for(i=0; i<nData; i++) {

                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %i, file2 value = %i .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %i\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_8S) {
        
                FRLONG *d0,*d1;

                d0 = (FRLONG *) vect0->data;
                d1 = (FRLONG *) vect1->data;

                for(i=0; i<nData; i++) {
                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %ld, file2 value = %ld .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %ld\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_4S) {
        
                int    *d0,*d1;

                d0 = (int *) vect0->data;
                d1 = (int *) vect1->data;

                for(i=0; i<nData; i++) {
                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %i, file2 value = %i .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }
                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %i\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_1U) {
       
                unsigned char   *d0,*d1;

                d0 = (unsigned char *)vect0->data;
                d1 = (unsigned char *)vect1->data;

                for(i=0; i<nData; i++) {
                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %c, file2 value = %c .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }
                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %c\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_2U) {
        
                unsigned short  *d0,*d1;

                d0 = (unsigned short *) vect0->data;
                d1 = (unsigned short *) vect1->data;

                for(i=0; i<nData; i++) {

                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %i, file2 value = %i .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }
                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %i\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_8U) {

                FRULONG *d0,*d1;
                d0 = (FRULONG *) vect0->data;
                d1 = (FRULONG *) vect1->data;

                for(i=0; i<nData; i++) {

                        if (d0[i] != d1[i]) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %ld, file2 value = %ld .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %ld\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_4U) {
                unsigned int    *d0,*d1;

                d0 = (unsigned int *) vect0->data;
                d1 = (unsigned int *) vect1->data;

                for(i=0; i<nData; i++) {

                        if (d0[i] != d1[i]) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %i, file2 value = %i .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %i\n",mytime,d0[i]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_8C){
        
                /*------------- complex float ------*/
                float *d0,*d1;

                d0 = (float *) vect0->data;
                d1 = (float *) vect1->data;

                for(i=0; i<nData; i++) {

                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[2*i], d1[2*i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[2*i],d1[2*i]);
                                return 0;
                        }

                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[2*i+1], d1[2*i+1])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[2*i+1],d1[2*i+1]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %g %g\n",mytime,d0[2*i],d0[2*i+1]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_16C) {
                
                double *d0, *d1;

                d0 = (double *) vect0->data;
                d1 = (double *) vect1->data;

                for(i=0; i<nData; i++) {
                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[2*i], d1[2*i])) {
                        if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                        if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[2*i],d1[2*i]);
                        return 0;
                        }
                        if (!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[2*i+1], d1[2*i+1])) {
                        if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                        if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[2*i+1],d1[2*i+1]);
                        return 0;
                        }
                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %g %g\n",mytime,d0[2*i],d0[2*i+1]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_8H)  {
                
                float *d0,*d1;

                d0 = (float *) vect0->data;
                d1 = (float *) vect1->data;

                for(i=0; i<nData/2; i++) {

                        if(!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if(d0[nData - i - 1] - d1[nData - i - 1]) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[nData - i - 1],d1[nData - i - 1]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %g %g\n",mytime,d0[i],d0[nData - i - 1]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_16H) {
                
                /*----half complex double -------*/
                double *d0, *d1;

                d0 = (double *) vect0->data;
                d1 = (double *) vect1->data;

                for(i=0; i<nData/2; i++) {

                        if(!ROOT::IOPlus::Helpers::EpsilonEqualTo(d0[i], d1[i])) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[i],d1[i]);
                                return 0;
                        }

                        if(d0[nData - i - 1] - d1[nData - i - 1]) {
                                if (debugLvl > 0) fprintf(stderr,"Data for channel %s differ: \n",name1);
                                if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %g, file2 value = %g .\n",mytime,d0[nData - i - 1],d1[nData - i - 1]);
                                return 0;
                        }

                        if (debugLvl > 2) fprintf(stdout,"%0.9lf %g %g\n",mytime,d0[i],d0[nData - i - 1]);
                        mytime += vect0->dx[0];
                }

        } else if(vect0->type == FR_VECT_STRING) {
        
                char   *d0,*d1;

                d0 = (char   *)vect0->data;
                d1 = (char   *)vect1->data;
                if ( strcmp(d0,d1) != 0) {
                        if (debugLvl > 0) fprintf(stderr,"String data for channel %s differ: \n",name1);
                        if (debugLvl > 0) fprintf(stderr,"for GPS time: %0.9lf file1 value = %s, file2 value = %s .\n",mytime,d0,d1);
                        return 0;
                }
                if (debugLvl > 2) fprintf(stdout,"%0.9lf %s\n",mytime,d0);
        
        } else {
                fprintf(stderr,"Unknown type: %d \n",vect0->type );
                return 0;
        }

        return 1;
}

void FrameDump(FrameH *frame, int level)
{
        return FrameDump(frame, stdout, level);
}

FrEndOfFrame *FrEndOfFrameReadAndReturn(FrFile *iFile)
/*---------------------------------------------------------------------------*/
{
        unsigned int run, frameNumber, instance, gTimeS, gTimeN, chkSumFlag, zero = 0;
        FRULONG nBytes;
        unsigned short instance2;

        if(iFile->fmtVersion == 4) {

                FrBack4EndOfFrameRead(iFile);
                if(iFile->error != FR_OK) return NULL;

                FrEndOfFrame *eof = new FrEndOfFrame();
                              eof->classe  = FrEndOfFrameDef();
                              eof->run     = iFile->run;
                              eof->frame   = iFile->curFrame->frame;
                              eof->chkType = 0;
                              eof->chkSum  = 0;

                return eof;
        }

        if(iFile->fmtVersion > 5) FrReadIntU  (iFile, &instance);
        else FrReadShortU(iFile, &instance2);

        FrReadIntU(iFile, &run);
        FrReadIntU(iFile, &frameNumber);

        FrEndOfFrame *eof = new FrEndOfFrame();
                      eof->classe  = FrEndOfFrameDef();
                      eof->run     = run;
                      eof->frame   = frameNumber;

        /*--------- checksum test for version lower than 8 ---*/
        if(iFile->fmtVersion < 8)
        {
                FrReadIntU(iFile, &iFile->chkTypeFrRead);

                chkSumFlag = iFile->chkSumFrFlag;
                if(iFile->chkTypeFrRead == 1) { /*-- chkSum is set to 0 when computing it--*/
                iFile->chkSumFrFlag = FR_NO;
                FrCksumGnu((char *) &zero, sizeof(int), &(iFile->chkSumFr));}
                FrReadIntU(iFile, &iFile->chkSumFrRead);
                iFile->chkSumFrFlag = chkSumFlag ? FR_YES : FR_NO;

                nBytes = iFile->nBytes - iFile->nBytesF;

                /*--------- check checksum ----------------------*/
                if((iFile->chkTypeFrRead == 1) && (iFile->chkSumFrFlag == FR_YES))
                {

                        FrCksumGnu( NULL, nBytes, &(iFile->chkSumFr));
                        if(iFile->chkSumFrRead  != iFile->chkSumFr) {
                                iFile->error = FR_ERROR_CHECKSUM;
                                return eof;
                        }

                        eof->chkType = iFile->chkTypeFrRead;
                        eof->chkSum  = iFile->chkSumFrRead;
                }

                /*-------------------- checksum for frame format v8 and above ---*/
                else {
                        FrReadIntU(iFile, &gTimeS);
                        FrReadIntU(iFile, &gTimeN);
                        FrReadStructChksum(iFile);
                }

                eof->chkType = iFile->chkTypeFrRead;
                eof->chkSum  = iFile->chkSumFrRead;
                iFile->endOfFrame = FR_YES;
        }
 else {
   FrReadIntU(iFile, &gTimeS);
   FrReadIntU(iFile, &gTimeN);
   FrReadStructChksum(iFile);}

 if(iFile->curFrame != NULL) {

   {if(run != iFile->curFrame->run)
     {FrError(3, (char*) "FrEndOfFrameRead", (char*) "run number missmatch");
      iFile->error = FR_ERROR_BAD_END_OF_FRAME;
      return eof;}

    if(frameNumber != iFile->curFrame->frame)
     {FrError(3, (char*) "FrEndOfFrameRead", (char*) "frame number missmatch");
      iFile->error = FR_ERROR_BAD_END_OF_FRAME;
      return eof;}}
 }
 iFile->endOfFrame = FR_YES;


        return eof;
}
FrameFormat::TFrAdcData *RootCast(FrAdcData *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrAdcData>();
             rootObject->name          = std::string(((FrAdcData*) ptrFr)->name ? ((FrAdcData*) ptrFr)->name : "");
             rootObject->comment       = std::string(((FrAdcData*) ptrFr)->comment ? ((FrAdcData*) ptrFr)->comment : "");
             rootObject->channelGroup  = ((FrAdcData*) ptrFr)->channelGroup;
             rootObject->channelNumber = ((FrAdcData*) ptrFr)->channelNumber;
             rootObject->nBits         = ((FrAdcData*) ptrFr)->nBits;
             rootObject->bias          = ((FrAdcData*) ptrFr)->bias;
             rootObject->slope         = ((FrAdcData*) ptrFr)->slope;
             rootObject->units         = ((FrAdcData*) ptrFr)->units;
             rootObject->sampleRate    = ((FrAdcData*) ptrFr)->sampleRate;
             rootObject->timeOffset    = ((FrAdcData*) ptrFr)->timeOffset;
             rootObject->fShift        = ((FrAdcData*) ptrFr)->fShift;
             rootObject->phase         = ((FrAdcData*) ptrFr)->phase;
             rootObject->dataValid     = ((FrAdcData*) ptrFr)->dataValid;

        if (ptrFr->data) rootObject->data = *RootCast(ptrFr->data);
        if (ptrFr->aux)  rootObject->aux  = *RootCast(ptrFr->aux);

        return rootObject.release();
}

FrameFormat::TFrDetector *RootCast(FrDetector *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrDetector>();
             rootObject->name         = std::string(((FrDetector*) ptrFr)->name ? ((FrDetector*) ptrFr)->name : "");
             rootObject->prefix[0]    = ((FrDetector*) ptrFr)->prefix[0];
             rootObject->prefix[1]    = ((FrDetector*) ptrFr)->prefix[1];
             rootObject->longitude    = ((FrDetector*) ptrFr)->longitude;
             rootObject->latitude     = ((FrDetector*) ptrFr)->latitude;
             rootObject->elevation    = ((FrDetector*) ptrFr)->elevation;
             rootObject->armXazimuth  = ((FrDetector*) ptrFr)->armXazimuth;
             rootObject->armYazimuth  = ((FrDetector*) ptrFr)->armYazimuth;
             rootObject->armXaltitude = ((FrDetector*) ptrFr)->armXaltitude;
             rootObject->armYaltitude = ((FrDetector*) ptrFr)->armYaltitude;
             rootObject->armXmidpoint = ((FrDetector*) ptrFr)->armXmidpoint;
             rootObject->armYmidpoint = ((FrDetector*) ptrFr)->armYmidpoint;

             rootObject->localTime    = ((FrDetector*) ptrFr)->localTime;

        if (ptrFr->aux) rootObject->aux     = *RootCast(ptrFr->aux);
        if (ptrFr->table) rootObject->table = *RootCast(ptrFr->table);

        return rootObject.release();
}

FrameFormat::TFrEvent *RootCast(FrEvent *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrEvent>();
             rootObject->name           = std::string(((FrEvent*) ptrFr)->name ? ((FrEvent*) ptrFr)->name : "");
             rootObject->comment        = std::string(((FrEvent*) ptrFr)->comment ? ((FrEvent*) ptrFr)->comment : "");
             rootObject->inputs         = std::string(((FrEvent*) ptrFr)->inputs ? ((FrEvent*) ptrFr)->inputs : "");
             rootObject->GTimeS         = ((FrEvent*) ptrFr)->GTimeS;
             rootObject->GTimeN         = ((FrEvent*) ptrFr)->GTimeN;
             rootObject->timeBefore     = ((FrEvent*) ptrFr)->timeBefore;
             rootObject->timeAfter      = ((FrEvent*) ptrFr)->timeAfter;
             rootObject->eventStatus    = ((FrEvent*) ptrFr)->eventStatus;
             rootObject->amplitude      = ((FrEvent*) ptrFr)->amplitude;
             rootObject->probability    = ((FrEvent*) ptrFr)->probability;
             rootObject->statistics     = std::string(((FrEvent*) ptrFr)->statistics ? ((FrEvent*) ptrFr)->statistics : "");

             rootObject->nParam         = ((FrEvent*) ptrFr)->nParam;
             rootObject->parameters     = std::vector<double> (((FrEvent*) ptrFr)->parameters    , ((FrEvent*) ptrFr)->parameters     + ((FrEvent*) ptrFr)->nParam);
             rootObject->parameterNames = std::vector<std::string> (((FrEvent*) ptrFr)->parameterNames, ((FrEvent*) ptrFr)->parameterNames + ((FrEvent*) ptrFr)->nParam);
        
        if(ptrFr->data)  rootObject->data  = *RootCast(ptrFr->data);
        if(ptrFr->table) rootObject->table = *RootCast(ptrFr->table);
    
        return rootObject.release();
}

FrameFormat::TFrHistory *RootCast(FrHistory *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrHistory>();
             rootObject->name    = std::string(((FrHistory*) ptrFr)->name);
             rootObject->time    = ((FrHistory*) ptrFr)->time;
             rootObject->comment = std::string(((FrHistory*) ptrFr)->comment);
        
        return rootObject.release();
}

FrameFormat::TFrMsg *RootCast(FrMsg *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrMsg>();
             rootObject->alarm    = std::string(((FrMsg*) ptrFr)->alarm ? ((FrMsg*) ptrFr)->alarm : "");
             rootObject->message  = std::string(((FrMsg*) ptrFr)->message ? ((FrMsg*) ptrFr)->message : "");
             rootObject->severity = ((FrMsg*) ptrFr)->severity;
             rootObject->GTimeS   = ((FrMsg*) ptrFr)->GTimeS;
             rootObject->GTimeN   = ((FrMsg*) ptrFr)->GTimeN;

        return rootObject.release();
}

FrameFormat::TFrProcData *RootCast(FrProcData *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrProcData>();
             rootObject->name         = std::string(((FrProcData*) ptrFr)->name ? ((FrProcData*) ptrFr)->name : "");
             rootObject->comment      = std::string(((FrProcData*) ptrFr)->comment ? ((FrProcData*) ptrFr)->comment : "");
             rootObject->type         = static_cast<FrameFormat::TFrProcData::TYPE>   (((FrProcData*) ptrFr)->type);
             rootObject->subType      = static_cast<FrameFormat::TFrProcData::SUBTYPE>(((FrProcData*) ptrFr)->subType);
             rootObject->timeOffset   = ((FrProcData*) ptrFr)->timeOffset;
             rootObject->tRange       = ((FrProcData*) ptrFr)->tRange;
             rootObject->fShift       = ((FrProcData*) ptrFr)->fShift;
             rootObject->phase        = ((FrProcData*) ptrFr)->phase;
             rootObject->fRange       = ((FrProcData*) ptrFr)->fRange;
             rootObject->BW           = ((FrProcData*) ptrFr)->BW;
     
             rootObject->nAuxParam     = ((FrProcData*) ptrFr)->nAuxParam;
             rootObject->auxParam      = std::vector<double>(((FrProcData*) ptrFr)->auxParam     , ((FrProcData*) ptrFr)->auxParam      + ((FrProcData*) ptrFr)->nAuxParam);
             rootObject->auxParamNames = std::vector<std::string>(((FrProcData*) ptrFr)->auxParamNames, ((FrProcData*) ptrFr)->auxParamNames + ((FrProcData*) ptrFr)->nAuxParam);

        if(ptrFr->data)    rootObject->data = *RootCast(ptrFr->data);
        if(ptrFr->aux)     rootObject->aux  = *RootCast(ptrFr->aux);
        if(ptrFr->table)   rootObject->table   = *RootCast(ptrFr->table);
        if(ptrFr->history) rootObject->history = *RootCast(ptrFr->history);
        
        return rootObject.release();
}

FrameFormat::TFrRawData *RootCast(FrRawData *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrRawData>();
             rootObject->name       = std::string(((FrRawData*) ptrFr)->name ? ((FrRawData*) ptrFr)->name : "");
        
        if(ptrFr->logMsg) rootObject->logMsg = *RootCast(ptrFr->logMsg);
        if(ptrFr->more)   rootObject->more   = *RootCast(ptrFr->more);

        // rootObject->firstTable = RootCast((FrTable*) ((FrRawData*) ptrFr)->firstTable);
        // rootObject->firstSer   = RootCast((FrSerData*) ((FrRawData*) ptrFr)->firstSer); // We dont keep first/next structure
        // rootObject->firstAdc   = RootCast((FrAdcData*) ((FrRawData*) ptrFr)->firstAdc);

        return rootObject.release();
}

FrameFormat::TFrSerData *RootCast(FrSerData *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrSerData>();
             rootObject->name       = std::string(((FrSerData*) ptrFr)->name ? ((FrSerData*) ptrFr)->name : "");
             rootObject->timeSec    = ((FrSerData*) ptrFr)->timeSec;
             rootObject->timeNsec   = ((FrSerData*) ptrFr)->timeNsec;
             rootObject->sampleRate = ((FrSerData*) ptrFr)->sampleRate;
             rootObject->data       = ((FrSerData*) ptrFr)->data;

        if(ptrFr->serial) rootObject->serial = *RootCast(ptrFr->serial);
        if(ptrFr->table)  rootObject->table  = *RootCast(ptrFr->table);
        
        return rootObject.release();
}

FrameFormat::TFrSimData *RootCast(FrSimData *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrSimData>();
             rootObject->name       = std::string(((FrSimData*) ptrFr)->name ? ((FrSimData*) ptrFr)->name : "");
             rootObject->comment    = std::string(((FrSimData*) ptrFr)->comment ? ((FrSimData*) ptrFr)->comment : "");
             rootObject->sampleRate = ((FrSimData*) ptrFr)->sampleRate;
             rootObject->timeOffset = ((FrSimData*) ptrFr)->timeOffset;
             rootObject->fShift     = ((FrSimData*) ptrFr)->fShift;
             rootObject->phase      = ((FrSimData*) ptrFr)->phase;

        if(ptrFr->data)  rootObject->data  = *RootCast(ptrFr->data);
        if(ptrFr->input) rootObject->input = *RootCast(ptrFr->input);
        if(ptrFr->table) rootObject->table = *RootCast(ptrFr->table);

        return rootObject.release();
}

FrameFormat::TFrSimEvent *RootCast(FrSimEvent *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrSimEvent>();
             rootObject->name    = std::string(ptrFr->name ? ptrFr->name : "");
             rootObject->comment = std::string(ptrFr->comment ? ptrFr->comment : "");
             rootObject->inputs  = std::string(ptrFr->inputs ? ptrFr->inputs : "");
             rootObject->GTimeS  = ptrFr->GTimeS;
             rootObject->GTimeN  = ptrFr->GTimeN;
     
             rootObject->timeBefore = ptrFr->timeBefore;
             rootObject->timeAfter  = ptrFr->timeAfter;
             rootObject->amplitude  = ptrFr->amplitude;
     
             rootObject->nParam         = ptrFr->nParam;
             rootObject->parameters     = std::vector<double>(ptrFr->parameters, ptrFr->parameters + ptrFr->nParam);
             rootObject->parameterNames = std::vector<std::string>(ptrFr->parameterNames, ptrFr->parameterNames + ptrFr->nParam);

        if(ptrFr->data)  rootObject->data  = *RootCast(ptrFr->data);
        if(ptrFr->table) rootObject->table = *RootCast(ptrFr->table);

        return rootObject.release();
}

FrameFormat::TFrStatData *RootCast(FrStatData *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrStatData>();
             rootObject->name           = std::string(((FrStatData*) ptrFr)->name ? ((FrStatData*) ptrFr)->name : "");
             rootObject->comment        = std::string(((FrStatData*) ptrFr)->comment ? ((FrStatData*) ptrFr)->comment : "");
             rootObject->representation = std::string(((FrStatData*) ptrFr)->representation ? ((FrStatData*) ptrFr)->representation : "");
             rootObject->timeStart      = ((FrStatData*) ptrFr)->timeStart;
             rootObject->timeEnd        = ((FrStatData*) ptrFr)->timeEnd;
             rootObject->version        = ((FrStatData*) ptrFr)->version;

        if(ptrFr->detector) rootObject->detector = *RootCast(ptrFr->detector);
        if(ptrFr->data)     rootObject->data  = *RootCast(ptrFr->data);
        if(ptrFr->table)    rootObject->table = *RootCast(ptrFr->table);

        return rootObject.release();
}

FrameFormat::TFrSummary *RootCast(FrSummary *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrSummary>();
             rootObject->name    = std::string(((FrSummary*) ptrFr)->name ? ((FrSummary*) ptrFr)->name : "");
             rootObject->comment = std::string(((FrSummary*) ptrFr)->comment ? ((FrSummary*) ptrFr)->comment : "");
             rootObject->test    = std::string(((FrSummary*) ptrFr)->test ? ((FrSummary*) ptrFr)->test : "");
             rootObject->GTimeS  = ((FrSummary*) ptrFr)->GTimeS;
             rootObject->GTimeN  = ((FrSummary*) ptrFr)->GTimeN;
 
        if(ptrFr->moments) rootObject->moments = *RootCast(ptrFr->moments);
        if(ptrFr->table)   rootObject->table   = *RootCast(ptrFr->table);

        return rootObject.release();
}

FrameFormat::TFrTable *RootCast(FrTable *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrTable>();
             rootObject->name       = std::string(((FrTable*) ptrFr)->name ? ((FrTable*) ptrFr)->name : "");
             rootObject->comment    = std::string(((FrTable*) ptrFr)->comment ? ((FrTable*) ptrFr)->comment : "");

             rootObject->nColumn    = ((FrTable*) ptrFr)->nColumn;
             rootObject->nRow       = ((FrTable*) ptrFr)->nRow;
             rootObject->columnName = std::vector<std::string> (((FrTable*) ptrFr)->columnName , ((FrTable*) ptrFr)->columnName  + ((FrTable*) ptrFr)->nColumn);

        if(ptrFr->column) rootObject->column = *RootCast(ptrFr->column);

        return rootObject.release();
}

FrameFormat::TFrVect *RootCast(FrVect *ptrFr)
{
        if(ptrFr == nullptr) return nullptr;

        auto rootObject = std::make_unique<FrameFormat::TFrVect>();
             rootObject->name     = std::string(ptrFr->name ? ptrFr->name : "");
             rootObject->compress = ptrFr->compress;
             rootObject->type     = static_cast<FrameFormat::TFrVect::TYPE> (ptrFr->type);

             rootObject->nDim     = ptrFr->nDim;
             rootObject->nBytes   = ptrFr->nBytes;
             rootObject->nData    = ptrFr->nData;

        if(ptrFr->nData > 0) {

                if (ptrFr->nDim > 0) {

                        rootObject->nx      = std::vector<uint>  (ptrFr->nx    , ptrFr->nx     + ptrFr->nDim);
                        rootObject->dx      = std::vector<double>(ptrFr->dx    , ptrFr->dx     + ptrFr->nDim);
                        rootObject->startX  = std::vector<double>(ptrFr->startX, ptrFr->startX + ptrFr->nDim);
                        rootObject->unitX   = std::vector<std::string> (ptrFr->unitX , ptrFr->unitX  + ptrFr->nDim);
                }

                if (ptrFr->nBytes > 0) 
                        rootObject->data = FrReadData(ptrFr);

                rootObject->unitY    = std::string(ptrFr->unitY ? ptrFr->unitY : "");
        }

        return rootObject.release();
}







FrDetector *FrCast(FrameFormat::TFrDetector *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());;
        FrDetector *ptrFr = FrDetectorNew("L1");
                    ptrFr->prefix[0]    = rootObject->prefix[0];
                    ptrFr->prefix[1]    = rootObject->prefix[1];
                    ptrFr->longitude    = rootObject->longitude;
                    ptrFr->latitude     = rootObject->latitude;
                    ptrFr->elevation    = rootObject->elevation;
                    ptrFr->armXazimuth  = rootObject->armXazimuth;
                    ptrFr->armYazimuth  = rootObject->armYazimuth;
                    ptrFr->armXaltitude = rootObject->armXaltitude;
                    ptrFr->armYaltitude = rootObject->armYaltitude;
                    ptrFr->armXmidpoint = rootObject->armXmidpoint;
                    ptrFr->armYmidpoint = rootObject->armYmidpoint;

                    ptrFr->localTime = rootObject->localTime;

        delete ptrFr->aux;
        ptrFr->aux   = (FrVect*)  FrCast(&(rootObject->aux));

        delete ptrFr->table;
        ptrFr->table = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        return ptrFr;
}

FrEvent *FrCast(FrameFormat::TFrEvent *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        FrEvent *ptrFr = FrEventNew(
                NULL, name, NULL, NULL, 
                gClockGPS->Timestamp(rootObject->GTimeS, rootObject->GTimeN),
                rootObject->timeBefore, rootObject->timeAfter,
                rootObject->amplitude, rootObject->probability, NULL, NULL, 0);

        ptrFr->comment     = strdup(rootObject->comment.c_str());
        ptrFr->inputs      = strdup(rootObject->inputs.c_str());
        ptrFr->eventStatus = rootObject->eventStatus;
        ptrFr->statistics  = strdup(rootObject->statistics.c_str());

        ptrFr->nParam = rootObject->nParam;
        ptrFr->parameters = new double[rootObject->nParam];
        std::copy(rootObject->parameters.begin(), rootObject->parameters.end(), ptrFr->parameters);

        ptrFr->parameterNames = new char*[rootObject->nParam];
        for(int i = 0; i < rootObject->nParam; ++i) {
                ptrFr->parameterNames[i] = strdup(rootObject->parameterNames[i].c_str());
        }

        delete ptrFr->data;
        ptrFr->data  = (FrVect*) FrCast(&(rootObject->data));

        delete ptrFr->table;
        ptrFr->table = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        return ptrFr;
}

FrHistory *FrCast(FrameFormat::TFrHistory *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *comment = strdup(rootObject->comment.c_str());
        FrHistory *ptrFr = FrHistoryAdd(NULL, comment);
                   ptrFr->name    = strdup(rootObject->name.c_str());
                   ptrFr->time    = rootObject->time;
        
        delete[] comment;
        return ptrFr;
}

FrMsg *FrCast(FrameFormat::TFrMsg *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *alarm    = strdup(rootObject->alarm.c_str());
        char *message  = strdup(rootObject->message.c_str());

        FrMsg *ptrFr = FrMsgAdd(NULL, alarm, message, rootObject->severity);
               ptrFr->GTimeS   = rootObject->GTimeS;
               ptrFr->GTimeN   = rootObject->GTimeN;

        delete[] alarm;
        delete[] message;
        return ptrFr;
}

FrProcData *FrCast(FrameFormat::TFrProcData *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        FrProcData *ptrFr = FrProcDataNew(NULL, name, rootObject->data.dx[0] ? 1./rootObject->data.dx[0] : 0, rootObject->data.nData, FrBits(rootObject->data.type));

        ptrFr->comment     = strdup(rootObject->comment.c_str());
        ptrFr->type        = static_cast<int>(rootObject->type);
        ptrFr->subType     = static_cast<int>(rootObject->subType);
        ptrFr->timeOffset  = rootObject->timeOffset;
        ptrFr->tRange      = rootObject->tRange;
        ptrFr->fShift      = rootObject->fShift;
        ptrFr->phase       = rootObject->phase;
        ptrFr->fRange      = rootObject->fRange;
        ptrFr->BW          = rootObject->BW;

        ptrFr->nAuxParam   = rootObject->nAuxParam;
        ptrFr->auxParam    = new double[rootObject->nAuxParam];
        std::copy(rootObject->auxParam.begin(), rootObject->auxParam.end(), ptrFr->auxParam);

        ptrFr->auxParamNames = new char*[rootObject->nAuxParam];
        for(int i = 0; i < rootObject->nAuxParam; ++i) {
                ptrFr->auxParamNames[i] = strdup(rootObject->auxParamNames[i].c_str());
        }

        delete ptrFr->data;
        ptrFr->data  = (FrVect*)  FrCast(&(rootObject->data));

        delete ptrFr->aux;
        ptrFr->aux   = (FrVect*)  FrCast(&(rootObject->aux));

        delete ptrFr->table;
        ptrFr->table = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        return ptrFr;
}

FrRawData *FrCast(FrameFormat::TFrRawData *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        FrRawData *ptrFr = FrRawDataNew(NULL);
        ptrFr->name = strdup(rootObject->name.c_str());
        ptrFr->logMsg  = (FrMsg*)  FrCast(&(rootObject->logMsg));
        ptrFr->more    = (FrVect*) FrCast(&(rootObject->more));

        return ptrFr;
}

FrSerData *FrCast(FrameFormat::TFrSerData *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        FrSerData *ptrFr = FrSerDataNew(NULL, name, gClockGPS->Timestamp(rootObject->timeSec, rootObject->timeNsec), "", rootObject->sampleRate);

        ptrFr->sampleRate = rootObject->sampleRate;
        ptrFr->data       = strdup(rootObject->data.c_str());

        delete ptrFr->serial;
        ptrFr->serial = (FrVect*)  FrCast(&(rootObject->serial));

        delete ptrFr->table;
        ptrFr->table  = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        return ptrFr;
}

FrSimData *FrCast(FrameFormat::TFrSimData *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        FrSimData *ptrFr = FrSimDataNew(NULL, name, rootObject->sampleRate, rootObject->data.nData, FrBits(rootObject->data.type));
        
        ptrFr->comment    = strdup(rootObject->comment.c_str());
        ptrFr->sampleRate = rootObject->sampleRate;
        ptrFr->timeOffset = rootObject->timeOffset;
        ptrFr->fShift     = rootObject->fShift;
        ptrFr->phase      = rootObject->phase;

        delete ptrFr->data;
        ptrFr->data  = (FrVect*)  FrCast(&(rootObject->data));

        delete ptrFr->input;
        ptrFr->input = (FrVect*)  FrCast(&(rootObject->input));

        delete ptrFr->table;
        ptrFr->table = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        return ptrFr;
}

FrSimEvent *FrCast(FrameFormat::TFrSimEvent *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        FrSimEvent *ptrFr = FrSimEventNew(
                NULL, name, NULL, NULL,
                gClockGPS->Timestamp(rootObject->GTimeS, rootObject->GTimeN),
                rootObject->timeBefore, rootObject->timeBefore,
                ptrFr->amplitude,0); 

                ptrFr->comment = strdup(rootObject->comment.c_str());
        ptrFr->inputs  = strdup(rootObject->inputs.c_str());

        ptrFr->timeBefore = rootObject->timeBefore;
        ptrFr->timeAfter  = rootObject->timeAfter;
        ptrFr->amplitude  = rootObject->amplitude;
        ptrFr->nParam = rootObject->nParam;

        ptrFr->parameters = new double[rootObject->nParam];
        std::copy(rootObject->parameters.begin(), rootObject->parameters.end(), ptrFr->parameters);

        ptrFr->parameterNames = new char*[rootObject->nParam];
        for(int i = 0; i < rootObject->nParam; ++i) {
                ptrFr->parameterNames[i] = strdup(rootObject->parameterNames[i].c_str());
        }

        delete ptrFr->data;        
        ptrFr->data  = (FrVect*) FrCast(&(rootObject->data));

        delete ptrFr->table;
        ptrFr->table = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        return ptrFr;
}

FrStatData *FrCast(FrameFormat::TFrStatData *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        FrStatData *ptrFr = FrStatDataNew(name, NULL, NULL, rootObject->timeStart, rootObject->timeEnd, rootObject->version, NULL, NULL);

        ptrFr->comment        = strdup(rootObject->comment.c_str());
        ptrFr->representation = strdup(rootObject->representation.c_str());
        ptrFr->timeStart      = rootObject->timeStart;
        ptrFr->timeEnd        = rootObject->timeEnd;
        ptrFr->version        = rootObject->version;

        delete ptrFr->detector;
        ptrFr->detector = (FrDetector*) FrCast(&(rootObject->detector));

        delete ptrFr->data;
        ptrFr->data     = (FrVect*) FrCast(&(rootObject->data));

        delete ptrFr->table;
        ptrFr->table    = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        return ptrFr;
}

FrSummary *FrCast(FrameFormat::TFrSummary *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        char *comment = strdup(rootObject->comment.c_str());
        
        FrSummary *ptrFr = FrSummaryNew(NULL,name,comment," ",NULL, NULL);
        ptrFr->test    = strdup(rootObject->test.c_str());
        ptrFr->GTimeS  = rootObject->GTimeS;
        ptrFr->GTimeN  = rootObject->GTimeN;

        delete ptrFr->moments;
        ptrFr->moments = (FrVect*) FrCast(&(rootObject->moments));

        delete ptrFr->table;
        ptrFr->table = (FrTable*) FrCast(&(rootObject->table));

        delete[] name;
        delete[] comment;
        return ptrFr;
}

FrTable *FrCast(FrameFormat::TFrTable *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name       = strdup(rootObject->name.c_str());
        FrTable *ptrFr = FrTableNew(NULL, name, NULL, 0 ,0);
                 ptrFr->comment    = strdup(rootObject->comment.c_str());
                 ptrFr->nColumn    = rootObject->nColumn;
                 ptrFr->nRow       = rootObject->nRow;
                 ptrFr->columnName = new char*[rootObject->nColumn];

        for (int i = 0; i < rootObject->nColumn; ++i) {
                ptrFr->columnName[i] = strdup(rootObject->columnName[i].c_str());
        }

        delete ptrFr->column;
        ptrFr->column = (FrVect*) FrCast(&(rootObject->column));

        delete[] name;
        return ptrFr;
}

FrAdcData *FrCast(FrameFormat::TFrAdcData *rootObject)
{
        if(rootObject == nullptr) return nullptr;

        char *name = strdup(rootObject->name.c_str());
        char *comment = strdup(rootObject->comment.c_str());
        
        FrAdcData *ptrFr = FrAdcDataNewF(
                NULL, name, comment,
                rootObject->channelGroup, rootObject->channelNumber,
                rootObject->nBits,
                rootObject->bias,
                rootObject->slope,
                strdup(rootObject->units.c_str()),
                rootObject->sampleRate,
                rootObject->data.data.size()
        );
        
        ptrFr->timeOffset    = rootObject->timeOffset;
        ptrFr->fShift        = rootObject->fShift;
        ptrFr->phase         = rootObject->phase;
        ptrFr->dataValid     = rootObject->dataValid;
        
        delete ptrFr->data; // free the default data, generated by FrAdcDataNewF
        ptrFr->data          = (FrVect*) FrCast(&(rootObject->data));

        delete ptrFr->aux; // free the default aux, generated by FrAdcDataNewF
        ptrFr->aux           = (FrVect*) FrCast(&(rootObject->aux));

        delete[] name;
        delete[] comment;

        return ptrFr;
}

FrVect *FrCast(FrameFormat::TFrVect *rootObject)
{
        if(rootObject == nullptr) return nullptr;
        
        FrVect *ptrFr = FrVectNew(int(rootObject->type), rootObject->nx, rootObject->dx, rootObject->unitX);
        ptrFr->compress = rootObject->compress;
        ptrFr->nData = rootObject->nData;
        ptrFr->nBytes = ptrFr->nData * ptrFr->wSize;
   
        if ( rootObject->nBytes > 0) {

                for(int i = 0; i < rootObject->nDim; ++i) {
                        ptrFr->startX[i] = rootObject->startX[i];
                        ptrFr->unitX[i] = strdup(rootObject->unitX[i].c_str());
                }

                std::vector<char> data;
                switch(ptrFr->type) {
                        case FR_VECT_C:
                                data = VcCast<char>(std::vector<char>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_2S:
                                data = VcCast<char>(std::vector<short>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_4R:
                                data = VcCast<char>(std::vector<float>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_8R:
                                data = VcCast<char>(std::vector<double>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_4S:
                                data = VcCast<char>(std::vector<int>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_8S:
                                data = VcCast<char>(std::vector<long>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_STRING:
                                data = VcCast<char>(std::vector<char>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_2U:
                                data = VcCast<char>(std::vector<unsigned short>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_4U:
                                data = VcCast<char>(std::vector<unsigned int>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_8U:
                                data = VcCast<char>(std::vector<unsigned long>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_1U:
                                data = VcCast<char>(std::vector<unsigned char>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_8C:
                                data = VcCast<char>(std::vector<std::complex<float>>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        case FR_VECT_16C:
                                data = VcCast<char>(std::vector<std::complex<double>>(rootObject->data.begin(), rootObject->data.end()));
                                break;
                        default:
                                throw std::runtime_error("Unsupported vector type");
                }

                for(int i = 0; i < data.size(); i++)
                        ptrFr->data[i] = data[i];
        }

        ptrFr->unitY    = strdup(rootObject->unitY.c_str());

        return ptrFr;
}

FrVect *FrVectNew(int type, std::vector<unsigned int> nx, std::vector<double> dx, std::vector<std::string> unitx)
{
        int nDim = nx.size();
        if (nDim != dx.size()) {
                FrError(3,"FrVectNew","nxi and dxi sizes should match");
                return(NULL);
        }

        unitx.resize(nDim, "");

        FrVect *vect;

        const char *unit;
        static int first = 0, wSize[FR_VECT_END];
        int i;

        if(first == 0)
        {first = 1;
        wSize[FR_VECT_4R]  = sizeof(float);
        wSize[FR_VECT_8R]  = sizeof(double);
        wSize[FR_VECT_C]   = sizeof(char);
        wSize[FR_VECT_1U]  = sizeof(char);
        wSize[FR_VECT_2S]  = sizeof(short);
        wSize[FR_VECT_2U]  = sizeof(short);
        wSize[FR_VECT_4S]  = sizeof(int);
        wSize[FR_VECT_4U]  = sizeof(int);
        wSize[FR_VECT_8S]  = sizeof(FRLONG);
        wSize[FR_VECT_8U]  = sizeof(FRLONG);
        wSize[FR_VECT_STRING] = sizeof(char *);
        wSize[FR_VECT_8C]   = 2*sizeof(float);
        wSize[FR_VECT_16C]  = 2*sizeof(double);
        wSize[FR_VECT_8H]   = sizeof(float);
        wSize[FR_VECT_16H]  = sizeof(double);}

        vect = (FrVect *) calloc(1,sizeof(FrVect));
        if(vect == NULL) return(NULL);
        vect->classe = FrVectDef();
        
        vect->compress = 0;
        if(type < 0) 
        vect->type = -type;
        else  vect->type =  type;
        vect->nDim = nDim;
        vect->nData = 1;
        vect->nx    = (FRULONG *) malloc(nDim*sizeof(FRULONG));
        vect->unitX = (char **) malloc(nDim*sizeof(char *));
        vect->dx    = (double *) malloc(nDim*sizeof(double));
        vect->startX= (double *) malloc(nDim*sizeof(double));
        if(vect->nx    == NULL || 
        vect->unitX == NULL || 
        vect->dx    == NULL || 
        vect->startX== NULL) 
        {FrError(3,"FrVectNew","malloc failed");
        return(NULL);}

        for(i=0; i<nDim; i++)
        {vect->nx[i] = nx[i];
        vect->dx[i] = dx[i];
        unit = (char *)unitx[i].c_str();
        FrStrCpy(&vect->unitX[i], unit);
        vect->nData = vect->nData*vect->nx[i];
        vect->startX[i] = 0.;}

        vect->wSize  = wSize[vect->type];
        vect->space  = vect->nData;
        vect->nBytes = vect->nData*vect->wSize;

        if(vect->nData > 0)
        {if(type >= 0)
        {vect->data = (char *) calloc(vect->nData, vect->wSize);}
        else {vect->data = (char *) malloc(vect->nBytes);}
        if(vect->data == NULL) 
        {FrError(3,"FrVectNew","calloc failed");
        return(NULL);}}
        else {vect->data = NULL;}

        FrVectMap(vect);

        vect->unitY = NULL;
        vect->next  = NULL;

        return(vect);
}

TString FrVectType(FrAdcData *ptrFr)
{
        return FrVectType(ptrFr->data);
}

TString FrVectType(FrVect *ptrFr)
{
        switch(ptrFr->type)
        {
                case FR_VECT_C:
                        return "char";
                case FR_VECT_2S:
                        return "short";
                case FR_VECT_4R:
                        return "float";
                case FR_VECT_8R:
                        return "double";
                case FR_VECT_4S:
                        return "int";
                case FR_VECT_8S:
                        return "long";
                case FR_VECT_STRING:
                        return "string";
                case FR_VECT_2U:
                        return "unsigned short";
                case FR_VECT_4U:
                        return "unsigned int";
                case FR_VECT_8U:
                        return "unsigned long";
                case FR_VECT_1U:
                        return "unsigned char";
                case FR_VECT_8C:
                        return "complex<float>";
                case FR_VECT_16C:
                        return "complex<double>";
                default:
                        throw std::runtime_error("Unsupported vector type");
        }

        return "";
}

std::vector<double> FrVectRead(FrVect *ptrFr)
{
        std::vector<char> bytes = std::vector<char>(ptrFr->data, ptrFr->data + ptrFr->nBytes);
        switch(ptrFr->type) {
                case FR_VECT_C:
                        return VcCast<char, double>(bytes);
                case FR_VECT_2S:
                        return VcCast<short, double>(bytes);
                case FR_VECT_4R:
                        return VcCast<float, double>(bytes);
                case FR_VECT_8R:
                        return VcCast<double, double>(bytes);
                case FR_VECT_4S:
                        return VcCast<int, double>(bytes);
                case FR_VECT_8S:
                        return VcCast<long, double>(bytes);
                case FR_VECT_STRING:
                        return VcCast<char, double>(bytes);
                case FR_VECT_2U:
                        return VcCast<unsigned short, double>(bytes);
                case FR_VECT_4U:
                        return VcCast<unsigned int, double>(bytes);
                case FR_VECT_8U:
                        return VcCast<unsigned long, double>(bytes);
                case FR_VECT_1U:
                        return VcCast<unsigned char, double>(bytes);
                case FR_VECT_8C:
                        return VcCast<std::complex<float>, double>(bytes);
                case FR_VECT_16C:
                        return VcCast<std::complex<double>, double>(bytes);
                default:
                        throw std::runtime_error("Unsupported vector type");
        }
}

int FrBits(FrameFormat::TFrVect::TYPE type) {
        
        int nBits;
        switch(type) {
                case FrameFormat::TFrVect::TYPE::FR_VECT_4S:
                        nBits = 32;
                        break;
                case FrameFormat::TFrVect::TYPE::FR_VECT_2S:
                        nBits = 16;
                        break;
                case FrameFormat::TFrVect::TYPE::FR_VECT_C:
                        nBits = 8;
                        break;
                case FrameFormat::TFrVect::TYPE::FR_VECT_4R:
                        nBits = -32;
                        break;
                case FrameFormat::TFrVect::TYPE::FR_VECT_8R:
                        nBits = -64;
                        break;
                default:
                        nBits = 0;
                        break;
        }

        return nBits;
}