/**
 **********************************************
 *
 * \file TFrameManager.cc
 * \brief Source code of the TFrameManager class
 * \author Marco Meyer \<marco.meyer@cern.ch\>
 *
 *********************************************
 */

#include "FrameFormat/TFrameManager.h"
ClassImp(FrameFormat::TFrameManager)

#include "FrameFormat/Impl/TFrameHeader.h"
#include "FrameL.h"

#include <algorithm>
#include <vector>

namespace FrameFormat {
    
    TFrameManager::TFrameManager(TString fMasterName, std::vector<int> frameAllocation)
    {
        this->SetDebug(0);

        //
        // Prepare master file
        this->fMaster = FrFileINew(fMasterName);

        //
        // Allocate reading heads
        TString fName = ROOT::IOPlus::Helpers::Implode(" ", this->GetFileNames(this->fMaster));
        this->fHead.push_back(new TFrameRHead(0, FrFileINew(fName)));

        if( ::ROOT::IsImplicitMTEnabled() ) { // Additional reading heads are declared if implicit MT enabled
            
            TPrint::Debug(10, __METHOD_NAME__, (TString) "Frame library using parallel reading: %d core(s)", ::ROOT::GetThreadPoolSize());
            for(int i = 0; i < ::ROOT::GetThreadPoolSize(); i++) {
                this->fHead.push_back(new TFrameRHead(i+1, FrFileINew(fName)));
            }
        }
        
        this->SetDebug(GetVerbosity());

        //
        // Allocate memory for reading heads
        this->frame[this->fMaster] = std::vector<std::vector<FrameH*>>(frameAllocation.size(), std::vector<FrameH*>({}));
        for(int i = 0; i < frameAllocation.size(); i++)
            this->frame[this->fMaster][i] = std::vector<FrameH*>(frameAllocation[i], NULL);

        for(int h = 0; h < this->fHead.size(); h++) {
            this->frame[this->fHead[h]->GetFile()] = std::vector<std::vector<FrameH*>>(frameAllocation.size(), std::vector<FrameH*>({}));
            for(int i = 0; i < frameAllocation.size(); i++)
                this->frame[this->fHead[h]->GetFile()][i] = std::vector<FrameH*>(frameAllocation[i], NULL);
        }
    }

    TFrameManager::~TFrameManager()
    {
        //
        // Deallocate Memory
        for(int h = 0; h < this->fHead.size(); h++) {

            for(int i = 0; i < this->frame[this->fHead[h]->GetFile()].size(); i++) {
                for(int j = 0; j < this->frame[this->fHead[h]->GetFile()][i].size(); j++) {

                    this->Free(this->fHead[h]->GetFile(), i, j);
                }
            }
        }

        //
        // Deallocate Heads
        for(int i = 0, N = fHead.size(); i < N; i++) {

            if (fHead[i] != NULL) {
 
                FrFileIEnd(fHead[i]->GetFile()); 
                fHead[i] = NULL;
            }
        }

        fHead.clear();
    }

    FrFile &TFrameManager::GetMaster(int fileID, int frameID)
    {
        this->Seek(this->fMaster, fileID, frameID);
        return *this->fMaster;
    }

    TFrameRHead *TFrameManager::GetRHead(int headID)
    {
        int poolSize = TMath::Max(1, (int) ::ROOT::GetThreadPoolSize());
        int modulo = TMath::Min((int) this->fHead.size(), poolSize);

        headID = headID - modulo * TMath::Floor(headID / modulo);
        for(int i = 0, N = this->fHead.size(); i < N; i++) {

            if (this->fHead[i]->GetID() == headID)
                return this->fHead[i];
        }

        return NULL;
    }

    FrFile *TFrameManager::GetFile(TFrameRHead &head, int fileID, int frameID)
    {
        FrFile *fHead = head.GetFile();
        return this->Seek(fHead, fileID, frameID) ? fHead : NULL;
    }

    int TFrameManager::GetFileID(FrFile *file, FrFileH *fileHeader)
    {
        if(file == NULL) return -1;

        fileHeader = fileHeader ? fileHeader : file->current;
        if(fileHeader == NULL) return -1;

        FrFileH *_fileHeader = file->fileH;
        for(int i = 0; _fileHeader != NULL; i++) {

            if(_fileHeader == fileHeader) return i;
            _fileHeader = _fileHeader->next;
        }

        return -1;
    }

    int TFrameManager::GetFrameID(FrFile *file, FrameH *frameHeader) // FrameHeader search could be implemented (if needed)
    {
        if(frameHeader != NULL) 
            throw std::invalid_argument("Finding FrameH in FrFile is not implemented yet.");

        if(file == NULL) return -1;
        
        this->SetDebug(0);
        FrTOC *toc = FrTOCReadFull(file);
        this->SetDebug(this->GetVerbosity());
        if(file) {

            this->SetDebug(0);
            int tellp = FrFileIOTell(file);
            this->SetDebug(this->GetVerbosity());
            
            for(int frameID = 0; frameID < toc->nFrame; frameID++) {

                if(tellp == toc->positionH[frameID])
                    return frameID;
            }
        }

        return -1;
    }

    int TFrameManager::GetNFiles(FrFile *file)
    {
        int nFiles = 0;
        if(file == NULL) return nFiles;

        FrFileH *_fileHeader = file->fileH;
        for(int i = 0; _fileHeader != NULL; i++) {
            _fileHeader = _fileHeader->next;
            nFiles++;
        }

        return nFiles;
    }

    int TFrameManager::GetNFrames(FrFile *file, int fileID)
    {
        int nFrames = 0;
        if(file == NULL) return nFrames;

        FrFileH *_fileHeader = file->fileH;
        for(int i = 0; _fileHeader != NULL; i++) {

            if(fileID == i || fileID < 0) {

                FrTOC *toc = this->ReadTOC(file, fileID);
                if(toc == NULL) {
                
                    TPrint::Warning(__METHOD_NAME__, Form("Failed to read TOC file #%d", fileID+1));
                    continue; // Failed to load file.. @TODO an alternative loading procedure if TOC not found
                }

                nFrames += toc->nFrame;
            }

            if(fileID == i) break;
            _fileHeader = _fileHeader->next;
        }

        return nFrames;
    }

    bool TFrameManager::Rewind(FrFile *f, int fileID)
    {
        if(fileID < 0) this->Seek(f, fileID);

        int i = this->Tell(f, EIndexing::kFile);
        TPrint::Debug(100, __METHOD_NAME__, (TString) "Rewind to file #%d", i);
        return this->Seek(f, i, 0);
    }

    bool TFrameManager::PreviousFile(FrFile *f)
    {
        int i = this->Tell(f, EIndexing::kFile);
        return this->Seek(f, i-1, 0);
    }

    bool TFrameManager::Previous(FrFile *f)
    {
        int i = this->Tell(f, EIndexing::kFile);
        int j = this->Tell(f, EIndexing::kFrame);
        return this->Seek(f, i-1);
    }

    bool TFrameManager::Next(FrFile *f)
    {
        int i = this->Tell(f, EIndexing::kFile);
        int j = this->Tell(f, EIndexing::kFrame);
        return this->Seek(f, i, j+1);
    }

    bool TFrameManager::NextFile(FrFile *f)
    {
        int i = this->Tell(f, EIndexing::kFile);
        return this->Seek(f, i+1, 0);
    }

    GPS TFrameManager::GetGPS(FrFile *file)
    {
        if(file == NULL) return GPS();
        return GPS(file->startTime, file->closingTime);
    }

    GPS TFrameManager::GetGPS(FrFile *f, int fileID, int frameID)
    {
        GPS gps;
        if(f == NULL) return gps;

        if (fileID < 0) {
            fileID = this->GetFileID(f);
            frameID = frameID < 0 ? 0 : frameID;
        }

        if (frameID < 0) {
            frameID = this->GetFrameID(f);
        }
        
        FrTOC *toc = this->ReadTOC(f, fileID);
        if(toc == NULL) {
        
            TPrint::Warning(__METHOD_NAME__, Form("Failed to read TOC file #%d", fileID+1));
            return gps;
        }

        gps.start = gClockGPS->Timestamp(toc->GTimeS[frameID], toc->GTimeN[frameID]);
        gps.end = gps.start + toc->dt[frameID];

        return gps;
    }

    double TFrameManager::GetTimeLength(FrFile *file)
    {
        if(file == NULL) return 0;
        return file->closingTime - file->startTime;
    }

    double TFrameManager::GetTimeLength(FrFile *f, int fileID, int frameID)
    {
        if(f == NULL) return 0;

        if (fileID < 0) {
            fileID = this->GetFileID(f);
            frameID = frameID < 0 ? 0 : frameID;
        }
        
        if (frameID < 0) {
            frameID = this->GetFrameID(f);
        }
        
        FrTOC *toc = this->ReadTOC(f, fileID);
        if(toc == NULL) {
        
            TPrint::Warning(__METHOD_NAME__, Form("Failed to read TOC file #%d", fileID+1));
            return 0;
        }

        return toc->dt[frameID];
    }

    unsigned int TFrameManager::GetDataQuality  (FrFile *f, int fileID, int frameID)
    {
        if(f == NULL) return 0;

        if (fileID < 0) {
            fileID = this->GetFileID(f);
            frameID = frameID < 0 ? 0 : frameID;
        }
        
        if (frameID < 0) {
            frameID = this->GetFrameID(f);
        }
        
        FrTOC *toc = this->ReadTOC(f, fileID);
        if(toc == NULL) {
        
            TPrint::Warning(__METHOD_NAME__, Form("Failed to read TOC file #%d", fileID+1));
            return 0;
        }

        return toc->dataQuality[frameID];
    }

    bool TFrameManager::Seek(FrFile *f, int fileID, int frameID)
    {
        if(f == NULL) return false;

        if(TPrint::IsDebugEnabled(100)) TPrint::SetSingleCarriageReturn();
        if((fileID == this->GetFileID(f) && frameID == this->GetFrameID(f)) || (fileID < 0 && frameID < 0)) {

            TPrint::Debug(100, __METHOD_NAME__, (TString) "Pointer in place, no seek required (fileID, frameID) = (%d, %d)", fileID, frameID);
            return true;
        }

        TPrint::Debug(100, __METHOD_NAME__, (TString) "Seek for (fileID, frameID) = (%d, %d) << currently at (%d, %d)", fileID, frameID, this->GetFileID(f), this->GetFrameID(f));
        if(fileID < 0) fileID = this->GetFileID(f); // Select current file

        if(this->GetFrameID(f) < 0 || this->GetFileID(f) > fileID || this->GetFrameID(f) > frameID) {
            TPrint::Debug(100, __METHOD_NAME__, (TString) "Rewind to file #%d at frame %d << currently at (%d, %d)", fileID+1, frameID+1, this->GetFileID(f), this->GetFrameID(f));

            this->SetDebug(0);
            FrFileIRewind(f); // Rewind at the beginning of the file..
            this->SetDebug(this->GetVerbosity());            
        }

        int _fileID = fileID - this->GetFileID(f);
        while(_fileID-- > 0) {

            if(f->current == NULL) {
                return false;
            }
            if(f->current->next == NULL) {
                return false;
            }

            this->SetDebug(0);
            if(FrFileINext(f, 0., 0., NULL, FR_NO) != 0) {
                
                this->SetDebug(this->GetVerbosity());
                TPrint::Warning(__METHOD_NAME__, (TString) "Failed to seek header file #"+gPrint->Pad(fileID, this->GetNFiles()) + ": " + TPrint::MakeItShorter(f->current->fileName));    
                return false;
            }
            this->SetDebug(this->GetVerbosity());
        }

        if(frameID > 0) {

            this->SetDebug(0);
            FrTOC *toc = FrTOCReadFull(f);
            this->SetDebug(this->GetVerbosity());

            if(frameID >= toc->nFrame) {
                TPrint::Debug(100, __METHOD_NAME__, (TString) "Seek next file (fileID, frameID) = (%d, %d) << currently at (%d, %d)", fileID+1, frameID - toc->nFrame, this->GetFileID(f), this->GetFrameID(f));
                return this->Seek(f, fileID+1, frameID - toc->nFrame);
            }

            if(toc != NULL) {

                if(frameID != this->GetFrameID(f)) {

                    if(FrFileIOSet(f, toc->positionH[frameID]) == -1) {
                        TPrint::Error(__METHOD_NAME__, (TString) "Reading table of content.. failed to seek to (fileID, frameID) = (%d, %d)", fileID, frameID);
                    } else {
                        TPrint::Debug(100, __METHOD_NAME__, (TString) "Reading table of content.. going straight to frame #%d", frameID);
                    }
                }

            } else { // When no TOC..... no so good to do, but ok..

                if(this->GetFrameID(f) >= 0)
                    frameID -= this->GetFrameID(f);

                TPrint::Debug(10, __METHOD_NAME__, (TString) "No TOC found.. Reading frame one by one..");
                while(frameID-- > 0) {

                    TPrint::Debug(10, __METHOD_NAME__, (TString) "Reading frame #%d", frameID+1);
                    FrameReadT(f, this->GetGPS(f).start, " ");
                }
            }
        }

        return true;
    }

    int TFrameManager::Tell(FrFile *file, EIndexing repr)
    {
        switch(repr)
        {
            case EIndexing::kFile:
                return this->GetFileID(file);
            break;

            case EIndexing::kFrame:
                return this->GetFrameID(file);
            break;
        }

        return 0;
    }

    FrameH* TFrameManager::ReadHeader(FrFile *file, int fileID)
    {
        if(!this->Seek(file, fileID)) return NULL;

        this->SetDebug(0);
        FrameH *header = FrameHRead(file);
        this->SetDebug(this->GetVerbosity());

        return header;
    }

    FrTOC* TFrameManager::ReadTOC(FrFile *file, int fileID)
    {
        if(!this->Seek(file, fileID, 0)) return NULL;
     
        this->SetDebug(0);
        FrTOC *toc = FrTOCReadFull(file);
        this->SetDebug(this->GetVerbosity());

        return toc;
    }

    void TFrameManager::DumpTOC(FrFile *file, int fileID, std::vector<TString> tags)
    {
        return DumpTOC(file, fileID, ROOT::IOPlus::Helpers::Implode(" ", tags).Data());
    }

    void TFrameManager::DumpTOC(FrFile *file, int fileID, const char *tags)
    { 
        FrTOC *toc = this->ReadTOC(fileID);
        FrTOCDump(toc, stdout, this->GetVerbosity(), (char*) tags);
    }

    FrameH *TFrameManager::Read(FrFile *file, int fileID, int frameID, std::vector<TString> tags) 
    { 
        return Read(file, fileID, frameID, ROOT::IOPlus::Helpers::Implode(" ", tags).Data());
    }

    TString TFrameManager::GetUniqueID(int fileID, int frameID)
    {
        if(frameID < 0) return TString::Itoa(fileID,10);
        return TString::Itoa(fileID,10)+"-"+TString::Itoa(frameID,10);
    }

    TString TFrameManager::GetUniqueID(TFrameRHead &head, int fileID, int frameID)
    {
        return this->GetUniqueID(fileID, frameID)+"@"+TString::Itoa(head.GetID(),10);
    }

    bool TFrameManager::CheckBounds(FrFile *file, int fileID, int frameID)
    {
        if (file == nullptr || this->frame.find(file) == this->frame.end()) {
            TPrint::Error(__METHOD_NAME__, "Invalid file pointer while reading frame file (are you using correct frame manager ?).");
            return false;
        }
        if (fileID < 0 || fileID >= this->frame[file].size()) {
            TPrint::Error(__METHOD_NAME__, Form("File ID#%d is out of bounds.", fileID));
            return false;
        }
        if (frameID < 0 || frameID >= this->frame[file][fileID].size()) {
            TPrint::Error(__METHOD_NAME__, Form("Frame ID#%d is out of bounds.", frameID));
            return false;
        }

        return true;
    }

    FrameH *TFrameManager::Read(FrFile *file, int fileID, int frameID, const char *tags)
    {
        if(!this->CheckBounds(file, fileID, frameID)) {
            return NULL;
        }

        if(!this->Seek(file, fileID, frameID)) {
            return NULL;
        }

        if(!this->Find(file, fileID, frameID)) {
            this->frame[file][fileID][frameID] = FrameReadT(file, this->GetGPS(file, fileID, frameID).start, tags);
        }

        return this->frame[file][fileID][frameID];
    }

    FrEndOfFile* TFrameManager::ReadEOF(FrFile *file, int fileID)
    {
        if(!this->Rewind(file, fileID)) return NULL;
        return FrEndOfFileReadAndReturn(file);
    }

    bool TFrameManager::Find(FrFile *file, int fileID, int frameID)
    {
        return this->Get(file, fileID, frameID) != NULL;
    }

    FrameH* TFrameManager::Get(FrFile *file, int fileID, int frameID)
    {
        if (fileID < 0) {
            fileID = this->GetFileID(file);
            frameID = frameID < 0 ? 0 : frameID;
        }
        
        if (frameID < 0) {
            frameID = this->GetFrameID(file);
        }

        return this->frame[file][fileID][frameID];
    }

    void TFrameManager::Free(FrFile *file, int fileID, int frameID) { 

        for(int i = 0, I = this->GetNFiles(file); i < I; i++) {

            if(fileID != i && fileID >= 0) continue;
            for(int j = 0, J = this->GetNFrames(file, i); j < J; j++) {

                if(frameID != j && frameID >= 0) continue;
                if(this->frame[file][i][j] == NULL) continue;

                FrameFree(frame[file][i][j]);
                this->frame[file][i][j] = NULL;
            }
        }
    }

    void TFrameManager::Free(FrFile *file, FrameH *frame) { 

        if(frame == NULL) return;

        for(int i = 0, I = this->GetNFiles(file); i < I; i++) {

            for(int j = 0, J = this->GetNFrames(file, i); j < J; j++) {

                if(this->frame[file][i][j] != frame) continue;

                FrameFree(this->frame[file][i][j]);
                this->frame[file][i][j] = NULL;
            }
        }
    }

    void TFrameManager::Clear(FrFile *file, int fileID, int frameID) { 

        for(int i = 0, I = this->GetNFiles(file); i < I; i++) {

            if(fileID != i && fileID >= 0) continue;
            for(int j = 0, J = this->GetNFrames(file, i); j < J; j++) {

                if(frameID != j && frameID >= 0) continue;
                if(this->frame[file][i][j] == NULL) continue;

                this->frame[file][i][j] = NULL;
            }
        }
    }

    void TFrameManager::Clear(FrFile *file, FrameH*frame) { 

        if(frame == NULL) return;

        for(int i = 0, I = this->GetNFiles(file); i < I; i++) {

            for(int j = 0, J = this->GetNFrames(file, i); j < J; j++) {

                if(this->frame[file][i][j] != frame) continue;
                this->frame[file][i][j] = NULL;
            }
        }
    }

    int TFrameManager::Write(FrFile *oFile, TFrame *frame)
    {
        if(TPrint::ErrorIf(oFile == NULL, __METHOD_NAME__, (TString) "No output file provided. Cannot save the selected frame."))
            return FR_ERROR_NO_FILE;

        FrameH *frameH = this->Read(&frame->GetRHead(), frame->GetFileID(), frame->GetFrameID());
        if(frameH == NULL) return FR_ERROR_NO_FRAME;

        // Filter out entries that match the type TFrRawData
        std::vector<TFrAbstractT> buffer = frame->GetDataBuffer();
        buffer.erase(std::remove_if(buffer.begin(), buffer.end(), [](const TFrAbstractT& data) {
            return std::holds_alternative<TFrRawData*>(data);
        }), buffer.end());

        for (auto& data : buffer) {

            std::visit([&frameH](auto&& arg) {

                using T = std::remove_pointer<std::decay_t<decltype(arg)>>::type;
                if(arg == NULL) return;

                if constexpr (std::is_same_v<T, TFrMsg>) {
                    FrMsgAdd(frameH, (char*) arg->alarm.c_str(), (char*) arg->message.c_str(), arg->severity);
                } else if constexpr (std::is_same_v<T, TFrTable>) {
                    frameH->rawData->firstTable = FrCast(arg); // this might be a problem if multiple tables are added 
                } else if constexpr (std::is_same_v<T, TFrProcData>) {
                    FrameAddProc(frameH, FrCast(arg));
                } else if constexpr (std::is_same_v<T, TFrSimData>) {
                    FrameAddSimData(frameH, FrCast(arg));
                } else if constexpr (std::is_same_v<T, TFrEvent>) {
                    FrameAddEvent(frameH, FrCast(arg));
                } else if constexpr (std::is_same_v<T, TFrSimEvent>) {
                    FrameAddSimEvent(frameH, FrCast(arg));
                } else if constexpr (std::is_same_v<T, TFrSummary>) {
                    FrameAddSum(frameH, FrCast(arg));
                } else if constexpr (std::is_same_v<T, TFrAdcData>) {
                    FrameAddAdc(frameH, FrCast(arg));
                } else if constexpr (std::is_same_v<T, TFrSerData>) {
                    FrameAddSer(frameH, FrCast(arg));
                } else if constexpr (std::is_same_v<T, TFrStatData>) {
                    FrameAddStatData(frameH, NULL, FrCast(arg));
                } else {
                    TPrint::Error(__METHOD_NAME__, (TString) "Unsupported data type: " + TFrameProxy<T>::GetType());
                }

            }, data);
        }

        std::vector<FrType> dataClasses = frame->GetDataClasses();
        const char *tags = ROOT::IOPlus::Helpers::Implode(" ", frame->GetDataTags()).Data();

        if(dataClasses.size() == 0) FrameTag(frameH, const_cast<char *>(tags));
        else {

            for(int i = 0, N = dataClasses.size(); i < N; i++) {

                FrType dataClass = dataClasses[i];
                switch(dataClass)
                {
                    case FrType::FrMsg     : 
                        FrameTagMsg(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrProcData: 
                        FrameTagProc(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrSimData : 
                        FrameTagSim(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrEvent   : 
                        FrameTagEvent(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrSimEvent: 
                        FrameTagSimEvt(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrSummary : 
                        FrameTagSum(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrAdcData : 
                        FrameTagAdc(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrSerData : 
                        FrameTagSer(frameH, const_cast<char *>(tags));
                        break;
                    case FrType::FrStatData : 
                        FrameTagStat(frameH, const_cast<char *>(tags));
                        break;

                    default:
                        break;
                }
            }
        }

        frameH->dataQuality = frame->GetDataQuality();

        this->SetDebug(0);
        int ret = FrameWrite(frameH, oFile);
        this->SetDebug(this->GetVerbosity());

        frame->ClearDataBuffer();
        return ret;
    }

    FrTrendData* TFrameManager::Trend(int fs, FrAdcData* channel, GPS gpstime)
    {
        std::vector<FrTrendData*> trend = this->Trend(fs, std::vector<FrAdcData*>({channel}), gpstime);
        return trend.size() ? trend[0] : NULL;
    }

    std::vector<FrTrendData*> TFrameManager::Trend(int fs, std::vector<FrAdcData*> adc, GPS gpstime)
    {
        std::vector<FrTrendData*> trendList;

        FrTrendData *lastTrend = NULL;
        for (auto *channel : adc) {

            if (!channel || !channel->data) continue;

            double duration = channel->data->dx[0] * channel->data->nData;
            if (channel->data->startX[0] < gpstime.start && !ROOT::IOPlus::Helpers::IsNaN(gpstime.start)) continue;
            if (channel->data->startX[0] + duration > gpstime.end && !ROOT::IOPlus::Helpers::IsNaN(gpstime.end)) continue;

            FrTrendData *trend = FrTrendDataNew(fs, channel);
            if(trend == NULL) continue;

            if(lastTrend) lastTrend->next = trend;
            lastTrend = trend;

            trendList.push_back(trend);
        }

        return trendList;
    }

    FrameH *TFrameManager::Reshape(std::vector<TFrame*> frameCandidates, FrameH *frameH)
    {
        if (frameCandidates.empty()) {
            return frameH;
        }

        if (!frameH) {

            // Compute timestamp and dt based on frameCandidates
            double timestamp = NAN;
            double dt = NAN;

            for (auto *candidate : frameCandidates) {
                
                if (!candidate) continue;

                if (std::isnan(timestamp) || candidate->GetGPS().start < timestamp) {
                    timestamp = candidate->GetGPS().start;
                }
                
                if (std::isnan(dt) || candidate->GetTimeLength() < dt) {
                    dt = candidate->GetTimeLength();
                }
            }

            if (std::isnan(timestamp) || std::isnan(dt)) {
                return nullptr;
            }
        }

        double timestamp = frameH->GTimeS + 1e-9 * frameH->GTimeN;
        GPS frameGPS = GPS(timestamp, timestamp + frameH->dt);

        frameH = FrameReshapeNew(frameH, 1, 0);

        gPrint->Debug(1, __METHOD_NAME__, (TString) "Reshape frame within GPS %.2f and length of %.2fs", timestamp, frameH->dt);
        for (auto *candidate : frameCandidates) {

            if (!candidate) continue;

            GPS candidateGPS = candidate->GetGPS();
            if (!frameGPS.Overlaps(candidateGPS)) continue;
            
            FrameH *candidateH = candidate->GetManager().Read(&candidate->GetRHead(), candidate->GetFileID(), candidate->GetFrameID());
            FrameReshapeAdd(frameH, candidateH);
            
            candidate->GetManager().Clear(candidate->GetRHead(), candidate->GetFileID(), candidate->GetFrameID());
        }

        FrameReshapeEnd(frameH);
        return frameH;
    }

    FrameH *TFrameManager::Reshape(std::vector<TFrame*> frameCandidates, TFrame *frame)
    {
        FrameH *frameH = frame->GetManager().Read(&frame->GetRHead(), frame->GetFileID(), frame->GetFrameID());
        return Reshape(frameCandidates, frameH);
    }

    FrameH *TFrameManager::Reshape(std::vector<TFrame*> frameCandidates, double timestamp, double frameLength)
    {
        if (frameCandidates.empty()) {
            TPrint::Error(__METHOD_NAME__, "No frame candidates provided. Cannot reshape.");
            return nullptr;
        }

        FrameH *frame = FrameNew("");
        if (!frame) {
            TPrint::Error(__METHOD_NAME__, "Failed to create reshaped frame.");
            return nullptr;
        }

        frame->GTimeS = gClockGPS->TimeS(timestamp);
        frame->GTimeN = gClockGPS->TimeN(timestamp);
        frame->ULeapS = gClockGPS->GetLeapSeconds(timestamp);
        frame->dt = frameLength;

        return Reshape(frameCandidates, frame);
    }
};
