/**
 **********************************************
 *
 * \file TFrameProxy.cc
 * \brief Source code of the TFrameProxy class
 * \author Marco Meyer \<marco.meyer@cern.ch\>
 *
 *********************************************
 */

#include "FrameFormat/TFrameProxy.h"
#include "FrameFormat/TFrame.h"

ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrAdcData>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrDetector>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrEvent>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrHistory>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrMsg>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrProcData>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrRawData>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrSerData>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrSimData>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrSimEvent>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrStatData>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrSummary>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrTable>)
ClassImp(FrameFormat::TFrameProxy<FrameFormat::TFrVect>)

#include "FrameFormat/Impl/TFrameHeader.h"

/** @cond */
const char *FrameFormat::TFrameProxyAbstract::DEFAULT_BRANCHNAME = "payload";
/** @endcond */

namespace FrameFormat {

    TH1* TFrameProxyAbstract::Histogram(TTree *tree, TString branchName, Option_t *option)
    {
        TString opt = option;
                opt.ToLower();

        if(tree == NULL) return NULL;

        TString dataPath = branchName+".data";
        if(tree->GetListOfLeaves()->FindObject(dataPath) == NULL) {

            TPrint::Warning(__METHOD_NAME__, "No branch \""+branchName+"\" with \"data\" found in \""+TString(tree->GetName())+"\".");    
            return NULL;
        }

        std::vector<double> data = TVarexp(branchName+".data").GetSelection(tree)[0];
        if ( data.size() < 1 ) return NULL;

        std::vector<std::vector<double>> meta = TVarexp(branchName+".startX:"+branchName+".dx:"+branchName+".nx:GTimeS:GTimeN").GetSelection(tree);

        double startX = meta[0].size() ? meta[0][0] : NAN;
        double dx     = meta[1].size() ? meta[1][0] : NAN;
        int    nx     = meta[2].size() ? meta[2][0] : 0;
        double GTimeS = meta[3].size() ? meta[3][0] : NAN;
        double GTimeN = meta[4].size() ? meta[4][0] : NAN;

        double GPS = gClockGPS->Timestamp(GTimeS, GTimeN) + startX;
        
        std::vector<double> axis;
        for(int i = 0; i < nx+1; i++) 
            axis.push_back(GPS + i*dx);

        TString hName = "h"+ROOT::IOPlus::Helpers::Ucfirst(tree->GetName());
        if(opt.Contains("delete")) {

            TObject *obj = NULL;
            gDirectory->GetObject(hName, obj);

            if(obj) obj->Delete();
        }

        TH1 *hData = ROOT::IOPlus::Helpers::Histogram(hName, tree->GetTitle(), axis, data);
        return gClockGPS->FormatTimeAxis(hData, "", "UTC", 2);
    }

    template<typename T>
    TFrameProxy<T>::TFrameProxy(TFrame *frame, TString _identifier, void *_ptr): identifier(_identifier), frame(frame)
    {
        do {

            TString name      = this->ReadName(_ptr);
            TString message   = this->ReadMessage(_ptr);
            int channelGroup  = this->ReadChannelGroup(_ptr);
            int channelNumber = this->ReadChannelNumber(_ptr);
            double sampleRate = this->ReadSampleRate(_ptr);

            if(!frame->IsValidTag(name))
                continue;
            if(frame->GetChannelGroups().size() > 0 && std::find(frame->GetChannelGroups().begin(),   frame->GetChannelGroups().end(), channelGroup)  == frame->GetChannelGroups().end())
                continue;
            if(frame->GetChannelNumbers().size() > 0 && std::find(frame->GetChannelNumbers().begin(), frame->GetChannelNumbers().end(), channelNumber) == frame->GetChannelNumbers().end())
                continue;

            ptr.push_back(_ptr);
            names.push_back(name);
            messages.push_back(message);
            channelGroups.push_back(channelGroup);
            channelNumbers.push_back(channelNumber);
            sampleRates.push_back(sampleRate);

            if (std::is_same<T, TFrAdcData>::value && this->frame->GetDecimate() > 1)
                FrAdcDataDecimate((FrAdcData *) _ptr, this->frame->GetDecimate());

        } while((_ptr = this->Next(_ptr)));
    }

    template <typename T>
    void *TFrameProxy<T>::GetPointer(int basket)
    {
        if(basket < 0 || basket >= ptr.size()) return NULL;
        return ptr[basket];
    }

    template <typename T>
    int TFrameProxy<T>::GetBasketID(TString objName)
    {
        for(int i = 0, N = this->names.size(); i < N; i++) {
            
            if(this->names[i].EqualTo(objName)) return i;
        }

        return -1;
    }
             
    template <typename T>
    TString TFrameProxy<T>::GetName(int basket)
    {
        if(basket < 0 || basket >= names.size()) return "";
        return this->names[basket];
    }

    template <typename T>
    TString TFrameProxy<T>::GetMessage(int basket)
    {
        if(basket < 0 || basket >= messages.size()) return "";
        return this->messages[basket];
    }

    template <typename T>
    int TFrameProxy<T>::GetChannelGroup(int basket)
    {
        if(basket < 0 || basket >= channelGroups.size()) return -1;
        return this->channelGroups[basket];
    }

    template <typename T>
    int TFrameProxy<T>::GetChannelNumber(int basket)
    {
        if(basket < 0 || basket >= channelNumbers.size()) return -1;
        return this->channelNumbers[basket];
    }

    template <typename T>
    double TFrameProxy<T>::GetSampleRate(int basket)
    {
        if(basket < 0 || basket >= sampleRates.size()) return 0;
        return this->sampleRates[basket];
    }

    template <typename T>
    int TFrameProxy<T>::GetBasketID(void *ptr)
    {
        if(ptr == nullptr) return -1;

        int index = std::find(this->ptr.begin(), this->ptr.end(), ptr) - this->ptr.begin();
        if(index == this->ptr.size()) return -1;

        return index;
    }

    template<typename T>
    void *TFrameProxy<T>::Next(void *ptr)
    {
        if(ptr == NULL) return NULL;

        if (std::is_same<T, TFrAdcData>::value)
            return ((FrAdcData *) ptr)->next;
        if (std::is_same<T, TFrDetector>::value)
             return ((FrDetector*) ptr)->next;
        if (std::is_same<T, TFrEvent   >::value)
             return ((FrEvent   *) ptr)->next;
        if (std::is_same<T, TFrHistory >::value)
             return ((FrHistory *) ptr)->next;
        if (std::is_same<T, TFrMsg     >::value)
             return ((FrMsg     *) ptr)->next;
        if (std::is_same<T, TFrProcData>::value)
             return ((FrProcData*) ptr)->next;
        if (std::is_same<T, TFrRawData >::value)
             return NULL;
        if (std::is_same<T, TFrSerData >::value)
             return ((FrSerData *) ptr)->next;
        if (std::is_same<T, TFrSimData >::value)
             return ((FrSimData *) ptr)->next;
        if (std::is_same<T, TFrSimEvent>::value)
             return ((FrSimEvent*) ptr)->next;
        if (std::is_same<T, TFrStatData>::value)
             return NULL;
        if (std::is_same<T, TFrSummary >::value)
             return ((FrSummary *) ptr)->next;
        if (std::is_same<T, TFrTable   >::value)
             return ((FrTable   *) ptr)->next;
        if (std::is_same<T, TFrVect    >::value)
             return ((FrVect    *) ptr)->next;

        return NULL;
    }

    template <typename T>
    TString TFrameProxy<T>::GetUniqueID(int basket)
    {
        return this->frame->GetUniqueID() 
            + (identifier.EqualTo("") ? "" : ("-" + identifier))
            + (basket < 0 ? "" : "-"+TString::Itoa(basket,10));
    }

    template <typename T>
    void TFrameProxy<T>::Dump(int basket)
    {
        if (basket < 0) { 

            for(int basket = 0, B = this->GetNBaskets(); basket < B; basket++)
                this->Dump(basket);

            return;
        }

        void *ptr = this->GetPointer(basket);
        int debugLvl = this->GetVerbosity();

        if (std::is_same<T, TFrAdcData>::value)
            FrAdcDataDump((FrAdcData*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrDetector>::value)
            FrDetectorDump((FrDetector*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrEvent   >::value)
            FrEventDump((FrEvent*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrHistory >::value)
            fprintf(stdout," %s %s\n", FrStrGTime(((FrHistory*) ptr)->time), ((FrHistory*) ptr)->comment);
        if (std::is_same<T, TFrMsg     >::value)
            FrMsgDump((FrMsg *) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrProcData>::value)
            FrProcDataDump((FrProcData*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrSerData >::value)
            FrSerDataDump((FrSerData*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrSimData >::value)
            FrSimDataDump((FrSimData*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrSimEvent>::value)
            FrSimEventDump((FrSimEvent*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrRawData>::value)
            FrRawDataDump((FrRawData*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrStatData>::value)
            FrStatDataDump((FrStatData*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrSummary >::value)
            FrSummaryDump((FrSummary*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrTable   >::value)
            FrTableDump((FrTable*) ptr, stdout, debugLvl);
        if (std::is_same<T, TFrVect    >::value)
            FrVectDump((FrVect*) ptr, stdout, debugLvl);
    }

    template <typename T>
    void TFrameProxy<T>::Print(int basket)
    {
        if (basket < 0) {

            for(int basket = 0, B = this->GetNBaskets(); basket < B; basket++)
                this->Print(basket);

            return;
        }

        TString objType = this->GetType().ReplaceAll("FrameFormat::", "");
        if(objType.EqualTo("TFrRawData")) return; // Raw Data is a data holding structure

        TString spacer = "";
        TString title = objType + ".Print";
        TString channelInfo = "";

        if(this->Verbose(Verbosity::kInfo)) {

            if(this->GetChannelNumber(basket) >= 0) 
                channelInfo = "Channel=" + TString::Itoa(this->GetChannelNumber(basket),10);
            if(this->GetChannelGroup(basket) >= 0) {

                if(!channelInfo.EqualTo("")) channelInfo = " ";
                channelInfo = "Group=" + TString::Itoa(this->GetChannelGroup(basket),10);
            }
        
            if(!channelInfo.EqualTo("")) channelInfo = " ("+channelInfo+")";
    
        } else {
        
            if(this->GetChannelNumber(basket) >= 0) 
                channelInfo = "at channel " + TString::Itoa(this->GetChannelNumber(basket),10);
            if(this->GetChannelGroup(basket) >= 0) {
                if(!channelInfo.EqualTo("")) channelInfo = " ";
                channelInfo = "in group " + TString::Itoa(this->GetChannelGroup(basket),10);
            }

            if(!channelInfo.EqualTo("")) channelInfo = " "+channelInfo;
        }

        TString name = TPrint::kPurple + "Name " + TPrint::kNoColor + " = \"" + this->GetName(basket) + "\"";
        std::stringstream ss;

        int maxClassLength = TString("TFrProcData.Print").Length();
        spacer = ROOT::IOPlus::Helpers::Spacer(maxClassLength - title.Length());
        ss << TPrint::kPurple << title << spacer << " " << TPrint::kNoColor;
        
        spacer = ROOT::IOPlus::Helpers::Spacer(TPrint::CharCount(this->GetNBaskets()) - TPrint::CharCount(basket+1));
        ss << "#" << (basket+1) << "/" << this->GetNBaskets() << spacer;

        if(this->Verbose(Verbosity::kDebug)) ss << TPrint::kPurple << " at " << TPrint::kNoColor << "\"" << this->GetPointer(basket) << "\"" << channelInfo;
        
        std::cout << ss.str() << "; ";
        std::cout << name;

        TString msg = this->GetMessage(basket);
        if(!msg.EqualTo("")) {
            
            int maxNameLength = 0;
            for(int i = 0; i < this->names.size(); i++)
                maxNameLength = TMath::Max(maxNameLength, this->names[i].Length());

            spacer = ROOT::IOPlus::Helpers::Spacer(maxNameLength - this->GetName(basket).Length());
            std::cout << "; " << spacer << TPrint::kPurple << "Comment " << TPrint::kNoColor << " = \"" << msg << "\"";
        }

        std::cout << std::endl;
    }
                   
    template<typename T>
    void TFrameProxy<T>::Print(TString objName)
    {
        TString objType = this->GetType().ReplaceAll("FrameFormat::", "");

        for(int basket = 0, B = this->GetNBaskets(); basket < B; basket++) {

            if(!objName.EqualTo("") && !objName.EqualTo(this->GetName(basket))) continue;
            this->Print(basket);
        }
    }

    template<typename T>
    void TFrameProxy<T>::Print(T *obj)
    {
        TString objType = TFrameProxy<T>::GetType().ReplaceAll("FrameFormat::", "");
        if(obj == NULL) {
            std::cerr << "TFrameProxy<" << objType << ">::Print: Object is NULL" << std::endl;
            return;
        }

        TString title = objType + ".Print";
        TString name  = "Name = \"" + obj->name + "\"";
        std::cout << TPrint::kOrange << title << TPrint::kNoColor << "; (Uncommitted Object); " << name << std::endl;

        int debugLvl = GetVerbosity();
        if(debugLvl > 0) Dump(obj);
    }
     
    template<typename T>
    void TFrameProxy<T>::Dump(T *obj)
    {
        int debugLvl = GetVerbosity();
        
        if (std::is_same<T, TFrAdcData>::value)
            FrAdcDataDump((FrAdcData*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrDetector>::value)
            FrDetectorDump((FrDetector*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrEvent   >::value)
            FrEventDump((FrEvent*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrHistory >::value)
            fprintf(stdout," %s %s\n", FrStrGTime(((FrHistory*) FrCast(obj))->time), ((FrHistory*) FrCast(obj))->comment);
        if (std::is_same<T, TFrMsg     >::value)
            FrMsgDump((FrMsg*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrProcData>::value)
            FrProcDataDump((FrProcData*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrSerData >::value)
            FrSerDataDump((FrSerData*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrSimData >::value)
            FrSimDataDump((FrSimData*)FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrSimEvent>::value)
            FrSimEventDump((FrSimEvent*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrRawData>::value)
            FrRawDataDump((FrRawData*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrStatData>::value)
            FrStatDataDump((FrStatData*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrSummary >::value)
            FrSummaryDump((FrSummary*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrTable   >::value)
            FrTableDump((FrTable*) FrCast(obj), stdout, debugLvl);
        if (std::is_same<T, TFrVect    >::value)
            FrVectDump((FrVect*) FrCast(obj), stdout, debugLvl);
    }
     
    template<typename T>
    void TFrameProxy<T>::Dump(TString objName)
    {
        TString objType = this->GetType().ReplaceAll("FrameFormat::", "");
        if(objType.EqualTo("TFrRawData")) return; // Raw Data is a data holding structure

        for(int i = 0; i < this->GetNBaskets(); i++) {

            if(!objName.EqualTo("") && !objName.EqualTo(this->GetName(i))) continue;
            this->Dump(i);
        }
    }

    template<typename T>
    TString TFrameProxy<T>::GetType(bool demangle) {

        if(!demangle) {
            return typeid(T).name();
        }
        
        int status;
        char* demangled = abi::__cxa_demangle(typeid(T).name(), nullptr, nullptr, &status);

        if (status == 0) {

            TString result(demangled);
            free(demangled);
            return result;
        }
        
        return "";
    }

    template<typename T>
    TTree *TFrameProxy<T>::EmptyTree(TString fname, TString ftitle)
    {
        TTree *tree = new TTree(fname, ftitle);
               tree->SetAutoSave(0);

        // C : a character string terminated by the 0 character
        // B : an 1 bytes signed integer (Char_t)
        // b : an 1 bytes unsigned integer (UChar_t)
        // S : a 2 bytes signed integer (Short_t)
        // s : a 2 bytes unsigned integer (UShort_t)
        // I : a 4 bytes signed integer (Int_t)
        // i : a 4 bytes unsigned integer (UInt_t)
        // F : a 4 bytes floating point (Float_t)
        // f : a 3 bytes floating point with truncated mantissa (Float16_t)
        // D : a 8 bytes floating point (Double_t)
        // d : a 3 bytes truncated floating point (Double32_t)
        // L : a 8 bytes signed integer (Long64_t)
        // l : a 8 bytes unsigned integer (ULong64_t)
        // G : a long signed integer, stored as 8 bytes (Long_t)
        // g : a long unsigned integer, stored as 8 bytes (ULong_t)
        // O : [the letter o, not a zero] a boolean (Bool_t)

        tree->Branch("name",        &_buff_name);
        tree->Branch("run",         &_buff_run,         "run/I"        );
        tree->Branch("basket",      &_buff_basket,      "basket/I"     );
        tree->Branch("frame",       &_buff_frame,       "frame/i"      );
        tree->Branch("dataQuality", &_buff_dataQuality, "dataQuality/i");
        tree->Branch("GTimeS",      &_buff_GTimeS,      "GTimeS/i"     );
        tree->Branch("GTimeN",      &_buff_GTimeN,      "GTimeN/i"     );
        tree->Branch("ULeapS",      &_buff_ULeapS,      "ULeapS/s"     );
        tree->Branch("dt",          &_buff_dt,          "dt/D"         );

        TString objType = this->GetType();

        tree->Branch(TString(DEFAULT_BRANCHNAME)+".", objType, &_buff);
        return tree;
    }

    template<typename T>
    T* TFrameProxy<T>::Get(int basket)
    {
        return Hydrate(this->GetPointer(basket));
    }

    template<typename T>
    TObject* TFrameProxy<T>::GetObject(int basket)
    {
        TString name = this->GetName(basket);

        TTree *t = this->EmptyTree(name);
        if(t) TFrameProxy<T>::Hydrate(basket, t);

        return t;
    }

    template<typename T>
    TObject* TFrameProxy<T>::GetObject(int basket, TString branchName, Option_t *option)
    {
        TTree *tree = (TTree *) this->GetObject(this->GetName(basket));
        TH1 *hData = TFrameProxyAbstract::Histogram(tree, branchName, option);

        tree->Delete();

        return (TObject *) hData;
    }

    template<typename T>
    int TFrameProxy<T>::Hydrate(int basket, TTree *t)
    {
        if(t == nullptr) return FR_ERROR_NO_FILE;
        if(basket < 0 || basket >= ptr.size()) return FR_ERROR_NO_FRAME;

        _buff_basket      = basket;
        _buff_name        = this->frame->ReadName().Data();
        _buff_run         = this->frame->ReadRunNumber();
        _buff_frame       = this->frame->ReadFrameNumber();
        _buff_dataQuality = this->frame->GetDataQuality();
        _buff_GTimeS      = this->frame->ReadGTimeS();
        _buff_GTimeN      = this->frame->ReadGTimeN();    
        _buff_ULeapS      = this->frame->ReadULeapS();
        _buff_dt          = this->frame->ReadLength();
        _buff             = *Hydrate(ptr[basket]);
        
        t->Fill();
        return FR_OK;
    }

    template<typename T>
    T* TFrameProxy<T>::Hydrate(void* ptr)
    {
        if constexpr (std::is_same_v<T, TFrDetector>) {
            return RootCast((FrDetector*) ptr);
        } else if constexpr (std::is_same_v<T, TFrEvent>) {
            return RootCast((FrEvent*) ptr);
        } else if constexpr (std::is_same_v<T, TFrAdcData>) {
            return RootCast((FrAdcData*) ptr);
        } else if constexpr (std::is_same_v<T, TFrHistory>) {
            return RootCast((FrHistory*) ptr);
        } else if constexpr (std::is_same_v<T, TFrMsg>) {
            return RootCast((FrMsg*) ptr);
        } else if constexpr (std::is_same_v<T, TFrProcData>) {
            return RootCast((FrProcData*) ptr);
        } else if constexpr (std::is_same_v<T, TFrRawData>) {
            return RootCast((FrRawData*) ptr);
        } else if constexpr (std::is_same_v<T, TFrSerData>) {
            return RootCast((FrSerData*) ptr);
        } else if constexpr (std::is_same_v<T, TFrSimData>) {
            return RootCast((FrSimData*) ptr);
        } else if constexpr (std::is_same_v<T, TFrSimEvent>) {
            return RootCast((FrSimEvent*) ptr);
        } else if constexpr (std::is_same_v<T, TFrStatData>) {
            return RootCast((FrStatData*) ptr);
        } else if constexpr (std::is_same_v<T, TFrSummary>) {
            return RootCast((FrSummary*) ptr);
        } else if constexpr (std::is_same_v<T, TFrTable>) {
            return RootCast((FrTable*) ptr);
        } else if constexpr (std::is_same_v<T, TFrVect>) {
            return RootCast((FrVect*) ptr);
        } else {
            static_assert(sizeof(T) == 0, "Unsupported type in FrameProxy::Hydrate");
        }

        return nullptr;
    }

    template<typename T>
    void *TFrameProxy<T>::Dehydrate(T* rootObject) 
    {
        if constexpr (std::is_same_v<T, TFrAdcData>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrDetector>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrEvent>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrHistory>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrMsg>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrAdcData>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrProcData>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrRawData>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrSerData>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrSimData>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrSimEvent>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrStatData>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrSummary>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrTable>) {
            return FrCast(rootObject);
        } else if constexpr (std::is_same_v<T, TFrVect>) {
            return FrCast(rootObject);
        } else {
            static_assert(sizeof(T) == 0, "Unsupported type in FrameProxy::Dehydrate");
        }

        return nullptr;
    }

    template <typename T>
    TString TFrameProxy<T>::ReadName(void *ptr)
    {
        if (std::is_same<T, TFrAdcData>::value)
            return ((FrAdcData  *) ptr)->name;
        if (std::is_same<T, TFrDetector>::value)
             return ((FrDetector*) ptr)->name;
        if (std::is_same<T, TFrEvent   >::value)
             return ((FrEvent   *) ptr)->name;
        if (std::is_same<T, TFrHistory >::value)
             return ((FrHistory *) ptr)->name;
        if (std::is_same<T, TFrMsg>::value)
             return ((FrMsg     *) ptr)->alarm;
        if (std::is_same<T, TFrProcData>::value)
             return ((FrProcData*) ptr)->name;
        if (std::is_same<T, TFrRawData >::value)
             return ((FrRawData *) ptr)->name;
        if (std::is_same<T, TFrSerData >::value)
             return ((FrSerData *) ptr)->name;
        if (std::is_same<T, TFrSimData >::value)
             return ((FrSimData *) ptr)->name;
        if (std::is_same<T, TFrSimEvent>::value)
             return ((FrSimEvent*) ptr)->name;
        if (std::is_same<T, TFrStatData>::value)
             return ((FrStatData*) ptr)->name;
        if (std::is_same<T, TFrSummary >::value)
             return ((FrSummary *) ptr)->name;
        if (std::is_same<T, TFrTable   >::value)
             return ((FrTable   *) ptr)->name;
        if (std::is_same<T, TFrVect    >::value)
             return ((FrVect    *) ptr)->name;

        return "";
    }

    template <typename T>
    double TFrameProxy<T>::ReadSampleRate(void *ptr)
    {
        if (std::is_same<T, TFrAdcData>::value)
            return ((FrAdcData  *) ptr)->sampleRate;
        if (std::is_same<T, TFrDetector>::value)
             return 0;
        if (std::is_same<T, TFrEvent   >::value)
             return 0;
        if (std::is_same<T, TFrHistory >::value)
             return 0;
        if (std::is_same<T, TFrMsg>::value)
             return 0;
        if (std::is_same<T, TFrProcData>::value) {
            double dx = *(((FrProcData *) ptr)->data->dx);
            return dx ? 1./dx : 0;
        }
        
        if (std::is_same<T, TFrRawData >::value)
             return 0;
        if (std::is_same<T, TFrSerData >::value)
             return ((FrSerData *) ptr)->sampleRate;
        if (std::is_same<T, TFrSimData >::value)
             return ((FrSimData *) ptr)->sampleRate;
        if (std::is_same<T, TFrSimEvent>::value)
             return 0;
        if (std::is_same<T, TFrStatData>::value)
             return 0;
        if (std::is_same<T, TFrSummary >::value)
             return 0;
        if (std::is_same<T, TFrTable   >::value)
             return 0;
        if (std::is_same<T, TFrVect    >::value) {
            double dx = *(((FrVect    *) ptr)->dx);
            return dx ? 1./dx : 0;
        }

        return 0;
    }

    template <typename T>
    TString TFrameProxy<T>::ReadMessage(void *ptr)
    {
        if (std::is_same<T, TFrAdcData>::value)
            return ((FrAdcData  *) ptr)->comment;
        if (std::is_same<T, TFrDetector>::value) {
            TString message;
            if (((FrDetector*) ptr)->prefix[0] != 0)
                message += ((FrDetector*) ptr)->prefix[0];
            if (((FrDetector*) ptr)->prefix[1] != 0)
                message += ((FrDetector*) ptr)->prefix[1];
            return message;
        }
        if (std::is_same<T, TFrEvent   >::value)
            return ((FrEvent   *) ptr)->comment;
        if (std::is_same<T, TFrHistory >::value)
            return ((FrHistory *) ptr)->comment;
        if (std::is_same<T, TFrMsg>::value)
            return ((FrMsg     *) ptr)->message;
        if (std::is_same<T, TFrProcData>::value)
            return ((FrProcData*) ptr)->comment;
        if (std::is_same<T, TFrRawData >::value)
            return "";
        if (std::is_same<T, TFrSerData >::value)
            return ((FrSerData *) ptr)->data;
        if (std::is_same<T, TFrSimData >::value)
            return ((FrSimData *) ptr)->comment;
        if (std::is_same<T, TFrSimEvent>::value)
            return ((FrSimEvent*) ptr)->comment;
        if (std::is_same<T, TFrStatData>::value)
            return ((FrSimEvent*) ptr)->comment;
        if (std::is_same<T, TFrSummary >::value)
            return ((FrSummary *) ptr)->comment;
        if (std::is_same<T, TFrTable   >::value)
            return ((FrTable   *) ptr)->comment;
        if (std::is_same<T, TFrVect    >::value)
            return ((FrVect    *) ptr)->unitY;

        return "";
    }

    template <typename T>
    TString TFrameProxy<T>::GetBranchName()
    {
        if (std::is_same<T, TFrAdcData>::value)
            return TString(DEFAULT_BRANCHNAME)+".data";
        if (std::is_same<T, TFrDetector>::value)
            return TString(DEFAULT_BRANCHNAME)+".aux";
        if (std::is_same<T, TFrEvent   >::value)
            return TString(DEFAULT_BRANCHNAME)+".data";
        if (std::is_same<T, TFrHistory >::value)
            return "";
        if (std::is_same<T, TFrMsg>::value)
            return "";
        if (std::is_same<T, TFrProcData>::value)
            return TString(DEFAULT_BRANCHNAME)+".data";
        if (std::is_same<T, TFrRawData >::value)
            return TString(DEFAULT_BRANCHNAME)+".more";
        if (std::is_same<T, TFrSerData >::value)
            return TString(DEFAULT_BRANCHNAME)+".serial";
        if (std::is_same<T, TFrSimData >::value)
            return TString(DEFAULT_BRANCHNAME)+".data";
        if (std::is_same<T, TFrSimEvent>::value)
            return TString(DEFAULT_BRANCHNAME)+".data";
        if (std::is_same<T, TFrStatData>::value)
            return TString(DEFAULT_BRANCHNAME)+".data";
        if (std::is_same<T, TFrSummary >::value)
            return TString(DEFAULT_BRANCHNAME)+".moments";
        if (std::is_same<T, TFrTable   >::value)
            return TString(DEFAULT_BRANCHNAME)+".column";
        if (std::is_same<T, TFrVect    >::value)
            return "";

        return "";
    }

    template <typename T>
    int TFrameProxy<T>::ReadChannelNumber(void *ptr)
    {
        if (std::is_same<T, TFrAdcData>::value) return -1;
        
        return ((FrAdcData *) ptr)->channelNumber;
    }

    template <typename T>
    int TFrameProxy<T>::ReadChannelGroup(void *ptr)
    {
        if (std::is_same<T, TFrAdcData>::value) return -1;
        
        return ((FrAdcData *) ptr)->channelGroup;
    }

    template class TFrameProxy<TFrAdcData>;
    template class TFrameProxy<TFrDetector>;
    template class TFrameProxy<TFrEvent>;
    template class TFrameProxy<TFrHistory>;
    template class TFrameProxy<TFrMsg>;
    template class TFrameProxy<TFrProcData>;
    template class TFrameProxy<TFrRawData>;
    template class TFrameProxy<TFrSerData>;
    template class TFrameProxy<TFrSimData>;
    template class TFrameProxy<TFrSimEvent>;
    template class TFrameProxy<TFrStatData>;
    template class TFrameProxy<TFrSummary>;
    template class TFrameProxy<TFrTable>;
    template class TFrameProxy<TFrVect>;

}
