#include <TApplication.h>
#include <Rioplus.h>
#include <TLorentzVector.h>
#include <TTree.h>
#include <TLegend.h>
#include <TFile.h>
#include <TKey.h>
#include <TH1.h>
#include <TH2.h>
#include <TH3.h>
#include <TGraph.h>
#include <TCanvas.h>
#include <TStyle.h>
#include <TSystem.h>
#include <TEntryList.h>
#include <TROOT.h>
#include <TFormula.h>
#include <TClass.h>

#include <TFrameFormat.h>
#include <TCli.h>

using namespace std;

template <typename T>
std::vector<T> intersection(std::vector<T> v1, std::vector<T> v2)
{
        std::vector<T> v3;

        std::sort(v1.begin(), v1.end());
        std::sort(v2.begin(), v2.end());

        std::set_intersection(v1.begin(),v1.end(),
                                v2.begin(),v2.end(),
                                back_inserter(v3));
        return v3;
}

int main(int argc, char **argv)
{
        std::vector<TString> ff;
        Bool_t bTOC;
        bool bDump, bDraw;

        TString options;
        int compression;
        int dataQuality;

        TString select;
        std::vector<int> globalIDs;
        std::vector<double> gps;

        double gpsMin, gpsMax;
        double minDuration, maxDuration;
        
        std::vector<int> channels;
        std::vector<int> groups;
        
        int skip, max;

        TString output;
        bool bRootObject;
        bool bMore, bHisto, bInteractive;
        bool duplicates;
        std::vector<TString> dataClassStr;

        TString tags;
        int decimate;
        bool bOutputTree, bOverride;
        int framePerFiles;

        TCli& cmdline = TCli::Instance(&argc, argv, "verbose multithread graphics");

              cmdline.AddMessage("This script is used to read frame format file..");
              cmdline.AddOption("--toc", "Read table of content", bTOC);
              cmdline.AddOption("--histo", "Draw data structures", bHisto);
              cmdline.AddOption("--options", "Custom chain options", options);
              cmdline.AddOption("--compression", "Compression level (0..255)", compression, 0);
              cmdline.AddOption("--duplicates", "Consider duplicates", duplicates);
              cmdline.AddOption("--dump", "Dump header information", bDump);
              cmdline.AddOption("--more", "Print more information", bMore);
              cmdline.AddOption("--non-interactive", "Do not use interactive UI", bInteractive, true);
              cmdline.AddOption("--data-quality", "Data quality requirements", dataQuality, 0);

              cmdline.AddOption("--data-class  |--class", "Only consider specific Fr structures (e.g. FrAdcData, FrDetector)", dataClassStr);
              cmdline.AddOption("--channel-tags|--tags   |--name", "Only display specific structure names (you can use wildcard, multiple tags and anti-tags)", tags, "*");
              cmdline.AddOption("--adc-number  |--adc|-c", "Only display specific channels (use blank space as separator ` `; NB: Only for ADC)", channels);
              cmdline.AddOption("--adc-group   |--group  |-g", "Only display specific groups (use blank space as separator ` `; NB: Only for ADC)", groups);
              cmdline.AddOption("--adc-decimate|--decimate", "Decimate ADC channel values by a given factor (e.g. 1,2,3,4)", decimate, 1);

              cmdline.AddOption("--frame", "Select specific global frame IDs (range or list)", globalIDs);
              cmdline.AddOption("--gps", "Select specific GPS times (range or list)", gps);
              cmdline.AddOption("--gps-min|--gpsmin", "Minimal permitted GPS value", gpsMin, NAN);
              cmdline.AddOption("--gps-max|--gpsmax", "Maximum permitted GPS value", gpsMax, NAN);
              cmdline.AddOption("--min-duration", "Select frame with minimal duration", minDuration);
              cmdline.AddOption("--max-duration", "Select frame with maximal duration", maxDuration);
              cmdline.AddOption("--select", "Select specific frame (either range, list, or formula [x = global ID], [y = GPStime], [z = duration], and [t = data quality])", select, "");

              cmdline.AddOption("--root", "Turn structures into ROOT objects", bRootObject);
              cmdline.AddOption("--output", "Data saved into the output file", output);
              cmdline.AddOption("--output-override", "Override existing files", bOverride);
              cmdline.AddOption("--output-frames-per-file", "Maximum number of frame per file", framePerFiles, -1);
              cmdline.AddOption("--output-gps-hierarchy", "Output tree structure including truncated GPS time information", bOutputTree);
              
              cmdline.AddVariadic(ff, "test.gwf", 1);

        if(!cmdline.Usage() ) return 1;

        std::vector<FrameFormat::FrType> dataClass = FrameFormat::GetType(dataClassStr);

        // Extract filenames and tags
        for(int i = 0; i < ff.size(); i++) {

                TString fObject = TFileReader::Obj(ff[i]).Strip(TString::EStripType::kBoth, '/');
                if (fObject.EqualTo("")) continue;

                if (!fObject.EqualTo("") && !tags.EqualTo("*") && !tags.EqualTo(fObject.Data())) {
                        std::cerr << "Ambiguous tag provided. (`" << fObject << "` != `" << tags << "`)" << std::endl;
                        return 1;
                }

                ff[i] = TFileReader::FileName(ff[i]);
                if (!tags.EqualTo("*") && !fObject.EqualTo(tags)) {
                        gPrint->Error(__METHOD_NAME__, "Multiple tagging provided. Please check files and --tags option.");
                        return 1;
                }

                if (!fObject.EqualTo("")) tags = fObject;
        }
        
        bRootObject |= output.EndsWith(".root");

        // Prepare chain
        std::vector<FrameFormat::TFrameChain*> fChain;
        for(int i = 0; i < ff.size(); i++) {

                FrameFormat::Verbosity verbosity = FrameFormat::Verbosity::kNormal;
                if(gPrint->IsDebugEnabled((int) TCli::Verbosity::kDebug    )) verbosity = FrameFormat::kAll;
                else if(gPrint->IsDebugEnabled((int) TCli::Verbosity::kAll )) verbosity = FrameFormat::kDebug;
                else if(gPrint->IsDebugEnabled((int) TCli::Verbosity::kInfo)) verbosity = FrameFormat::kInfo;

                FrameFormat::TFrameChain *f = new FrameFormat::TFrameChain(ff[i].Data(), options + TString(duplicates ? "d" : ""), tags, dataClass, decimate, groups, channels, verbosity);
                if( TPrint::WarningIf(!f->IsOpen(), __METHOD_NAME__, "Failed to open `%s`", ff[i].Data()) )
                        continue;

                if (fChain.size())
                        fChain[0]->AddFriend(f);

                if( TPrint::ErrorIf(ff.size() > 0 && f == NULL, __METHOD_NAME__, "Failed to open main element `%s`", ff[0].Data()) )
                        return 1;

                bMore = (f->GetNFiles() == 1);

                //
                // Select and save data
                std::vector<int> selection = f->Select(select.EqualTo("") ? "true" : select);

                // Select data quality
                if(dataQuality != 0) {
                
                        std::vector<int> selectionDQ = f->SelectByDQ(dataQuality);
                        selection = intersection(selection, selectionDQ);
                }
                
                if(gps.size() > 0) { 

                        std::vector<int> selectionGPS;
                        switch(gps.size()) {

                                case 0: break;

                                case 1: selectionGPS = f->SelectByGPS(gps);
                                break;

                                case 2: selectionGPS = f->SelectByGPS(gps[0], gps[1]);
                                break;

                                default: selectionGPS = f->SelectByGPS(gps);
                        }

                        selection = intersection(selection, selectionGPS);
                } 

                if(!ROOT::IOPlus::Helpers::IsNaN(gpsMin) || !ROOT::IOPlus::Helpers::IsNaN(gpsMax)) {

                        std::vector<int> selectionGPS = f->SelectByGPS(gpsMin, gpsMax);
                        selection = intersection(selection, selectionGPS);
                }
                
                if(maxDuration > 0) { 

                        std::vector<int> selectionDuration = f->SelectByMaxDuration(maxDuration);
                                        selection = intersection(selection, selectionDuration);
                } 

                if(minDuration > 0) { 

                        std::vector<int> selectionDuration = f->SelectByMinDuration(minDuration);
                                        selection = intersection(selection, selectionDuration);
                } 

                if (globalIDs.size()) {

                        std::vector<int> selectionFrame;
                        switch(globalIDs.size()) {

                                case 0: break;

                                case 1: selectionFrame = f->SelectByFrame(globalIDs);
                                break;

                                case 2: selectionFrame = f->SelectByFrame(globalIDs[0], globalIDs[1]);
                                break;

                                default: selectionFrame = f->SelectByFrame(globalIDs);
                        }

                        selection = intersection(selection, selectionFrame);
                }

                bDraw = cmdline.HasGraphics();
                bInteractive |= bDraw;
                
                bMore |= !tags.EqualTo("*") || dataClass.size();
                if(bMore) options += " more";
                if(bDump) options += " dump";
                if(bOutputTree || bRootObject) options += " tree";
                if(bOverride) options += "recreate";

                f->Print(TString(bMore ? "read " : "") + TString(bDraw ? "draw " : "")+ TString(bDump ? "dump " : "")+ (bTOC ? "toc" : "")+ (bInteractive ? "interactive" : ""), selection);

                // One can maybe implement a single file merger ? (multiple detector --> one output)
                if(!output.EqualTo("")) f->Write(output, options, framePerFiles, selection);
                else if(!bMore) {

                        std::cout << std::endl;
                        if(!bMore && !bDump && !bTOC) {
                                TPrint::Message(__METHOD_NAME__, "Consider using --toc, --more, --selection or --dump, if you need to access more information");
                        }

                        TPrint::Message(__METHOD_NAME__, "You're all set.. Bye !");
                }

                std::cout << std::endl;
        }
        
        for(int i = 0; i < fChain.size(); i++)
                delete fChain[i];

        fChain.clear();
        return 0;
}
